#ifndef my_vector_h
#define my_vector_h

#include <cstdlib>
#include <iterator>
#include <new>

template <class T>
class my_vector
{
public:
        typedef T value_type;
        typedef T* iterator;
        typedef const T* const_iterator;
        typedef int difference_type;
        typedef int size_type;
private:
        static const size_type INIT = 16;
        T* a;
        size_type a_size, a_alloc;
public:
        my_vector() : a(0), a_size(0), a_alloc(0) {}

        my_vector(size_type n, const T& x) : a(0), a_size(0), a_alloc(0)
        {
                if (n == 0) return;
                a = static_cast<T*>(std::malloc(n * sizeof(T)));
                if (a == 0) throw std::bad_alloc();
                a_size = a_alloc = n;
                T* const aN = a + n;
                for (T* ai = a; ai != aN; ++ai) new(ai) T(x);
        }

        explicit my_vector(size_type n) : a(0), a_size(0), a_alloc(0)
        {
                new(this) my_vector(n, T());
        }

        template <class FI>
        my_vector(FI x0, FI xN) : a(0), a_size(0), a_alloc(0)
        {
                const size_type n = std::distance(x0, xN);
                if (n == 0) return;
                a = static_cast<T*>(std::malloc(n * sizeof(T)));
                if (a == 0) throw std::bad_alloc();
                a_size = a_alloc = n;
                T* ai = a;
                for (FI xi = x0; xi != xN; ++xi, ++ai) new(ai) T(*xi);
        }

        my_vector(const my_vector& v) : a(0), a_size(0), a_alloc(0)
        {
                new(this) my_vector(v.begin(), v.end());
        }

        ~my_vector()
        {
                T* const aN = a + a_size;
                for (T* ai = a; ai != aN; ++ai) ai->~T();
                free(a);
        }

        void clear()
        {
                T* const aN = a + a_size;
                for (T* ai = a; ai != aN; ++ai) ai->~T();
                free(a);
                a = 0; a_size = a_alloc = 0;
        }

        void assign(size_type n, const T& x)
        {
                clear();
                new(this) my_vector(n, x);
        }

        template <class FI>
        void assign(FI x0, FI xN)
        {
                clear();
                new(this) my_vector(x0, xN);
        }

        my_vector& operator=(const my_vector& v)
        {
                if (this != &v) {
                        clear();
                        new(this) my_vector(v.begin(), v.end());
                }
                return *this;
        }

        iterator begin() { return a; }
        iterator end() { return a + a_size; }
        const_iterator begin() const { return a; }
        const_iterator end() const { return a + a_size; }
        size_type size() const { return a_size; }
        bool empty() const { return a_size == 0; }

        void push_back(const T& x)
        {
                if (a_size == a_alloc) {
                        if (a_alloc == 0)
                                a_alloc = INIT;
                        else
                                a_alloc *= 2;
                                //a_alloc += a_alloc / 2;
                        T* t = static_cast<T*>(
                                std::realloc(a, a_alloc * sizeof(T))
                        );
                        if (t == 0) throw std::bad_alloc();
                        a = t;
                }
                new(a + a_size++) T(x);
        }

        void resize(size_type n)
        {
                if (a_alloc < n) {
                        a_alloc = n;
                        T* t = static_cast<T*>(
                                std::realloc(a, a_alloc * sizeof(T))
                        );
                        if (t == 0) throw std::bad_alloc();
                        a = t;
                }
                T* const aM = a + a_size;
                T* const aN = a + n;
                for (T* ai = aN; ai < aM; ++ai) ai->~T();
                for (T* ai = aM; ai < aN; ++ai) new(ai) T();
                a_size = n;
        }

        const T& operator[](size_type i) const { return a[i]; }
        T& operator[](size_type i) { return a[i]; }

        bool operator==(const my_vector& x) const
        {
                const_iterator p = begin(), q = x.begin(), r = end();
                while (p != r) {
                        if (*p == *q) ++p, ++q;
                        else return false;
                }
                return true;
        }

        bool operator<(const my_vector& x) const
        {
                const_iterator ap = begin(), aN = end();
                const_iterator bp = x.begin(), bN = x.end();
                while (ap != aN && bp != bN) {
                        if (*ap < *bp) return true;
                        else if (*bp < *ap) return false;
                        ++ap, ++bp;
                }
                if (ap != aN) return false;
                if (bp != bN) return true;
                return false;
        }

        bool operator>(const my_vector& x) const { return x < *this; }
        bool operator!=(const my_vector& x) const { return !(*this == x); }
        bool operator<=(const my_vector& x) const { return !(*this > x); }
        bool operator>=(const my_vector& x) const { return !(*this < x); }
};

#endif // my_vector_h
