#include <cstdlib>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <strings.h>
#include "StopWatch.hpp"
using namespace std;

#if !defined(__BLOCKS__)
template <typename T>
inline T flsl(T x)
{
  T n = 0;
  for (; x>1; x>>=1, n++);
  return n+1;
}
#endif

#define DEPTH(n) ((flsl(n)-1)*2)

inline void introsort_limiter(size_t n, size_t *depth)
{
  size_t d = 0, N = n;
  for (; N>1; N>>=1, ++d);
  d *= 2;
  *depth = d;
}
inline void pintrosort_limiter(size_t n, size_t *depth, size_t *n_threshold)
{
  size_t d = 0, N = n;
  for (; N>1; N>>=1, ++d);
  d *= 2;
  *depth = d;
  *n_threshold = size_t(sqrt(double(n)));
}

inline void qsort_limiter(size_t n, size_t *depth)
{
  *depth = 2*(flsl(n) - 1);
}
static
size_t isqrt(size_t x)
{
  size_t s = 1;
  s <<= flsl(x)/2;
  return (s + x/s)/2;
}
inline void psort_limiter(size_t n, size_t *depth, size_t *n_threshold)
{
  *depth = (flsl(n)-1)*2;
  *n_threshold = isqrt(n);
}

inline void introsort_r_limiter(size_t n, size_t *depth)
{
#if !defined(__BLOCKS__)
  size_t N = n, fls0 = 0;
  for (; N>1; N>>=1, ++fls0);
  *depth = fls0*2;
#else
  size_t fls = flsl(n);
  *depth = (fls-1)*2;
#endif
}
inline void pintrosort_r_limiter(size_t n, size_t *depth, size_t *n_threshold)
{
#if !defined(__BLOCKS__)
  size_t N = n, fls0 = 0, s = 1;
  for (; N>1; N>>=1, ++fls0);
  *depth = fls0*2;
  s <<= (fls0+1)/2;
  *n_threshold = (s + n/s)/2;
#else
  size_t fls = flsl(n), s = 1;
  *depth = (fls-1)*2;
  s <<= fls/2;
  *n_threshold = (s + n/s)/2;
#endif
}

template <typename T>
inline T fls0l(T x)
{
  T n = 0;
  for (; x>1; x>>=1, n++);
  return n;
}
template <typename S>
inline void intro_sort_limiter(S n, S &depth)
{
#if !defined(__BLOCKS__)
  depth = fls0l(n)*2;
#else
  depth = (flsl(n)-1)*2;
#endif
}

template <typename T>
inline T flsqrt(T x, T fls)
{
  T s = 1;
  s <<= fls/2;
  return (s + x/s)/2;
}
template <typename S>
inline void parallel_intro_sort_limiter(S n, S &depth, S &n_threshold)
{
#if !defined(__BLOCKS__)
  S fls0 = fls0l(n);
  depth = fls0*2;
  n_threshold = flsqrt(n, fls0+1);
#else
  S fls = flsl(n);
  depth = (fls-1)*2;
  n_threshold = flsqrt(n, fls);
#endif
}

#define check_limiter(n) (depth==size_t(DEPTH(n)))
#define check_limiter_parallel(n) (check_limiter(n) && n_threshold==isqrt(n))
#define do_bench_limiter_common(op) {\
  sw.start();\
  for (i=0; i<sizeof(d)/sizeof(d[0]); i++)\
    for (m=n=1; n<ULONG_MAX/d[i]; n*=d[i], m=n) op;\
  sw.stop();\
}
#define do_bench_limiter(op) {\
  size_t m;\
  do_bench_limiter_common(op);\
  n = m; op;\
  cout << left << setw(56) << #op << (check_limiter(n)?"":"!") << sw.get_ctime() << endl; \
  if (!check_limiter(n))\
    cout << n << "\t" <<\
      depth << "\t" <<\
      DEPTH(n) <<\
      endl;\
}
#define do_bench_limiter_parallel(op) {\
  size_t m;\
  do_bench_limiter_common(op);\
  n = m; op;\
  cout << left << setw(56) << #op << (check_limiter_parallel(n)?"":"!") << sw.get_ctime() << endl; \
  if (!check_limiter_parallel(n))\
    cout << n << "\t" <<\
      depth << "\t" <<\
      DEPTH(n) << "\t" <<\
      n_threshold << "\t" <<\
      isqrt(n) << "\t" <<\
      sqrt(n) <<\
      endl;\
}

int main()
{
  StopWatch sw;
  size_t n;
  size_t depth;
  size_t n_threshold;

  srandomdev();
  size_t i, d[1000000];
  for (i=0; i<sizeof(d)/sizeof(d[0]); i++)
    d[i] = 2 + random()%(10 - 2 + 1);

  do_bench_limiter(introsort_limiter(n, &depth));
  do_bench_limiter(qsort_limiter(n, &depth));
  do_bench_limiter_parallel(psort_limiter(n, &depth, &n_threshold));
  do_bench_limiter_parallel(pintrosort_limiter(n, &depth, &n_threshold));
  do_bench_limiter(introsort_r_limiter(n, &depth));
  do_bench_limiter_parallel(pintrosort_r_limiter(n, &depth, &n_threshold));
  do_bench_limiter(intro_sort_limiter(n, depth));
  do_bench_limiter_parallel(parallel_intro_sort_limiter(n, depth, n_threshold));
  return 0;
}
