/* Copyright (C) 1999 Lucent Technologies */
/* From 'Programming Pearls' by Jon Bentley */

/* sort.cpp -- test and time sorting algorithms
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include "sort_r.h"
#include "sort.hpp"
#include "quick_sort.cpp"
#include "pquick_sort.cpp"
#include "pquick_sort2.cpp"
#include "merge_sort.cpp"
#include "pmerge_sort.cpp"
#include "pmerge_sort2.cpp"
#include "mmerge_sort.cpp"
#include "heap_sort1.cpp"
#include "heap_sort2.cpp"
#include "heap_sort3.cpp"
#include "intro_sort.cpp"

int i_cmp(const void *a, const void *b)
{
  return ((*(int *)a < *(int *)b) ? -1 : (*(int *)a > *(int *)b) ? 1 : 0);
}
int i_cmp_r(void *thunk, const void *a, const void *b)
{
  return ((*(int *)a < *(int *)b) ? -1 : (*(int *)a > *(int *)b) ? 1 : 0);
}

// To change from C++ back to C, remove the following two lines
// and the call to sort in main
#include <algorithm>
using namespace std;

/* Data and supporting functions */

#define MAXN 10000000

typedef int DType;

DType realx[MAXN];
int *x = realx;	/* allow x to shift for heaps */
int n;

void swap(int i, int j)
{	DType t = x[i];
	x[i] = x[j];
	x[j] = t;
}
/*#define swap(i, j) { int I=i, J=j; DType t = x[I]; x[I] = x[J]; x[J] = t; }*/

int randint(int l, int u)
{	/*return l + (RAND_MAX*rand() + rand()) % (u-l+1);*/ /* buggy */
	return l + (int)((double)(u - l + 1)*rand()/(RAND_MAX+1.0));
}

/* LIBRARY QSORT */

int intcomp(const void *x, const void *y)
{	return *(const int *)x - *(const int *)y;
}

/* INSERTION SORTS */

/* Simplest insertion sort */
void isort1()
{	int i, j;
	for (i = 1; i < n; i++)
		for (j = i; j > 0 && x[j-1] > x[j]; j--)
			swap(j-1, j);
}

/* Write swap function inline */
void isort2()
{	int i, j;
	DType t;
	for (i = 1; i < n; i++)
		for (j = i; j > 0 && x[j-1] > x[j]; j--) {
			t = x[j];
			x[j] = x[j-1];
			x[j-1] = t;
		}
}

/* Move assignments to and from t out of loop */
void isort3()
{	int i, j;
	DType t;
	for (i = 1; i < n; i++) {
		t = x[i];
		for (j = i; j > 0 && x[j-1] > t; j--)
			x[j] = x[j-1];
		x[j] = t;
	}
}

/* QUICKSORTS */

/* Simplest version, Lomuto partitioning */
void qsort1(int l, int u)
{	int i, m;
	if (l >= u)
		return;
	m = l;
	for (i = l+1; i <= u; i++)
		if (x[i] < x[l])
			{ ++m; swap(m, i); }
	swap(l, m);
	qsort1(l, m-1);
	qsort1(m+1, u);
}

/* Sedgewick's version of Lomuto, with sentinel */
void qsort2(int l, int u)
{	int i, m;
	if (l >= u)
		return;
	m = i = u+1;
	do {
		do i--; while (x[i] < x[l]);
		--m; swap(m, i);
	} while (i > l);
	qsort2(l, m-1);
	qsort2(m+1, u);
}

/* Two-way partitioning */
void qsort3(int l, int u)
{	int i, j;
	DType t;
	if (l >= u)
		return;
	t = x[l];
	i = l;
	j = u+1;
	for (;;) {
		do i++; while (i <= u && x[i] < t);
		do j--; while (x[j] > t);
		if (i > j)
			break;
		swap(i, j);
	}
	swap(l, j);
	qsort3(l, j-1);
	qsort3(j+1, u);
}

/* qsort3 + randomization + isort small subarrays + swap inline */
int cutoff = 50;
void qsort4(int l, int u)
{	int i, j;
	DType t, temp;
	if (u - l < cutoff)
		return;
	swap(l, randint(l, u));
	t = x[l];
	i = l;
	j = u+1;
	for (;;) {
		do i++; while (i <= u && x[i] < t);
		do j--; while (x[j] > t);
		if (i > j)
			break;
		temp = x[i]; x[i] = x[j]; x[j] = temp;
	}
	swap(l, j);
	qsort4(l, j-1);
	qsort4(j+1, u);
}

/* selection */

void select1(int l, int u, int k)
{	int i, j;
	DType t, temp;
	if (l >= u)
		return;
	swap(l, randint(l, u));
	t = x[l];
	i = l;
	j = u+1;
	for (;;) {
		do i++; while (i <= u && x[i] < t);
		do j--; while (x[j] > t);
		if (i > j)
			break;
		temp = x[i]; x[i] = x[j]; x[j] = temp;
	}
	swap(l, j);
	if (j < k)
		select1(j+1, u, k);
	else if (j > k)
		select1(l, j-1, k);
}

/* HEAP SORTS */

void siftup(int u)
{	int i, p;
	i = u;
	for (;;) {
		if (i == 1)
			break;
		p = i / 2;
		if (x[p] >= x[i])
			break;
		swap(p, i);
		i = p;
	}
}

void siftdown1(int l, int u)
{	int i, c;
	i = l;
	for (;;) {
		c = 2*i;
		if (c > u)
			break;
		if (c+1 <= u && x[c+1] > x[c])
			c++;
		if (x[i] > x[c])
			break;
		swap(i, c);
		i = c;
	}
}

void siftdown1b(int l, int u) /* More C-ish version of 1 */
{	int i, c;
	for (i = l; (c = 2*i) <= u; i = c) {
		if (c+1 <= u && x[c+1] > x[c])
			c++;
		if (x[i] > x[c])
			break;
		swap(i, c);
	}
}

void hsort1()
{	int i;
	x--;
	for (i = 2; i <= n; i++)
		siftup(i);
	for (i = n; i >= 2; i--) {
		swap(1, i);
		siftdown1(1, i-1);
	}
	x++;
}

void hsort2()
{	int i;
	x--;
	for (i = n/2; i >= 1; i--)
		siftdown1(i, n);
	for (i = n; i >= 2; i--) {
		swap(1, i);
		siftdown1(1, i-1);
	}
	x++;
}

void siftdown3(int l, int u) /* push to bottom, then back up */
{	int i, c, p;
	i = l;
	for (;;) {
		c = 2*i;
		if (c > u)
			break;
		if (c+1 <= u && x[c+1] > x[c])
			c++;
		swap(i, c);
		i = c;
	}
	for (;;) {
		p = i/2;
		if (p < l)
			break;
		if (x[p] > x[i])
			break;
		swap(p, i);
		i = p;
	}
}

void hsort3()
{	int i;
	x--;
	for (i = n/2; i >= 1; i--)
		siftdown3(i, n);
	for (i = n; i >= 2; i--) {
		swap(1, i);
		siftdown3(1, i-1);
	}
	x++;
}

void siftdown4(int l, int u) /* replace swap with assignments */
{	int i, c, p;
	DType t;
	t = x[l];
	i = l;
	for (;;) {
		c = 2*i;
		if (c > u)
			break;
		if (c+1 <= u && x[c+1] > x[c])
			c++;
		x[i] = x[c];
		i = c;
	}
	x[i] = t;
	for (;;) {
		p = i/2;
		if (p < l)
			break;
		if (x[p] > x[i])
			break;
		swap(p, i);
		i = p;
	}
}

void hsort4()
{	int i;
	x--;
	for (i = n/2; i >= 1; i--)
		siftdown4(i, n);
	for (i = n; i >= 2; i--) {
		swap(1, i);
		siftdown4(1, i-1);
	}
	x++;
}


/* Other Sorts -- Exercises in Column 11 */

void selsort()  /* Selection sort */
{	int i, j;
	for (i = 0; i < n-1; i++)
		for (j = i; j < n; j++)
			if (x[j] < x[i])
				swap(i, j);
}

void shellsort()
{	int i, j, h;
	for (h = 1; h < n; h = 3*h + 1)
		;
	for (;;) {
		h /= 3;
		if (h < 1) break;
		for (i = h; i < n; i++) {
			for (j = i; j >= h; j -= h) {
				if (x[j-h] < x[j]) break;
				swap(j-h, j);
			}
		}
	}
}



/* SCAFFOLDING */

/* Timing */

double utime()
{
  struct timeval tv;

  gettimeofday(&tv, NULL);
  return tv.tv_sec + 1E-6*tv.tv_usec;
}

#define echo_do_this(op) { printf("%-48s\t", #op); start = utime(); op; sec = utime() - start; }
void timedriver()
{	int i, k, algnum, mod;
	double start = 0, sec = 0;
	size_t ta, tm;
	int algnums[] = { 0,1,2,3,4,5,6,7, 10,11,12,
#if defined(__BLOCKS__)
			  13,
#endif
			  14,15,16,
			  100,101,102,103,104,105,106,107,108,109,110, 120,121,122, /*21,22,23,*/ /*31,32,*/33,34, /*41,*/ 51,52,53,54, /*61,62,*/ };
	n = /*1000000*/100000;
	int mods[] = { n, 10 };
	for (tm=0; tm<sizeof(mods)/sizeof(mods[0]); tm++) { mod = mods[tm]; printf("%-48s\t%s\t%s\t%s\t%s\t%s\n", "operation", "algnum", "n", "mod", "second", "nsec/elem"); for (ta=0; ta<sizeof(algnums)/sizeof(algnums[0]); ta++) {
		algnum = algnums[ta];
		if (mod <= 0)
			mod = 10*n;
		for (i = 0; i < n; i++)
			x[i] = randint(0, mod-1);
		k = n/2;
		switch (algnum) {
		case  0: echo_do_this(shellsort_r(x, n, sizeof(int), NULL, i_cmp_r)); break;
		case  1: echo_do_this(heapsort_r(x, n, sizeof(int), NULL, i_cmp_r)); break;
		case  2: echo_do_this(mergesort_r(x, n, sizeof(int), NULL, i_cmp_r)); break;
		case  3: echo_do_this(qsort_r(x, n, sizeof(int), NULL, i_cmp_r)); break;
		case  4: echo_do_this(introsort_r(x, n, sizeof(int), NULL, i_cmp_r)); break;
		case  5: echo_do_this(pshellsort_r(x, n, sizeof(int), NULL, i_cmp_r)); break;
		case  6: echo_do_this(pmergesort_r(x, n, sizeof(int), NULL, i_cmp_r)); break;
		case  7: echo_do_this(pintrosort_r(x, n, sizeof(int), NULL, i_cmp_r)); break;

		case 10: echo_do_this(mergesort(x, n, sizeof(int), intcomp)); break;
		case 11: echo_do_this(heapsort(x, n, sizeof(int), intcomp)); break;
		case 12: echo_do_this(qsort(x, n, sizeof(int), intcomp)); break;
#if defined(__BLOCKS__)
		case 13: echo_do_this(psort(x, n, sizeof(int), intcomp)); break;
#endif
		case 14: echo_do_this(shell_sort(x, x+n)); break;
		case 15: echo_do_this(stable_sort(x, x+n)); break;
		case 16: echo_do_this(sort(x, x+n)); break;

		case 100: echo_do_this(quick_sort(x, x+n)); break;
		case 101: echo_do_this(pquick_sort(x, x+n)); break;
		case 102: echo_do_this(pquick_sort2(x, x+n)); break;
		case 103: echo_do_this(heap_sort1(x, x+n)); break;
		case 104: echo_do_this(heap_sort2(x, x+n)); break;
		case 105: echo_do_this(heap_sort3(x, x+n)); break;
		case 106: echo_do_this(merge_sort(x, x+n)); break;
		case 107: echo_do_this(pmerge_sort(x, x+n)); break;
		case 108: echo_do_this(pmerge_sort2(x, x+n)); break;
		case 109: echo_do_this(mmerge_sort(x, x+n)); break;
		case 110: echo_do_this(intro_sort(x, x+n)); break;

		case 120: echo_do_this(parallel_shell_sort(x, x+n)); break;
		case 121: echo_do_this(parallel_stable_sort(x, x+n)); break;
		case 122: echo_do_this(parallel_sort(x, x+n)); break;

		case 21: echo_do_this(isort1()); break;
		case 22: echo_do_this(isort2()); break;
		case 23: echo_do_this(isort3()); break;
		case 31: echo_do_this(qsort1(0, n-1)); break;
		case 32: echo_do_this(qsort2(0, n-1)); break;
		case 33: echo_do_this(qsort3(0, n-1)); break;
		case 34: echo_do_this(qsort4(0, n-1); isort3()); break;
		case 41: echo_do_this(select1(0, n-1, k)); break;
		case 51: echo_do_this(hsort1()); break;
		case 52: echo_do_this(hsort2()); break;
		case 53: echo_do_this(hsort3()); break;
		case 54: echo_do_this(hsort4()); break;
		case 61: echo_do_this(selsort()); break;
		case 62: echo_do_this(shellsort()); break;
		}
		if (algnum == 41) { /* Test selection */
			for (i = 0; i < k; i++)
				if (x[i] > x[k]) {
					printf("%d\tSELECT BUG i=%d\n", algnum, i);
					break;
				}
			for (i = k+1; i < n; i++)
				if (x[i] < x[k]) {
					printf("%d\tSELECT BUG i=%d\n", algnum, i);
					break;
				}
		} else { /* Test sort */
			for (i = 0; i < n-1; i++)
				if (x[i] > x[i+1]) {
					printf("%d\tSORT BUG i=%d\n", algnum, i);
					break;
				}
		}
		printf("%d\t%d\t%d\t%g\t%g\n",
			algnum, n, mod, sec, 1e9*sec/n);
	}}
}

/* Main */

int main()
{
	timedriver();
	return 0;
}
