/*
  sort.hpp - yet another sort algorithm

  Copyright (C) 2010 Taiji Yamada <taiji@aihara.co.jp>

  Distributed under the Boost Software License, Version 1.0. See
  http://www.boost.org/LICENSE_1_0.txt for more details.

  References:
  [1] http://www.moon.sannet.ne.jp/okahisa/sort/sort.html, 2010.
  [2] http://threadpool.sourceforge.net/, 2005-2008.
 */
#ifndef _sort_hpp_
#define _sort_hpp_

#include <iterator>
#include <functional>
#include <algorithm>
#include <vector>
#include <boost/swap.hpp>
#include <boost/thread.hpp>
#define USE_BOOST_THREADPOOL
#if !defined(USE_BOOST_THREADPOOL)
#include <queue>
#include <boost/function.hpp>
#include <boost/bind.hpp>
#include <boost/thread/condition.hpp>

class threadpool {
  typedef boost::function0<void> job_type;
  std::queue<job_type> jobs;
  boost::thread_group thrs;
  int c_idle, c_job;
  bool flag_wait, flag_quit;
  boost::mutex mutex;
  boost::condition cond_work, cond_wait;
  void work()
  {
    do {
      job_type job = 0;
      {
	boost::mutex::scoped_lock lock(mutex);
	++c_idle;
	while (jobs.empty() && !flag_quit)
	  cond_work.wait(lock);
	--c_idle;
	if (!jobs.empty()) {
	  job = jobs.front();
	  jobs.pop();
	}
      }
      if (job)
	job();
      {
	boost::mutex::scoped_lock lock(mutex);
	if (job)
	  --c_job;
	if (flag_wait && jobs.empty()) {
	  flag_wait = false;
	  cond_wait.notify_all();
	}
      }
    } while (!flag_quit);
  }
public:
  threadpool(size_t size) : c_idle(0), c_job(0), flag_wait(false), flag_quit(false)
  {
    for (size_t i=0; i<size; i++)
      thrs.create_thread(boost::bind(&threadpool::work, this));
  }
  void push(job_type job)
  {
    boost::mutex::scoped_lock lock(mutex);
    jobs.push(job);
    ++c_job;
    if (c_idle > 0)
      cond_work.notify_one();
  }
  void wait()
  {
    boost::mutex::scoped_lock lock(mutex);
    while (c_job) {
      flag_wait = true;
      cond_wait.wait(lock);
    }
  }
  void join()
  {
    {
      boost::mutex::scoped_lock lock(mutex);
      while (c_job) {
	flag_wait = true;
	cond_wait.wait(lock);
      }
      flag_quit = true;
      cond_work.notify_all();
    }
    thrs.join_all();
  }
};
#else
#include <boost/threadpool.hpp>
#endif

template <typename T>
inline T fls0l(T x)
{
  T n = 0;
  for (; x>1; x>>=1, n++);
  return n;
}
template <typename S>
inline void intro_sort_limiter(S n, S &depth)
{
#if !defined(__BLOCKS__)
  depth = fls0l(n)*2;
#else
  depth = (flsl(n)-1)*2;
#endif
}

template <typename T>
inline T flsqrt(T x, T fls)
{
  T s = 1;
  s <<= fls/2;
  return (s + x/s)/2;
}
template <typename S>
inline void parallel_intro_sort_limiter(S n, S &depth, S &n_threshold)
{
#if !defined(__BLOCKS__)
  S fls0 = fls0l(n);
  depth = fls0*2;
  n_threshold = flsqrt(n, fls0+1);
#else
  S fls = flsl(n);
  depth = (fls-1)*2;
  n_threshold = flsqrt(n, fls);
#endif
}

template <class FI, class C>
void foam_sort(FI a0, FI aN, C cmp)
{
  if (a0 == aN) return;
  FI i, j, k = aN, m, a1 = a0;

  std::advance(a1, 1);
  do {
    m = a0;
    for (i=j=a0, ++j; j!=k; ++i, ++j) {
      if (cmp(*j, *i)) {
	boost::swap(*i, *j);
	m = j;
      }
    }
    k = m;
  } while (k > a1);
}
template <class FI>
inline void foam_sort(FI a0, FI aN)
{
  typedef typename std::iterator_traits<FI>::value_type val_t;
  foam_sort(a0, aN, std::less<val_t>());
}

/* Sequential oddeven sort */
template <class FI, class C>
void oddeven_sort(FI a0, FI aN, C cmp)
{
  if (a0 == aN) return;
  FI i, j, k = aN;
  bool flag = false;

  while (!flag) {
    flag = true;
    for (i=j=a0, ++j; j<k; i+=2, j+=2)
      if (cmp(*j, *i)) {
	boost::swap(*i, *j);
	flag = false;
      }
    for (i=j=a0, ++i, j+=2; j<k; i+=2, j+=2)
      if (cmp(*j, *i)) {
	boost::swap(*i, *j);
	flag = false;
      }
  }
}
template <class FI>
inline void oddeven_sort(FI a0, FI aN)
{
  typedef typename std::iterator_traits<FI>::value_type val_t;
  oddeven_sort(a0, aN, std::less<val_t>());
}
/* Parallel oddeven sort */
template <class FI, class C>
struct thr_parallel_oddeven_sort {
  FI n0, n1;
  C cmp;
  bool *flag;
  thr_parallel_oddeven_sort(FI _n0, FI _n1, C _cmp, bool *_flag) :
    n0(_n0), n1(_n1), cmp(_cmp), flag(_flag) {}
  void operator()()
  {
    FI i, j;

    for (i=j=n0, ++j; i<n1; i+=2, j+=2)
      if (cmp(*j, *i)) {
	boost::swap(*i, *j);
	*flag = false;
      }
    for (i=j=n0, ++i, j+=2; i<n1; i+=2, j+=2)
      if (cmp(*j, *i)) {
	boost::swap(*i, *j);
	*flag = false;
      }
  }
};
#if 0
/* Thread version oddeven sort */
template <class FI, class C>
inline void parallel_oddeven_sort(FI a0, FI aN, C cmp)
{
  typedef typename std::iterator_traits<FI>::difference_type dif_t;
  const dif_t N = std::distance(a0, aN);
  if (N < 2) return;
  unsigned int n_cpu = boost::thread::hardware_concurrency();
  if (n_cpu > 1) {
    FI i;
    bool flag = false;

    while (!flag) {
      size_t k, m = min(N-1, dif_t(n_cpu));
      boost::thread_group thrs;
      flag = true;
      for (i=a0, k=0; k<m; k++) {
	thr_parallel_oddeven_sort<FI, C>
	  functor(i, i + (N-1)/m + ((k < (N-1)%m)?1:0), cmp, &flag);
	thrs.create_thread(functor);
	i = functor.n1;
      }
      thrs.join_all();
    }
  }
  else
    oddeven_sort(a0, aN, cmp);
}
#elif !defined(USE_BOOST_THREADPOOL)
/* Thread pool version oddeven sort */
template <class FI, class C>
inline void parallel_oddeven_sort(FI a0, FI aN, C cmp)
{
  typedef typename std::iterator_traits<FI>::difference_type dif_t;
  const dif_t N = std::distance(a0, aN);
  if (N < 2) return;
  unsigned int n_cpu = boost::thread::hardware_concurrency();
  if (n_cpu > 1) {
    threadpool pool(n_cpu);
    FI i;
    bool flag = false;

    while (!flag) {
      size_t k, m = min(N-1, dif_t(n_cpu));
      flag = true;
      for (i=a0, k=0; k<m; k++) {
	thr_parallel_oddeven_sort<FI, C>
	  functor(i, i + (N-1)/m + ((k < (N-1)%m)?1:0), cmp, &flag);
	pool.push(functor);
	i = functor.n1;
      }
      pool.wait();
    }
    pool.join();
  }
  else
    oddeven_sort(a0, aN, cmp);
}
#else
/* Thread pool version oddeven sort */
template <class FI, class C>
inline void parallel_oddeven_sort(FI a0, FI aN, C cmp)
{
  typedef typename std::iterator_traits<FI>::difference_type dif_t;
  const dif_t N = std::distance(a0, aN);
  if (N < 2) return;
  unsigned int n_cpu = boost::thread::hardware_concurrency();
  if (n_cpu > 1) {
    boost::threadpool::pool pool(n_cpu);
    FI i;
    bool flag = false;

    while (!flag) {
      size_t k, m = min(N-1, dif_t(n_cpu));
      flag = true;
      for (i=a0, k=0; k<m; k++) {
	thr_parallel_oddeven_sort<FI, C>
	  functor(i, i + (N-1)/m + ((k < (N-1)%m)?1:0), cmp, &flag);
	pool.schedule(functor);
	i = functor.n1;
      }
      pool.wait();
    }
  }
  else
    oddeven_sort(a0, aN, cmp);
}
#endif
template <class FI>
inline void parallel_oddeven_sort(FI a0, FI aN)
{
  typedef typename std::iterator_traits<FI>::value_type val_t;
  parallel_oddeven_sort(a0, aN, std::less<val_t>());
}

template <class BI, class C>
#if 0
void insert_sort(BI a0, BI aN, C cmp)
{
  if (a0 == aN) return;
  BI i = a0;
  for (++i; i!=aN; ++i) {
    BI j = i, k = i;
    while (a0 != j && cmp(*j, *(--k))) {
      boost::swap(*j, *k);
      j = k;
    }
  }
}
#elif !0
void insert_sort(BI a0, BI aN, C cmp)
{
  typedef typename std::iterator_traits<BI>::value_type val_t;
  if (a0 == aN) return;
  BI i = a0;
  for (++i; i!=aN; ++i) {
    const val_t tmp = *i;
    BI j = i, k = i;
    while (a0 != j && cmp(tmp, *(--k))) {
      *j = *k;
      j = k;
    }
    *j = tmp;
  }
}
#else /* not faster than above */
void insert_sort(BI a0, BI aN, C cmp)
{
  typedef typename std::iterator_traits<BI>::value_type val_t;
  if (a0 == aN) return;
  BI i = a0;
  for (++i; i!=aN; ++i) {
    const val_t tmp = *i;
    BI j = i, k = i;
    if (cmp(tmp, *a0))
      while (a0 != j) {
	--k;
	*j = *k;
	j = k;
      }
    else 
      while (cmp(tmp, *(--k))) {
	*j = *k;
	j = k;
      }
    *j = tmp;
  }
}
#endif
template <class BI>
inline void insert_sort(BI a0, BI aN)
{
  typedef typename std::iterator_traits<BI>::value_type val_t;
  insert_sort(a0, aN, std::less<val_t>());
}

/* Sequential shell sort */
template <class BI, class C>
void shell_sort(BI a0, BI aN, C cmp)
{
  typedef typename std::iterator_traits<BI>::difference_type dif_t;
  const dif_t N = std::distance(a0, aN);
  if (N < 2) return;
  dif_t H;
  for (H=1; H<N; H=H*3+1); H /= 3;
  {
    while (H != 1) {
      BI i, j, k, aH = a0 + H;
      for (k=a0; k!=aH; ++k)
	for (i=k; i<aN; i+=H)
	  for (j=i; j>aH && cmp(*j, *(j-H)); j-=H)
	    boost::swap(*(j-H), *j);
      H /= 3;
    }
  }
  insert_sort(a0, aN, cmp);
}
template <class BI>
inline void shell_sort(BI a0, BI aN)
{
  typedef typename std::iterator_traits<BI>::value_type val_t;
  shell_sort(a0, aN, std::less<val_t>());
}

template <class BI, class C>
struct thr_parallel_shell_sort {
  typedef typename std::iterator_traits<BI>::difference_type dif_t;
  BI a0, aN;
  C cmp;
  BI k0, k1;
  dif_t H;
  thr_parallel_shell_sort(BI _a0, BI _aN, C _cmp, BI _k0, BI _k1, dif_t _H) :
    a0(_a0), aN(_aN), cmp(_cmp), k0(_k0), k1(_k1), H(_H) {}
  void operator()()
  {
    BI i, j, k, aH = a0 + H;
    for (k=k0; k!=k1; ++k)
      for (i=k; i<aN; i+=H)
	for (j=i; j>aH && cmp(*j, *(j-H)); j-=H)
	  boost::swap(*(j-H), *j);
  }
};
/* Parallel shell sort */
#if 0
/* Thread version shell sort */
template <class BI, class C>
void parallel_shell_sort(BI a0, BI aN, C cmp)
{
  typedef typename std::iterator_traits<BI>::difference_type dif_t;
  const dif_t N = std::distance(a0, aN);
  if (N < 2) return;
  unsigned int n_cpu = boost::thread::hardware_concurrency();
  if (n_cpu > 1) {
    dif_t H;
    for (H=1; H<N; H=H*3+1); H /= 3;
    BI i;
    {
      while (H != 1) {
	size_t k, m = min(H, dif_t(n_cpu));
	boost::thread_group thrs;
	for (i=a0, k=0; k<m; k++) {
	  thr_parallel_shell_sort<BI, C>
	    functor(a0, aN, cmp, i, i + H/m + ((k < H%m)?1:0), H);
	  thrs.create_thread(functor);
	  i = functor.k1;
	}
	thrs.join_all();
	H /= 3;
      }
    }
    insert_sort(a0, aN, cmp);
  }
  else
    shell_sort(a0, aN, cmp);
}
#elif !defined(USE_BOOST_THREADPOOL)
/* Thread pool version shell sort */
template <class BI, class C>
void parallel_shell_sort(BI a0, BI aN, C cmp)
{
  typedef typename std::iterator_traits<BI>::difference_type dif_t;
  const dif_t N = std::distance(a0, aN);
  if (N < 2) return;
  unsigned int n_cpu = boost::thread::hardware_concurrency();
  if (n_cpu > 1) {
    threadpool pool(n_cpu);
    dif_t H;
    for (H=1; H<N; H=H*3+1); H /= 3;
    BI i;
    {
      while (H != 1) {
	size_t k, m = min(H, dif_t(n_cpu));
	boost::thread_group thrs;
	for (i=a0, k=0; k<m; k++) {
	  thr_parallel_shell_sort<BI, C>
	    functor(a0, aN, cmp, i, i + H/m + ((k < H%m)?1:0), H);
	  pool.push(functor);
	  i = functor.k1;
	}
	pool.wait();
	H /= 3;
      }
    }
    pool.join();
    insert_sort(a0, aN, cmp);
  }
  else
    shell_sort(a0, aN, cmp);
}
#else
/* Thread pool version shell sort */
template <class BI, class C>
void parallel_shell_sort(BI a0, BI aN, C cmp)
{
  typedef typename std::iterator_traits<BI>::difference_type dif_t;
  const dif_t N = std::distance(a0, aN);
  if (N < 2) return;
  unsigned int n_cpu = boost::thread::hardware_concurrency();
  if (n_cpu > 1) {
    boost::threadpool::pool pool(n_cpu);
    dif_t H;
    for (H=1; H<N; H=H*3+1); H /= 3;
    BI i;
    {
      while (H != 1) {
	size_t k, m = min(H, dif_t(n_cpu));
	boost::thread_group thrs;
	for (i=a0, k=0; k<m; k++) {
	  thr_parallel_shell_sort<BI, C>
	    functor(a0, aN, cmp, i, i + H/m + ((k < H%m)?1:0), H);
	  pool.schedule(functor);
	  i = functor.k1;
	}
	pool.wait();
	H /= 3;
      }
    }
    insert_sort(a0, aN, cmp);
  }
  else
    shell_sort(a0, aN, cmp);
}
#endif
template <class BI>
inline void parallel_shell_sort(BI a0, BI aN)
{
  typedef typename std::iterator_traits<BI>::value_type val_t;
  parallel_shell_sort(a0, aN, std::less<val_t>());
}

/* Sequential merge sort */
template <class BI, class C, class B, class D>
void sequential_merge_sort(BI a0, BI aN, C cmp, B b0, D N)
{
  if (N < 34) { insert_sort(a0, aN, cmp); return; }
  D M = N >> 1;
  BI aM = a0 + M;
  B bM = b0 + M;
  B bN = b0 + N;
  sequential_merge_sort(a0, aM, cmp, b0, M);
  sequential_merge_sort(aM, aN, cmp, bM, N - M);
  BI ak = a0;
  B bi = b0, bj = bM;
  while (bi != bM && bj != bN) {
    if (cmp(*bj, *bi)) {
      *ak = *bj;
      ++ak;
      ++bj;
    }
    else {
      *ak = *bi;
      ++ak;
      ++bi;
    }
  }
  ak = std::copy(bi, bM, ak);
  ak = std::copy(bj, bN, ak);
}
template <class BI, class C>
inline void sequential_stable_sort(BI a0, BI aN, C cmp)
{
  typedef typename std::iterator_traits<BI>::value_type val_t;
  typedef typename std::iterator_traits<BI>::difference_type dif_t;
  typedef typename std::vector<val_t>::iterator it;
  const dif_t N = std::distance(a0, aN);
  std::vector<val_t> b(N);
  sequential_merge_sort(a0, aN, cmp, b.begin(), N);
}
/* Parallel merge sort */
#if !0
/* Thread version merge sort */
template <class BI, class C, class B>
struct parallel_merge_sort {
  typedef typename std::iterator_traits<BI>::difference_type dif_t;
  BI a0, aN;
  C cmp;
  B b0;
  dif_t N, n_threshold;
  parallel_merge_sort(BI _a0, BI _aN, C _cmp, B _b0, dif_t _N, dif_t _n_threshold) :
    a0(_a0), aN(_aN), cmp(_cmp), b0(_b0), N(_N), n_threshold(_n_threshold) {}
  void operator()()
  {
    if (N < 34) { insert_sort(a0, aN, cmp); return; }
    dif_t M = N >> 1;
    BI aM = a0 + M;
    B bM = b0 + M;
    B bN = b0 + N;
    parallel_merge_sort
      functor1(a0, aM, cmp, b0, M, n_threshold),
      functor2(aM, aN, cmp, bM, N - M, n_threshold);

    if (M > n_threshold) {
      boost::thread thr(functor1);
      functor2();
      thr.join();
    }
    else {
      functor1();
      functor2();
    }
    BI ak = a0;
    B bi = b0, bj = bM;
    while (bi != bM && bj != bN) {
      if (cmp(*bj, *bi)) {
	*ak = *bj;
	++ak;
	++bj;
      }
      else {
	*ak = *bi;
	++ak;
	++bi;
      }
    }
    ak = std::copy(bi, bM, ak);
    ak = std::copy(bj, bN, ak);
  }
};
template <class BI, class C>
inline void parallel_stable_sort(BI a0, BI aN, C cmp)
{
  unsigned int n_cpu = boost::thread::hardware_concurrency();
  if (n_cpu > 1) {
    typedef typename std::iterator_traits<BI>::value_type val_t;
    typedef typename std::iterator_traits<BI>::difference_type dif_t;
    typedef typename std::vector<val_t>::iterator it;
    const dif_t N = std::distance(a0, aN), n_threshold = N/n_cpu;
    std::vector<val_t> b(N);
    parallel_merge_sort<BI, C, it>
      functor(a0, aN, cmp, b.begin(), N, n_threshold);
    functor();
  }
  else
    sequential_stable_sort(a0, aN, cmp);
}
#elif !defined(USE_BOOST_THREADPOOL) /* not so rapid */
/* Thread pool version merge sort */
template <class BI, class C, class B>
struct parallel_merge_sort {
  typedef typename std::iterator_traits<BI>::difference_type dif_t;
  BI a0, aN;
  C cmp;
  B b0;
  dif_t N, n_threshold;
  threadpool *pool;
  parallel_merge_sort(BI _a0, BI _aN, C _cmp, B _b0, dif_t _N, dif_t _n_threshold, threadpool *_pool) :
    a0(_a0), aN(_aN), cmp(_cmp), b0(_b0), N(_N), n_threshold(_n_threshold), pool(_pool) {}
  void operator()()
  {
    if (N < 34) { insert_sort(a0, aN, cmp); return; }
    dif_t M = N >> 1;
    BI aM = a0 + M;
    B bM = b0 + M;
    B bN = b0 + N;
    parallel_merge_sort
      functor1(a0, aM, cmp, b0, M, n_threshold, pool),
      functor2(aM, aN, cmp, bM, N - M, n_threshold, pool);

    if (M > n_threshold) {
      pool->push(functor1);
      functor2();
      pool->wait();
    }
    else {
      functor1();
      functor2();
    }
    BI ak = a0;
    B bi = b0, bj = bM;
    while (bi != bM && bj != bN) {
      if (cmp(*bj, *bi)) {
	*ak = *bj;
	++ak;
	++bj;
      }
      else {
	*ak = *bi;
	++ak;
	++bi;
      }
    }
    ak = std::copy(bi, bM, ak);
    ak = std::copy(bj, bN, ak);
  }
};
template <class BI, class C>
inline void parallel_stable_sort(BI a0, BI aN, C cmp)
{
  unsigned int n_cpu = boost::thread::hardware_concurrency();
  if (n_cpu > 1) {
    typedef typename std::iterator_traits<BI>::value_type val_t;
    typedef typename std::iterator_traits<BI>::difference_type dif_t;
    typedef typename std::vector<val_t>::iterator it;
    const dif_t N = std::distance(a0, aN), n_threshold = N/n_cpu;
    std::vector<val_t> b(N);
    threadpool pool(n_cpu);
    parallel_merge_sort<BI, C, it>
      functor(a0, aN, cmp, b.begin(), N, n_threshold, &pool);
    functor();
    pool.join();
  }
  else
    sequential_stable_sort(a0, aN, cmp);
}
#else /* not so rapid */
/* Thread pool version merge sort */
template <class BI, class C, class B>
struct parallel_merge_sort {
  typedef typename std::iterator_traits<BI>::difference_type dif_t;
  BI a0, aN;
  C cmp;
  B b0;
  dif_t N, n_threshold;
  boost::threadpool::pool *pool;
  parallel_merge_sort(BI _a0, BI _aN, C _cmp, B _b0, dif_t _N, dif_t _n_threshold, boost::threadpool::pool *_pool) :
    a0(_a0), aN(_aN), cmp(_cmp), b0(_b0), N(_N), n_threshold(_n_threshold), pool(_pool) {}
  void operator()()
  {
    if (N < 34) { insert_sort(a0, aN, cmp); return; }
    dif_t M = N >> 1;
    BI aM = a0 + M;
    B bM = b0 + M;
    B bN = b0 + N;
    parallel_merge_sort
      functor1(a0, aM, cmp, b0, M, n_threshold, pool),
      functor2(aM, aN, cmp, bM, N - M, n_threshold, pool);

    if (M > n_threshold) {
      pool->schedule(functor1);
      functor2();
      pool->wait();
    }
    else {
      functor1();
      functor2();
    }
    BI ak = a0;
    B bi = b0, bj = bM;
    while (bi != bM && bj != bN) {
      if (cmp(*bj, *bi)) {
	*ak = *bj;
	++ak;
	++bj;
      }
      else {
	*ak = *bi;
	++ak;
	++bi;
      }
    }
    ak = std::copy(bi, bM, ak);
    ak = std::copy(bj, bN, ak);
  }
};
template <class BI, class C>
inline void parallel_stable_sort(BI a0, BI aN, C cmp)
{
  unsigned int n_cpu = boost::thread::hardware_concurrency();
  if (n_cpu > 1) {
    typedef typename std::iterator_traits<BI>::value_type val_t;
    typedef typename std::iterator_traits<BI>::difference_type dif_t;
    typedef typename std::vector<val_t>::iterator it;
    const dif_t N = std::distance(a0, aN), n_threshold = N/n_cpu;
    std::vector<val_t> b(N);
    boost::threadpool::pool pool(n_cpu);
    parallel_merge_sort<BI, C, it>
      functor(a0, aN, cmp, b.begin(), N, n_threshold, &pool);
    functor();
  }
  else
    sequential_stable_sort(a0, aN, cmp);
}
#endif
template <class BI>
inline void parallel_stable_sort(BI a0, BI aN)
{
  typedef typename std::iterator_traits<BI>::value_type val_t;
  parallel_stable_sort(a0, aN, std::less<val_t>());
}

template <class RAI, class T, class C>
inline void heap_up(RAI a, T c, C cmp)
{
  typedef typename std::iterator_traits<RAI>::value_type val_t;
  const val_t tmp = a[c];
  while (c > 0) {
    T p = (c - 1) >> 1;
    if (!cmp(a[p], tmp)) break;
    a[c] = a[p];
    c = p;
  }
  a[c] = tmp;
}
template <class RAI, class T, class C>
inline void heap_pop(RAI a, T q, T r, C cmp)
{
  typedef typename std::iterator_traits<RAI>::value_type val_t;
  T p = 0, c;
  const val_t tmp = a[r];
  a[r] = a[0];
  while ((c = (p << 1) + 1) <= q) {
    if (c < q && cmp(a[c], a[c + 1])) c++;
    a[p] = a[c];
    p = c;
  }
  a[p] = tmp;
  heap_up(a, p, cmp);
}
template <class RAI, class C>
void uppop_heap_sort(RAI a0, RAI aN, C cmp)
{
  typedef typename std::iterator_traits<RAI>::difference_type dif_t;
  const dif_t N = aN - a0;
  if (N < 2) return;
  for (dif_t c=1; c<N; c++) heap_up(a0, c, cmp);
  for (dif_t q=N-1; q>0; q--) heap_pop(a0, q - 1, q, cmp);
}
template <class RAI>
inline void uppop_heap_sort(RAI a0, RAI aN)
{
  typedef typename std::iterator_traits<RAI>::value_type val_t;
  uppop_heap_sort(a0, aN, std::less<val_t>());
}

template <class T, class C>
inline T median3(const T &a, const T &b, const T &c, C cmp)
{
  return cmp(a, b) ?
    (cmp(b, c) ? b : cmp(c, a) ? a : c) :
    (cmp(c, b) ? b : cmp(a, c) ? a : c);
}

/* Sequential intro sort */
template <class RAI, class C, class D>
static void sequential_intro_sort(RAI a0, RAI aN, C cmp, D d)
{
  typedef typename std::iterator_traits<RAI>::difference_type dif_t;
  typedef typename std::iterator_traits<RAI>::value_type val_t;
  const dif_t N = aN - a0;
  const dif_t L = N - 1;
  if (N < 40) { insert_sort(a0, aN, cmp); return; }
  if (d == 0) { uppop_heap_sort(a0, aN, cmp); return; }
  const RAI& a = a0;
  const dif_t M = N >> 1;
  val_t l, p = a[M], u;
  if (N > 7) {
    l = a[0];
    u = a[L];
    if (N > 40) {
      const dif_t s = M >> 1, d = s >> 1;
      l = median3(l, a[d], a[s], cmp);
      p = median3(a[M-d], p, a[M+d], cmp);
      u = median3(a[N-s], a[N-d], u, cmp);
    }
    p = median3(l, p, u, cmp);
  }
  RAI i = a0, j = a0 + L;
  while (!0) {
    while (cmp(*i, p)) ++i;
    while (cmp(p, *j)) --j;
    if (i >= j) break;
    boost::swap(*i, *j);
    ++i;
    --j;
  }
  ++j;
  const dif_t d2 = d - 1;
  if (i - a0 > 1) sequential_intro_sort(a0, i, cmp, d2);
  if (aN - j > 1) sequential_intro_sort(j, aN, cmp, d2);
}
template <class RAI, class C>
inline void sequential_sort(RAI a0, RAI aN, C cmp)
{
  typedef typename std::iterator_traits<RAI>::difference_type dif_t;
  dif_t N = aN - a0, depth;
  intro_sort_limiter(N, depth);
  sequential_intro_sort(a0, aN, cmp, depth);
}
template <class RAI>
inline void sequential_sort(RAI a0, RAI aN)
{
  typedef typename std::iterator_traits<RAI>::value_type val_t;
  sequential_sort(a0, aN, std::less<val_t>());
}
/* Parallel intro sort */
#if 0 /* very dull */
/* Thread version intro sort */
template <class RAI, class C>
struct parallel_intro_sort {
  typedef typename std::iterator_traits<RAI>::difference_type dif_t;
  typedef typename std::iterator_traits<RAI>::value_type val_t;
  RAI a0, aN;
  C cmp;
  dif_t d;
  size_t *n_thread, m_thread;
  boost::thread **thrs;
  boost::mutex *mutex;
  parallel_intro_sort(RAI _a0, RAI _aN, C _cmp, dif_t _d, size_t *_n_thread, size_t _m_thread, boost::thread **_thrs, boost::mutex *_mutex) :
    a0(_a0), aN(_aN), cmp(_cmp), d(_d), n_thread(_n_thread), m_thread(_m_thread), thrs(_thrs), mutex(_mutex) {}
  void operator()()
  {
    const dif_t N = aN - a0;
    const dif_t L = N - 1;
    if (N < 40) { insert_sort(a0, aN, cmp); return; }
    if (d == 0) { uppop_heap_sort(a0, aN, cmp); return; }
    const RAI& a = a0;
    const dif_t M = N >> 1;
    val_t l, p = a[M], u;
    if (N > 7) {
      l = a[0];
      u = a[L];
      if (N > 40) {
	const dif_t s = M >> 1, d = s >> 1;
	l = median3(l, a[d], a[s], cmp);
	p = median3(a[M-d], p, a[M+d], cmp);
	u = median3(a[N-s], a[N-d], u, cmp);
      }
      p = median3(l, p, u, cmp);
    }
    RAI i = a0, j = a0 + L;
    while (!0) {
      while (cmp(*i, p)) ++i;
      while (cmp(p, *j)) --j;
      if (i >= j) break;
      boost::swap(*i, *j);
      ++i;
      --j;
    }
    ++j;
    const dif_t d2 = d - 1;
    parallel_intro_sort<RAI, C>
      functor1(a0, i, cmp, d2, n_thread, m_thread, thrs, mutex),
      functor2(j, aN, cmp, d2, n_thread, m_thread, thrs, mutex);
    if (i - a0 > 1) {
      bool flag;
      {
	boost::mutex::scoped_lock lock(*mutex);
	if ((flag = *n_thread < m_thread)) {
	  thrs[*n_thread] = new boost::thread(functor1);
	  ++(*n_thread);
	}
      }
      if (!flag) functor1();
    }
    if (aN - j > 1) {
      functor2();
    }
  }
};
template <class RAI, class C>
inline void parallel_sort(RAI a0, RAI aN, C cmp)
{
  typedef typename std::iterator_traits<RAI>::difference_type dif_t;
  dif_t N = aN - a0;
  size_t depth;
  intro_sort_limiter(N, depth);
  size_t i, n_thread = 0, m_thread = 16*boost::thread::hardware_concurrency();
  boost::thread *thrs[m_thread];
  boost::mutex mutex;
  parallel_intro_sort<RAI, C> functor(a0, aN, cmp, depth, &n_thread, m_thread, thrs, &mutex);
  functor();
  for (i=0; i<n_thread; i++)
    thrs[i]->join();
}
#elif 0 /* does not work since recursive create_thread is not supported */
/* Thread group version intro sort */
template <class RAI, class C>
struct parallel_intro_sort {
  typedef typename std::iterator_traits<RAI>::difference_type dif_t;
  typedef typename std::iterator_traits<RAI>::value_type val_t;
  RAI a0, aN;
  C cmp;
  dif_t d;
  size_t m_thread;
  boost::thread_group *thrs;
  parallel_intro_sort(RAI _a0, RAI _aN, C _cmp, dif_t _d, size_t _m_thread, boost::thread_group *_thrs) :
    a0(_a0), aN(_aN), cmp(_cmp), d(_d), m_thread(_m_thread), thrs(_thrs) {}
  void operator()()
  {
    const dif_t N = aN - a0;
    const dif_t L = N - 1;
    if (N < 40) { insert_sort(a0, aN, cmp); return; }
    if (d == 0) { uppop_heap_sort(a0, aN, cmp); return; }
    const RAI& a = a0;
    const dif_t M = N >> 1;
    val_t l, p = a[M], u;
    if (N > 7) {
      l = a[0];
      u = a[L];
      if (N > 40) {
	const dif_t s = M >> 1, d = s >> 1;
	l = median3(l, a[d], a[s], cmp);
	p = median3(a[M-d], p, a[M+d], cmp);
	u = median3(a[N-s], a[N-d], u, cmp);
      }
      p = median3(l, p, u, cmp);
    }
    RAI i = a0, j = a0 + L;
    while (!0) {
      while (cmp(*i, p)) ++i;
      while (cmp(p, *j)) --j;
      if (i >= j) break;
      boost::swap(*i, *j);
      ++i;
      --j;
    }
    ++j;
    const dif_t d2 = d - 1;
    parallel_intro_sort<RAI, C>
      functor1(a0, i, cmp, d2, m_thread, thrs),
      functor2(j, aN, cmp, d2, m_thread, thrs);
    if (i - a0 > 1) {
      if (thrs->size() < m_thread)
	thrs->create_thread(functor1);
      else
	functor1();
    }
    if (aN - j > 1) {
      functor2();
    }
  }
};
template <class RAI, class C>
inline void parallel_sort(RAI a0, RAI aN, C cmp)
{
  typedef typename std::iterator_traits<RAI>::difference_type dif_t;
  dif_t N = aN - a0;
  size_t depth;
  intro_sort_limiter(N, depth);
  size_t m_thread = 16*boost::thread::hardware_concurrency();
  boost::thread_group thrs;
  parallel_intro_sort<RAI, C> functor(a0, aN, cmp, depth, m_thread, &thrs);
  functor();
  thrs.join_all();
}
#elif !defined(USE_BOOST_THREADPOOL)
/* Thread pool version intro sort */
template <class RAI, class C>
struct parallel_intro_sort {
  typedef typename std::iterator_traits<RAI>::difference_type dif_t;
  typedef typename std::iterator_traits<RAI>::value_type val_t;
  RAI a0, aN;
  C cmp;
  dif_t d, *n_threshold;
  threadpool *pool;
  parallel_intro_sort(RAI _a0, RAI _aN, C _cmp, dif_t _d, dif_t *_n_threshold, threadpool *_pool) :
    a0(_a0), aN(_aN), cmp(_cmp), d(_d), n_threshold(_n_threshold), pool(_pool) {}
  void operator()()
  {
    const dif_t N = aN - a0;
    const dif_t L = N - 1;
    if (N < 40) { insert_sort(a0, aN, cmp); return; }
    if (d == 0) { uppop_heap_sort(a0, aN, cmp); return; }
    const RAI& a = a0;
    const dif_t M = N >> 1;
    val_t l, p = a[M], u;
    if (N > 7) {
      l = a[0];
      u = a[L];
      if (N > 40) {
	const dif_t s = M >> 1, d = s >> 1;
	l = median3(l, a[d], a[s], cmp);
	p = median3(a[M-d], p, a[M+d], cmp);
	u = median3(a[N-s], a[N-d], u, cmp);
      }
      p = median3(l, p, u, cmp);
    }
    RAI i = a0, j = a0 + L;
    while (!0) {
      while (cmp(*i, p)) ++i;
      while (cmp(p, *j)) --j;
      if (i >= j) break;
      boost::swap(*i, *j);
      ++i;
      --j;
    }
    ++j;
    const dif_t d2 = d - 1;
    dif_t h;
    parallel_intro_sort<RAI, C>
      functor1(a0, i, cmp, d2, n_threshold, pool),
      functor2(j, aN, cmp, d2, n_threshold, pool);
    if ((h = i - a0) > 1) {
      if (h > *n_threshold) pool->push(functor1);
      else functor1();
    }
    if (aN - j > 1) {
      functor2();
    }
  }
};
template <class RAI, class C>
inline void parallel_sort(RAI a0, RAI aN, C cmp)
{
  unsigned int n_cpu = boost::thread::hardware_concurrency();
  if (n_cpu > 1) {
    typedef typename std::iterator_traits<RAI>::difference_type dif_t;
    dif_t N = aN - a0;
    size_t depth, n_threshold;
    parallel_intro_sort_limiter(N, depth, n_threshold);
    threadpool pool(n_cpu/*-1*/);
    parallel_intro_sort<RAI, C> functor(a0, aN, cmp, depth, &n_threshold, &pool);
    pool.push(functor);
    pool.join();
  }
  else
    sequential_sort(a0, aN, cmp);
}
#else
/* Thread pool version intro sort */
template <class RAI, class C>
struct parallel_intro_sort {
  typedef typename std::iterator_traits<RAI>::difference_type dif_t;
  typedef typename std::iterator_traits<RAI>::value_type val_t;
  RAI a0, aN;
  C cmp;
  dif_t d, *n_threshold;
  boost::threadpool::pool *pool;
  parallel_intro_sort(RAI _a0, RAI _aN, C _cmp, dif_t _d, dif_t *_n_threshold, boost::threadpool::pool *_pool) :
    a0(_a0), aN(_aN), cmp(_cmp), d(_d), n_threshold(_n_threshold), pool(_pool) {}
  void operator()()
  {
    const dif_t N = aN - a0;
    const dif_t L = N - 1;
    if (N < 40) { insert_sort(a0, aN, cmp); return; }
    if (d == 0) { uppop_heap_sort(a0, aN, cmp); return; }
    const RAI& a = a0;
    const dif_t M = N >> 1;
    val_t l, p = a[M], u;
    if (N > 7) {
      l = a[0];
      u = a[L];
      if (N > 40) {
	const dif_t s = M >> 1, d = s >> 1;
	l = median3(l, a[d], a[s], cmp);
	p = median3(a[M-d], p, a[M+d], cmp);
	u = median3(a[N-s], a[N-d], u, cmp);
      }
      p = median3(l, p, u, cmp);
    }
    RAI i = a0, j = a0 + L;
    while (!0) {
      while (cmp(*i, p)) ++i;
      while (cmp(p, *j)) --j;
      if (i >= j) break;
      boost::swap(*i, *j);
      ++i; --j;
    }
    ++j;
    const dif_t d2 = d - 1;
    dif_t h;
    parallel_intro_sort<RAI, C>
      functor1(a0, i, cmp, d2, n_threshold, pool),
      functor2(j, aN, cmp, d2, n_threshold, pool);
    if ((h = i - a0) > 1) {
      if (h > *n_threshold) pool->schedule(functor1);
      else functor1();
    }
    if (aN - j > 1) {
      functor2();
    }
  }
};
template <class RAI, class C>
inline void parallel_sort(RAI a0, RAI aN, C cmp)
{
  unsigned int n_cpu = boost::thread::hardware_concurrency();
  if (n_cpu > 1) {
    typedef typename std::iterator_traits<RAI>::difference_type dif_t;
    dif_t N = aN - a0, depth, n_threshold;
    parallel_intro_sort_limiter(N, depth, n_threshold);
    boost::threadpool::pool pool(n_cpu/*-1*/);
    parallel_intro_sort<RAI, C> functor(a0, aN, cmp, depth, &n_threshold, &pool);
    pool.schedule(functor);
  }
  else
    sequential_sort(a0, aN, cmp);
}
#endif
template <class RAI>
inline void parallel_sort(RAI a0, RAI aN)
{
  typedef typename std::iterator_traits<RAI>::value_type val_t;
  parallel_sort(a0, aN, std::less<val_t>());
}

#endif // _sort_hpp_
