/*
  sort_r.c - reentrant sort variations such as qsort_r with search

  Copyright (C) 2010 Taiji Yamada <taiji@aihara.co.jp>

  Distributed under the Boost Software License, Version 1.0. See
  http://www.boost.org/LICENSE_1_0.txt for more details.

  References:
  [1] J. Bentley, ``Programming Pearls, second edition,'' Addison-Wesley, 2000.
  [2] http://www.moon.sannet.ne.jp/okahisa/sort/sort.html, 2010.
  [3] http://fxr.watson.org/fxr/source/stdlib/qsort.c?v=FREEBSD7-LIBC, 1992.
  [4] http://www.opensource.apple.com/source/Libc/Libc-262.2.12/stdlib/merge.c, 2003.
  [5] http://docs.sun.com/app/docs/doc/819-0390/, 2008.
 */
#include <stdlib.h>
#include <stdint.h>
#include <sys/types.h>
#if defined(__APPLE__) || defined(__FreeBSD__)
#include <sys/sysctl.h>
#endif
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <math.h>
#include <unistd.h>

typedef struct thread_pool {
  struct pool_job {
    void *(*func)(void *);
    void *arg;
    struct pool_job *next;
  } *job_head, *job_tail;
  pthread_t *thrs;
  size_t n_thread, c_idle, c_job;
  pthread_mutex_t mutex;
  pthread_cond_t cond_work, cond_wait;
  unsigned long flags;
} thread_pool_t;
static
void *thread_pool_work(void *arg)
{
  struct thread_pool *pool = (struct thread_pool *)arg;
  struct pool_job *job;

  pthread_mutex_lock(&pool->mutex);
  do {
    pool->c_idle++;
    while (!pool->job_head && !(pool->flags & (1UL<<1)))
      pthread_cond_wait(&pool->cond_work, &pool->mutex);
    pool->c_idle--;
    if (pool->job_head) {
      struct pool_job the_job = *(job = pool->job_head);

      pool->job_head = job->next;
      if (job == pool->job_tail)
	pool->job_tail = NULL;
      pthread_mutex_unlock(&pool->mutex);
      free(job);
      the_job.func(the_job.arg);
      pthread_mutex_lock(&pool->mutex);
      pool->c_job--;
    }
    if ((pool->flags & (1UL<<0)) && !pool->job_head) {
      pool->flags &= ~(1UL<<0);
      pthread_cond_broadcast(&pool->cond_wait);
    }
  } while (!(pool->flags & (1UL<<1)));
  pthread_mutex_unlock(&pool->mutex);
  return NULL;
}
thread_pool_t *thread_pool_create(int n_thread, pthread_attr_t *attr)
{
  thread_pool_t *pool;
  int i;

  if (!(pool = malloc(sizeof(thread_pool_t)))) return NULL;
  pool->job_head = pool->job_tail = NULL;
  pthread_mutex_init(&pool->mutex, NULL);
  pthread_cond_init(&pool->cond_work, NULL);
  pthread_cond_init(&pool->cond_wait, NULL);
  if (!(pool->thrs = malloc(sizeof(pthread_t)*(pool->n_thread=n_thread)))) { free(pool); return NULL; }
  pool->c_idle = pool->c_job = 0;
  pool->flags = 0;
  for (i=0; i<n_thread; i++)
    pthread_create(&pool->thrs[i], attr, thread_pool_work, pool);
  return pool;
}
int thread_pool_queue(thread_pool_t *pool, void *(*func)(void *), void *arg)
{
  struct pool_job *job;

  if (!(job = malloc(sizeof(struct pool_job)))) return 1;
  job->func = func;
  job->arg = arg;
  job->next = NULL;

  pthread_mutex_lock(&pool->mutex);
  if (!pool->job_head)
    pool->job_head = job;
  else
    pool->job_tail->next = job;
  pool->job_tail = job;
  pool->c_job++;
  if (pool->c_idle > 0)
    pthread_cond_signal(&pool->cond_work);
  pthread_mutex_unlock(&pool->mutex);
  return 0;
}
void thread_pool_wait(thread_pool_t *pool)
{
  pthread_mutex_lock(&pool->mutex);
  while (pool->c_job) {
    pool->flags |= (1UL<<0);
    pthread_cond_wait(&pool->cond_wait, &pool->mutex);
  }
  pthread_mutex_unlock(&pool->mutex);
}
void thread_pool_join(thread_pool_t *pool)
{
  int i;

  pthread_mutex_lock(&pool->mutex);
  while (pool->c_job) {
    pool->flags |= (1UL<<0);
    pthread_cond_wait(&pool->cond_wait, &pool->mutex);
  }
  pool->flags |= (1UL<<1);
  pthread_cond_broadcast(&pool->cond_work);
  pthread_mutex_unlock(&pool->mutex);

  for (i=0; i<pool->n_thread; i++)
    pthread_join(pool->thrs[i], NULL);

  free(pool->thrs);
  free(pool);
}

#define min(a, b) ((a) < (b) ? (a) : (b))

/* swap for arbitrary type */
#define TSWAP(T, a, b) do {\
  T tmp;\
  tmp = *(T *)(a);\
  *(T *)(a) = *(T *)(b);\
  *(T *)(b) = tmp;\
} while (0)
#define SWAP(a, b) do {\
  struct tmp_t { char t[sizeof(*a)]; };\
  TSWAP(struct tmp_t, a, b);\
} while(0)

/* swap for number of arbitrary type */
#define TSWAPS(T, a, b, n) do {\
  int c;\
  for (c=0; c<(n); c++)\
    TSWAP(T, &((T *)a)[c], &((T *)b)[c]);\
} while(0)
#define SWAPS(a, b, n) do {\
  int c;\
  for (c=0; c<(n); c++)\
    SWAP(&(a)[c], &(b)[c]);\
} while(0)

/* fast swap for special type */
#define FSWAP(a, b) do {\
  if (sizeof(*(a)) == sizeof(int8_t))		TSWAP(int8_t, a, b);\
  else if (sizeof(*(a)) == sizeof(int16_t))	TSWAP(int16_t, a, b);\
  else if (sizeof(*(a)) == sizeof(int32_t))	TSWAP(int32_t, a, b);\
  else if (sizeof(*(a)) == sizeof(int64_t))	TSWAP(int64_t, a, b);\
  else if (!(sizeof(*(a))%sizeof(int64_t)))	TSWAPS(int64_t, a, b, sizeof(*(a))/sizeof(int64_t));\
  else if (!(sizeof(*(a))%sizeof(int32_t)))	TSWAPS(int32_t, a, b, sizeof(*(a))/sizeof(int32_t));\
  else if (!(sizeof(*(a))%sizeof(int16_t)))	TSWAPS(int16_t, a, b, sizeof(*(a))/sizeof(int16_t));\
  else SWAP(a, b);\
} while (0)

/* fast swap for number of special type */
#if 0
#define FSWAPS(a, b, n) do {\
  int d;\
  for (d=0; d<(n); d++)\
    FSWAP(&(a)[d], &(b)[d]);\
} while(0)
#else
#define FSWAPS(a, b, n) do {\
  int d;\
  if (sizeof(*(a)) == sizeof(int8_t))		for (d=0; d<(n); d++) TSWAP(int8_t, &(a)[d], &(b)[d]);\
  else if (sizeof(*(a)) == sizeof(int16_t))	for (d=0; d<(n); d++) TSWAP(int16_t, &(a)[d], &(b)[d]);\
  else if (sizeof(*(a)) == sizeof(int32_t))	for (d=0; d<(n); d++) TSWAP(int32_t, &(a)[d], &(b)[d]);\
  else if (sizeof(*(a)) == sizeof(int64_t))	for (d=0; d<(n); d++) TSWAP(int64_t, &(a)[d], &(b)[d]);\
  else if (!(sizeof(*(a))%sizeof(int64_t)))	for (d=0; d<(n); d++) TSWAPS(int64_t, &(a)[d], &(b)[d], sizeof(*(a))/sizeof(int64_t));\
  else if (!(sizeof(*(a))%sizeof(int32_t)))	for (d=0; d<(n); d++) TSWAPS(int32_t, &(a)[d], &(b)[d], sizeof(*(a))/sizeof(int32_t));\
  else if (!(sizeof(*(a))%sizeof(int16_t)))	for (d=0; d<(n); d++) TSWAPS(int16_t, &(a)[d], &(b)[d], sizeof(*(a))/sizeof(int16_t));\
  else SWAPS(a, b, n);\
} while (0)
#endif

/* substitution for arbitrary type */
#define TCOPY(T, a, b) do {\
  *((T *)b) = *((T *)a);\
} while (0)

/* substitution for number of arbitrary type */
#define TCOPYS(T, a, b, n) do {\
  int c;\
  for (c=0; c<(n); c++)\
    TCOPY(T, &((T *)a)[c], &((T *)b)[c]);\
} while (0)

/* median of 3 values under cmp for qsort_r */
#define MED3(a, b, c, th, cmp) (cmp(th, a, b) < 0 ? (cmp(th, b, c) < 0 ? (b) : (cmp(th, a, c) < 0 ? (c) : (a))) : (cmp(th, b, c) > 0 ? (b) : (cmp(th, a, c) < 0 ? (a) : (c))))

inline void introsort_r_limiter(size_t n, size_t *depth)
{
#if !defined(__BLOCKS__)
  size_t N = n, fls0 = 0;
  for (; N>1; N>>=1, ++fls0);
  *depth = fls0*2;
#else
  size_t fls = flsl(n);
  *depth = (fls-1)*2;
#endif
}
inline void pintrosort_r_limiter(size_t n, size_t *depth, size_t *n_threshold)
{
#if !defined(__BLOCKS__)
  size_t N = n, fls0 = 0, s = 1;
  for (; N>1; N>>=1, ++fls0);
  *depth = fls0*2;
  s <<= (fls0+1)/2;
  *n_threshold = (s + n/s)/2;
#else
  size_t fls = flsl(n), s = 1;
  *depth = (fls-1)*2;
  s <<= fls/2;
  *n_threshold = (s + n/s)/2;
#endif
}

/* Exchange sort variations */
#if 0
/* Simple implementation */
void bubblesort_r(void *base, size_t n, size_t z,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base;
  size_t i;
  int flag = 0;

  if (n < 2) return;
  while (!flag) {
    flag = !0;
    for (i=0; i<n-1; i++)
      if (cmp(thunk, &b[i], &b[i+1]) > 0) {
	FSWAP(&b[i], &b[i+1]);
	flag = 0;
      }
  }
}
#elif 0
/* Optimized implementation */
void bubblesort_r(void *base, size_t n, size_t z,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base;
  size_t i, k = n, m;

  if (!n) return;
  do {
    m = 0;
    for (i=0; i+1<k; i++)
      if (cmp(thunk, &b[i], &b[i+1]) > 0) {
	FSWAP(&b[i], &b[i+1]);
	m = i+1;
      }
    k = m;
  } while (k > 1);
}
#else
/* Optimized implementation */
#define BUBBLESORT(T) do {\
  T *b = (T *)base;\
  size_t i, k = n, m;\
\
  if (!n) return;\
  do {\
    m = 0;\
    for (i=0; i+1<k; i++)\
      if (cmp(thunk, &b[i], &b[i+1]) > 0) {\
	TSWAP(T, &b[i], &b[i+1]);\
	m = i+1;\
      }\
    k = m;\
  } while (k > 1);\
} while (0)
void bubblesort_r(void *base, size_t n, size_t z,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (z == sizeof(int8_t))
    BUBBLESORT(int8_t);
  else if (z == sizeof(int16_t))
    BUBBLESORT(int16_t);
  else if (z == sizeof(int32_t))
    BUBBLESORT(int32_t);
  else if (z == sizeof(int64_t))
    BUBBLESORT(int64_t);
  else {
    typedef struct { char t[z]; } temp_t;
    BUBBLESORT(temp_t);
  }
}
#endif

#if 0
void oddevensort_r(void *base, size_t n, size_t z,
		   void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base;
  size_t i;
  int flag = 0;

  if (n < 2) return;
  while (!flag) {
    flag = !0;
    for (i=0; i<n-1; i+=2)
      if (cmp(thunk, &b[i], &b[i+1]) > 0) {
	FSWAP(&b[i], &b[i+1]);
	flag = 0;
      }
    for (i=1; i<n-1; i+=2)
      if (cmp(thunk, &b[i], &b[i+1]) > 0) {
	FSWAP(&b[i], &b[i+1]);
	flag = 0;
      }
  }
}
#else
#define ODDEVENSORT(T) do {\
  T *b = (T *)base;\
  size_t i;\
  int flag = 0;\
\
  if (n < 2) return;\
  while (!flag) {\
    flag = !0;\
    for (i=0; i<n-1; i+=2)\
      if (cmp(thunk, &b[i], &b[i+1]) > 0) {\
	TSWAP(T, &b[i], &b[i+1]);\
	flag = 0;\
      }\
    for (i=1; i<n-1; i+=2)\
      if (cmp(thunk, &b[i], &b[i+1]) > 0) {\
	TSWAP(T, &b[i], &b[i+1]);\
	flag = 0;\
      }\
  }\
} while (0)
void oddevensort_r(void *base, size_t n, size_t z,
		   void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (z == sizeof(int8_t))
    ODDEVENSORT(int8_t);
  else if (z == sizeof(int16_t))
    ODDEVENSORT(int16_t);
  else if (z == sizeof(int32_t))
    ODDEVENSORT(int32_t);
  else if (z == sizeof(int64_t))
    ODDEVENSORT(int64_t);
  else {
    typedef struct { char t[z]; } temp_t;
    ODDEVENSORT(temp_t);
  }
}
#endif

#if 0
void combsort_r(void *base, size_t n, size_t z,
		void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base;
  size_t i, j, gap = n;
  int flag = 0;

  if (n < 2) return;
  while (!flag || gap > 1) {
    gap /= 1.3;
    if (gap == 9 || gap == 10) gap = 11;
    else if (!gap) gap = 1;
    flag = !0;
    for (i=0, j=gap; j!=n; i++, j++)
      if (cmp(thunk, &b[i], &b[j]) > 0) {
	FSWAP(&b[i], &b[j]);
	flag = 0;
      }
  }
}
#else
#define COMBSORT(T) do {\
  T *b = (T *)base;\
  size_t i, j, gap = n;\
  int flag = 0;\
\
  if (n < 2) return;\
  while (!flag || gap > 1) {\
    gap /= 1.3;\
    if (gap == 9 || gap == 10) gap = 11;\
    else if (!gap) gap = 1;\
    flag = !0;\
    for (i=0, j=gap; j!=n; i++, j++)\
      if (cmp(thunk, &b[i], &b[j]) > 0) {\
	TSWAP(T, &b[i], &b[j]);\
	flag = 0;\
      }\
  }\
} while (0)
void combsort_r(void *base, size_t n, size_t z,
		void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (z == sizeof(int8_t))
    COMBSORT(int8_t);
  else if (z == sizeof(int16_t))
    COMBSORT(int16_t);
  else if (z == sizeof(int32_t))
    COMBSORT(int32_t);
  else if (z == sizeof(int64_t))
    COMBSORT(int64_t);
  else {
    typedef struct { char t[z]; } temp_t;
    COMBSORT(temp_t);
  }
}
#endif

/* Insertion sort variations */
#if 0
void insertsort_r(void *base, size_t n, size_t z,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base, *p, *q;

#if !0
  for (p=b+1; p<b+n; p++)
    for (q=p; q>b && cmp(thunk, q-1, q) > 0; q--)
      FSWAP((q-1), q);
#else /* not rapid */
  for (p=b+1; p<b+n; p++) {
    tmp = *p;
    for (q=p; q>b && cmp(thunk, q-1, &tmp) > 0; q--)
      *q = *(q-1);
    *q = tmp;
  }
#endif
}
#else
#if !0
#define INSERTSORT(T, base) do {\
  T *b = (T *)base, *p, *q;\
\
  for (p=b+1; p<b+n; p++)\
    for (q=p; q>b && cmp(thunk, q-1, q) > 0; q--)\
      TSWAP(T, (q-1), q);\
} while (0)
#elif 0 /* not rapid */
#define INSERTSORT(T, base) do {\
  T tmp, *b = (T *)base, *p, *q;\
\
  for (p=b+1; p<b+n; p++) {\
    tmp = *p;\
    for (q=p; q>b && cmp(thunk, q-1, &tmp) > 0; q--)\
      *q = *(q-1);\
    *q = tmp;\
  }\
} while (0)
#elif 0 /* not rapid */
#define INSERTSORT(T, base) do {\
  T *b = (T *)base, *p, *q, *r;\
\
  for (p=b+1; p<b+n; p++) {\
    q = r = p;\
    if (cmp(thunk, r, b) < 0)\
      while (b != q) { --r; TSWAP(T, r, q); q = r; }\
    else\
      while (cmp(thunk, q, --r) < 0) { TSWAP(T, q, r); q = r; }\
  }\
} while (0)
#else /* not rapid */
#define INSERTSORT(T, base) do {\
  T tmp, *b = (T *)base, *p, *q, *r;\
\
  for (p=b+1; p<b+n; p++) {\
    tmp = *p;\
    q = r = p;\
    if (cmp(thunk, &tmp, b) < 0)\
      while (b != q) { --r; *q = *r; q = r; }\
    else\
      while (cmp(thunk, &tmp, --r) < 0) { *q = *r; q = r; }\
    *q = tmp;\
  }\
} while (0)
#endif
void insertsort_r(void *base, size_t n, size_t z,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (z == sizeof(int8_t))
    INSERTSORT(int8_t, base);
  else if (z == sizeof(int16_t))
    INSERTSORT(int16_t, base);
  else if (z == sizeof(int32_t))
    INSERTSORT(int32_t, base);
  else if (z == sizeof(int64_t))
    INSERTSORT(int64_t, base);
  else {
    typedef struct { char t[z]; } temp_t;
    INSERTSORT(temp_t, base);
  }
}
#endif

#if 0
void shellsort_r(void *base, size_t n, size_t z,
		 void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base;
  size_t h, i, k;
  int j;

  if (n < 2) return;
  for (h=1; h<n; h=h*3+1); h /= 3;
  {
    while (h != 1) {
      for (k=0; k<h; k++) {
#if !0
	for (i=k; i<n; i+=h)
	  for (j=i; j>h && cmp(thunk, &b[j-h], &b[j]) > 0; j-=h)
	    FSWAP(&b[j-h], &b[j]);
#else /* not rapid */
	struct temp_t tmp;

	for (i=k; i<n; i+=h) {
	  for (tmp = b[i], j=i; j>h && cmp(thunk, &b[j-h], &tmp) > 0; j-=h)
	    b[j] = b[j-h];
	  b[j] = tmp;
	}
#endif
      }
      h /= 3;
    }
  }
  insertsort_r(base, n, z, thunk, cmp);
}
#else
#define SHELLSORT(T) do {\
  T *b = (T *)base;\
  size_t h, i, k;\
  int j;\
\
  if (n < 2) return;\
  for (h=1; h<n; h=h*3+1); h /= 3;\
  {\
    while (h != 1) {\
      for (k=0; k<h; k++)\
	for (i=k; i<n; i+=h)\
	  for (j=i; j>h && cmp(thunk, &b[j-h], &b[j]) > 0; j-=h)\
	    TSWAP(T, &b[j-h], &b[j]);\
      h /= 3;\
    }\
  }\
  INSERTSORT(T, base);\
} while (0)
void shellsort_r(void *base, size_t n, size_t z,
		 void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (z == sizeof(int8_t))
    SHELLSORT(int8_t);
  else if (z == sizeof(int16_t))
    SHELLSORT(int16_t);
  else if (z == sizeof(int32_t))
    SHELLSORT(int32_t);
  else if (z == sizeof(int64_t))
    SHELLSORT(int64_t);
  else {
    typedef struct { char t[z]; } temp_t;
    SHELLSORT(temp_t);
  }
}
#endif

#if 0
void icombsort_r(void *base, size_t n, size_t z,
		 void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base;
  size_t i, j, gap = n;

  if (n < 2) return;
  while (gap > 1) {
    gap /= 1.37;
    if (gap == 9 || gap == 10) gap = 11;
    for (i=0, j=gap; j!=n; i++, j++)
      if (cmp(thunk, &b[i], &b[j]) > 0)
	FSWAP(&b[i], &b[j]);
  }
  insertsort_r(base, n, z, thunk, cmp);
}
#else
#define ICOMBSORT(T) do {\
  T *b = (T *)base;\
  size_t i, j, gap = n;\
\
  if (n < 2) return;\
  while (gap > 1) {\
    gap /= 1.37;\
    if (gap == 9 || gap == 10) gap = 11;\
    for (i=0, j=gap; j!=n; i++, j++)\
      if (cmp(thunk, &b[i], &b[j]) > 0)\
	TSWAP(T, &b[i], &b[j]);\
  }\
  INSERTSORT(T, base);\
} while (0)
void icombsort_r(void *base, size_t n, size_t z,
		 void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (z == sizeof(int8_t))
    ICOMBSORT(int8_t);
  else if (z == sizeof(int16_t))
    ICOMBSORT(int16_t);
  else if (z == sizeof(int32_t))
    ICOMBSORT(int32_t);
  else if (z == sizeof(int64_t))
    ICOMBSORT(int64_t);
  else {
    typedef struct { char t[z]; } temp_t;
    ICOMBSORT(temp_t);
  }
}
#endif

/* Heap sort variations */

#define HEAPUP(u) do {\
  int j, p;\
\
  j = u;\
  for (;;) {\
    if (j == 1) break;\
    p = j / 2;\
    if (cmp(thunk, &a[p], &a[j]) >= 0) break;\
    FSWAP(&a[p], &a[j]);\
    j = p;\
  }\
} while (0)
#if !0
#define HEAPDOWN(l, u) do {\
  int j, k, p;\
\
  j = l;\
  for (;;) {\
    k = 2*j;\
    if (k > u) break;\
    if (k+1 <= u && cmp(thunk, &a[k+1], &a[k]) > 0)\
      k++;\
    a[j] = a[k];\
    j = k;\
  }\
  for (;;) {\
    p = j/2;\
    if (p < l) break;\
    if (cmp(thunk, &a[p], &a[j]) > 0) break;\
    FSWAP(&a[p], &a[j]);\
    j = p;\
  }\
} while (0)
#else /* not rapid */
#define HEAPDOWN(l, u) do {\
  int j, k, p;\
\
  tmp = a[l];\
  j = l;\
  for (;;) {\
    k = 2*j;\
    if (k > u) break;\
    if (k+1 <= u && cmp(thunk, &a[k+1], &a[k]) > 0)\
      k++;\
    a[j] = a[k];\
    j = k;\
  }\
  a[j] = tmp;\
  for (;;) {\
    p = j/2;\
    if (p < l) break;\
    if (cmp(thunk, &a[p], &a[j]) > 0) break;\
    FSWAP(&a[p], &a[j]);\
    j = p;\
  }\
} while (0)
#endif

#if !0
#define HEAP_DOWN(T, r, q) do {\
  int k, p = (r);\
\
  while ((k = (p << 1) + 1) <= q) {\
    if (k < q && cmp(thunk, &b[k], &b[k+1]) < 0) k++;\
    if (cmp(thunk, &b[p], &b[k]) >= 0) break;\
    TSWAP(T, &b[p], &b[k]);\
    p = k;\
  }\
} while (0)
#else /* not rapid */
#define HEAP_DOWN(T, r, q) do {\
  int k, p = (r);\
\
  tmp = b[p];\
  while ((k = (p << 1) + 1) <= q) {\
    if (k < q && cmp(thunk, &b[k], &b[k+1]) < 0) k++;\
    if (cmp(thunk, &tmp, &b[k]) >= 0) break;\
    b[p] = b[k];\
    p = k;\
  }\
  b[p] = tmp;\
} while (0)
#endif
#if !0
#define HEAP_UP(T, u) do {\
  int k = (u);\
\
  while (k > 0) {\
    int p = (k - 1) >> 1;\
    if (cmp(thunk, &b[p], &b[k]) >= 0) break;\
    TSWAP(T, &b[k], &b[p]);\
    k = p;\
  }\
} while (0)
#else /* not rapid */
#define HEAP_UP(T, u) do {\
  int k = (u);\
\
  tmp = b[k];\
  while (k > 0) {\
    int p = (k - 1) >> 1;\
    if (cmp(thunk, &b[p], &tmp) >= 0) break;\
    b[k] = b[p];\
    k = p;\
  }\
  b[k] = tmp;\
} while (0)
#endif
#if !0
#define HEAP_POP(T, q, r) do {\
  int p = 0, k;\
\
  TSWAP(T, &b[0], &b[r]);\
  while ((k = (p << 1) + 1) <= q) {\
    if (k < q && cmp(thunk, &b[k], &b[k+1]) < 0) k++;\
    TSWAP(T, &b[p], &b[k]);\
    p = k;\
  }\
  HEAP_UP(T, p);\
} while (0)
#else /* not rapid */
#define HEAP_POP(T, q, r) do {\
  int p = 0, k;\
\
  tmp = b[r];\
  b[r] = b[0];\
  while ((k = (p << 1) + 1) <= q) {\
    if (k < q && cmp(thunk, &b[k], &b[k+1]) < 0) k++;\
    b[p] = b[k];\
    p = k;\
  }\
  b[p] = tmp;\
  HEAP_UP(T, p);\
} while (0)
#endif

inline
void heap_down(void *base, int r, int q, size_t z,
	       void *thunk, int (*cmp)(void *, const void *, const void *))
{
  typedef struct { char t[z]; } temp_t;
  temp_t /*tmp, */*b = (temp_t *)base;

  HEAP_DOWN(temp_t, r, q);
}
inline
void heap_up(void *base, int u, size_t z,
	     void *thunk, int (*cmp)(void *, const void *, const void *))
{
  typedef struct { char t[z]; } temp_t;
  temp_t /*tmp, */*b = (temp_t *)base;

  HEAP_UP(temp_t, u);
}
inline
void heap_pop(void *base, int q, int r, size_t z,
	      void *thunk, int (*cmp)(void *, const void *, const void *))
{
  typedef struct { char t[z]; } temp_t;
  temp_t /*tmp, */*b = (temp_t *)base;

  HEAP_POP(temp_t, q, r);
}

#if 0
int heapsort_r(void *base, size_t n, size_t z,
	       void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base, *a;
  int i;

  a = b-1;
  for (i = 2; i <= n; i++)
    HEAPUP(i);
  for (i = n; i >= 2; i--) {
    FSWAP(&a[1], &a[i]);
    HEAPDOWN(1, i-1);
  }
  return 0;
}
#elif 0
int heapsort_r(void *base, size_t n, size_t z,
	       void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base, *a;
  int i;

  a = b-1;
  for (i = n/2; i >= 1; i--)
    HEAPDOWN(i, (int)n);
  for (i = n; i >= 2; i--) {
    FSWAP(&a[1], &a[i]);
    HEAPDOWN(1, i-1);
  }
  return 0;
}
#elif 0
int heapsort_r(void *base, size_t n, size_t z,
	       void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base;
  int i, m = n-1;

  for (i=n>>1; i>0; ) {
    --i;
    HEAP_DOWN(struct temp_t, i, m);
  }
  for (i=m; i>0; i--) {
    FSWAP(&b[0], &b[i]);
    HEAP_DOWN(struct temp_t, 0, i-1);
  }
  return 0;
}
#elif 0
int heapsort_r(void *base, size_t n, size_t z,
	       void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base;
  int i, m = n-1;

  for (i=1; i<n; i++)
    HEAP_UP(struct temp_t, i);
  for (i=m; i>0; i--) {
    FSWAP(&b[0], &b[i]);
    HEAP_DOWN(struct temp_t, 0, i-1);
  }
  return 0;
}
#elif 0
int heapsort_r(void *base, size_t n, size_t z,
	       void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } tmp, *b = (struct temp_t *)base;
  int i, m = n-1;

  for (i=1; i<(int)n; i++)
    HEAP_UP(struct temp_t, i);
  for (i=m; i>0; i--)
    HEAP_POP(struct temp_t, i-1, i);
  return 0;
}
#else
#define HEAPSORT(T) do {\
  T /*tmp, */*b = (T *)base;\
  int i, m = n-1;\
\
  for (i=1; i<(int)n; i++)\
    HEAP_UP(T, i);\
  for (i=m; i>0; i--)\
    HEAP_POP(T, i-1, i);\
} while (0)
int heapsort_r(void *base, size_t n, size_t z,
	       void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (z == sizeof(int8_t))
    HEAPSORT(int8_t);
  else if (z == sizeof(int16_t))
    HEAPSORT(int16_t);
  else if (z == sizeof(int32_t))
    HEAPSORT(int32_t);
  else if (z == sizeof(int64_t))
    HEAPSORT(int64_t);
  else {
    typedef struct { char t[z]; } temp_t;
    HEAPSORT(temp_t);
  }
  return 0;
}
#endif

#if 0
#define PARTIALSORT(T) do {\
  T /*tmp, */*b = (T *)base;\
  int i, m = l-1;\
\
  if (!l) return;\
  for (i=1; i<(int)l; i++)\
    HEAP_UP(T, i);\
  for (i=(int)l; i<(int)n; i++)\
    if (cmp(thunk, &b[i], &b[0]) < 0) {\
      TSWAP(T, &b[0], &b[i]);\
      HEAP_DOWN(T, 0, m);\
    }\
  for (i=m; i>0; i--) {\
    TSWAP(T, &b[0], &b[i]);\
    HEAP_DOWN(T, 0, i-1);\
  }\
} while (0)
#else
#define PARTIALSORT(T) do {\
  T /*tmp, */*b = (T *)base;\
  int i, m = l-1;\
\
  if (!l) return;\
  for (i=1; i<(int)l; i++)\
    HEAP_UP(T, i);\
  for (i=(int)l; i<(int)n; i++)\
    if (cmp(thunk, &b[i], &b[0]) < 0)\
      HEAP_POP(T, m, i);\
  for (i=m; i>0; i--)\
    HEAP_POP(T, i-1, i);\
} while (0)
#endif
void partialsort_r(void *base, size_t l, size_t n, size_t z,
		   void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (z == sizeof(int8_t))
    PARTIALSORT(int8_t);
  else if (z == sizeof(int16_t))
    PARTIALSORT(int16_t);
  else if (z == sizeof(int32_t))
    PARTIALSORT(int32_t);
  else if (z == sizeof(int64_t))
    PARTIALSORT(int64_t);
  else {
    typedef struct { char t[z]; } temp_t;
    PARTIALSORT(temp_t);
  }
}

/* Merge sort variations */
#if 0
static
void _mergesort_r(void *area, void *an,
		  void *base, void *bn, size_t z, size_t n,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *a0 = (struct temp_t *)area, *am = a0, *b0 = (struct temp_t *)base, *bm = b0, *ai, *aj, *bk/*, *l*/;
  int m = n>>1;

  if (n < 34) {
    insertsort_r(area, n, z, thunk, cmp);
    return;
  }
  am += m;
  bm += m;
  _mergesort_r(a0, am, b0, bm, z, m, thunk, cmp);
  _mergesort_r(am, an, bm, bn, z, n - m, thunk, cmp);
  memcpy(b0, a0, (char *)am - (char *)a0);/*for (bk = b0, l=a0; l!=am; l++, bk++) *bk = *l;*/
  ai = a0;
  aj = am;
  bk = b0;
  while (bk != bm && aj != an) {
    if (cmp(thunk, aj, bk) < 0) {
      *ai = *aj;
      ++ai;
      ++aj;
    }
    else {
      *ai = *bk;
      ++ai;
      ++bk;
    }
  }
  memcpy(ai, bk, (char *)bm - (char *)bk);/*for (l=bk; l!=bm; l++, ai++) *ai = *l;*/
}
int mergesort_r(void *base, size_t n, size_t z,
		void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *a = malloc(z*n), *b = (struct temp_t *)base;
  if (!a) return -1;
  _mergesort_r(b, &b[n], a, &a[n], z, n, thunk, cmp);
  free(a);
  return 0;
}
#elif !0
#define MERGESORT(T) do {\
  T *a0 = (T *)area, *am = a0, *b0 = (T *)base, *bm = b0, *ai, *aj, *bk/*, *l*/;\
  int m = n>>1;\
\
  if (n < 34) {\
    INSERTSORT(T, area);\
    return;\
  }\
  am += m;\
  bm += m;\
  _mergesort_r(a0, am, b0, bm, z, m, thunk, cmp);\
  _mergesort_r(am, an, bm, bn, z, n - m, thunk, cmp);\
  memcpy(b0, a0, (char *)am - (char *)a0);/*for (bk = b0, l=a0; l!=am; l++, bk++) *bk = *l;*/\
  ai = a0;\
  aj = am;\
  bk = b0;\
  while (bk != bm && aj != an) {\
    if (cmp(thunk, aj, bk) < 0) {\
      *ai = *aj;\
      ++ai;\
      ++aj;\
    }\
    else {\
      *ai = *bk;\
      ++ai;\
      ++bk;\
    }\
  }\
  memcpy(ai, bk, (char *)bm - (char *)bk);/*for (l=bk; l!=bm; l++, ai++) *ai = *l;*/\
} while (0)
static
void _mergesort_r(void *area, void *an,
		  void *base, void *bn, size_t z, size_t n,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (z == sizeof(int8_t))
    MERGESORT(int8_t);
  else if (z == sizeof(int16_t))
    MERGESORT(int16_t);
  else if (z == sizeof(int32_t))
    MERGESORT(int32_t);
  else if (z == sizeof(int64_t))
    MERGESORT(int64_t);
  else {
    typedef struct { char t[z]; } temp_t;
    MERGESORT(temp_t);
  }
}
int mergesort_r(void *base, size_t n, size_t z,
		void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *a = malloc(z*n), *b = (struct temp_t *)base;
  if (!a) return -1;
  _mergesort_r(b, &b[n], a, &a[n], z, n, thunk, cmp);
  free(a);
  return 0;
}
#elif 0
static
void _mergesort_r(void *area, void *an,
		  void *base, void *bn, size_t z, size_t n,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *a0 = (struct temp_t *)area, *am = a0, *b0 = (struct temp_t *)base, *bm = b0, *ai, *aj, *bk/*, *l*/;
  int m = n>>1;

  if (n < 34) {
    insertsort_r(base, n, z, thunk, cmp);
    return;
  }
  am += m;
  bm += m;
  _mergesort_r(b0, bm, a0, am, z, m, thunk, cmp);
  _mergesort_r(bm, bn, am, an, z, n - m, thunk, cmp);
  ai = a0;
  aj = am;
  bk = b0;
  while (ai != am && aj != an) {
    if (cmp(thunk, aj, ai) < 0) {
      *bk = *aj;
      ++bk;
      ++aj;
    }
    else {
      *bk = *ai;
      ++bk;
      ++ai;
    }
  }
  memcpy(bk, ai, (char *)am - (char *)ai);/*for (l=ai; l!=am; l++, bk++) *bk = *l;*/
  memcpy((char *)bk + ((char *)am - (char *)ai), aj, (char *)an - (char *)aj);/*for (l=aj; l!=an; l++, bk++) *bk = *l;*/
}
int mergesort_r(void *base, size_t n, size_t z,
		void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *a = malloc(z*n), *b = (struct temp_t *)base;
  if (!a) return -1;
  _mergesort_r(a, &a[n], b, &b[n], z, n, thunk, cmp);
  free(a);
  return 0;
}
#elif 0
#define MERGESORT(T) do {\
  T *a0 = (T *)area, *am = a0, *b0 = (T *)base, *bm = b0, *ai, *aj, *bk/*, *l*/;\
  int m = n>>1;\
\
  if (n < 34) {\
    INSERTSORT(T, base);\
    return;\
  }\
  am += m;\
  bm += m;\
  _mergesort_r(b0, bm, a0, am, z, m, thunk, cmp);\
  _mergesort_r(bm, bn, am, an, z, n - m, thunk, cmp);\
  ai = a0;\
  aj = am;\
  bk = b0;\
  while (ai != am && aj != an) {\
    if (cmp(thunk, aj, ai) < 0) {\
      *bk = *aj;\
      ++bk;\
      ++aj;\
    }\
    else {\
      *bk = *ai;\
      ++bk;\
      ++ai;\
    }\
  }\
  memcpy(bk, ai, (char *)am - (char *)ai);/*for (l=ai; l!=am; l++, bk++) *bk = *l;*/\
  memcpy((char *)bk + ((char *)am - (char *)ai), aj, (char *)an - (char *)aj);/*for (l=aj; l!=an; l++, bk++) *bk = *l;*/\
} while (0)
static
void _mergesort_r(void *area, void *an,
		  void *base, void *bn, size_t z, size_t n,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (z == sizeof(int8_t))
    MERGESORT(int8_t);
  else if (z == sizeof(int16_t))
    MERGESORT(int16_t);
  else if (z == sizeof(int32_t))
    MERGESORT(int32_t);
  else if (z == sizeof(int64_t))
    MERGESORT(int64_t);
  else {
    struct temp_t { char t[z]; };
    MERGESORT(struct temp_t);
  }
}
int mergesort_r(void *base, size_t n, size_t z,
		void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *a = malloc(z*n), *b = (struct temp_t *)base;
  if (!a) return -1;
  _mergesort_r(a, &a[n], b, &b[n], z, n, thunk, cmp);
  free(a);
  return 0;
}
#else /* not so rapid */
#define MERGESORT(T) do {\
  T *a0 = (T *)area, *am = a0, *b0 = (T *)base, *bm = b0, *ai, *aj, *bk/*, *l*/;\
  int m = n>>1;\
\
  if (n < 34) {\
    INSERTSORT(T, base);\
    return;\
  }\
  am += m;\
  bm += m;\
  _mergesort_r(b0, bm, a0, am, z, m, thunk, cmp);\
  _mergesort_r(bm, bn, am, an, z, n - m, thunk, cmp);\
  ai = a0;\
  aj = am;\
  bk = b0;\
  while (ai != am && aj != an) {\
    if (cmp(thunk, aj, ai) < 0) {\
      *bk = *aj;\
      ++bk;\
      ++aj;\
    }\
    else {\
      *bk = *ai;\
      ++bk;\
      ++ai;\
    }\
  }\
  memcpy(bk, ai, (char *)am - (char *)ai);/*for (l=ai; l!=am; l++, bk++) *bk = *l;*/\
  memcpy((char *)bk + ((char *)am - (char *)ai), aj, (char *)an - (char *)aj);/*for (l=aj; l!=an; l++, bk++) *bk = *l;*/\
} while (0)
static
void _mergesort_r(void *area, void *an,
		  void *base, void *bn, size_t z, size_t n,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (z == sizeof(char))
    MERGESORT(char);
  else if (z == sizeof(int64_t))
    MERGESORT(int64_t);
  else {
    struct temp_t { char t[z]; };
    MERGESORT(struct temp_t);
  }
}
#define MERGESORT_REVERSE(T, a, b) do {\
  T *s = (T *)(b);\
  do {\
    TSWAP(T, a, s);\
    (a) += sizeof(T); s--;\
  } while ((T *)(a) < s);\
} while (0)
#define MERGESORT_NATURAL_MODE
#define MERGESORT_EVAL(p) (u_char **)((u_char *)0 + (((u_char *)p + sizeof(u_char *) - 1 - (u_char *)0) & ~(sizeof(u_char *) - 1)))
#define MERGESORT_NATURAL(T) do {\
  p2 = list2;\
  f1 = list1;\
  sense = (cmp(thunk, f1, f1 + z) > 0);\
  for (; f1 < last; sense = !sense) {\
    int length = 2;\
    u_char *f2, *l2;\
\
    /* Find pairs with same sense. */\
    for (f2 = f1 + z2; f2 < last; f2 += z2) {\
      if ((cmp(thunk, f2, f2+ z) > 0) != sense)\
	break;\
      length += 2;\
    }\
    if (length < 16) {	/* Pairwise merge */\
      do {\
	p2 = *MERGESORT_EVAL(p2) = f1 + z2 - list1 + list2;\
	if (sense > 0)\
	  TSWAP(T, f1, (f1 + z));\
      } while ((f1 += z2) < f2);\
    }\
    else {	/* Natural merge */\
      l2 = f2;\
      for (f2 = f1 + z2; f2 < l2; f2 += z2) {\
	if ((cmp(thunk, f2-z, f2) > 0) != sense) {\
	  p2 = *MERGESORT_EVAL(p2) = f2 - list1 + list2;\
	  if (sense > 0)\
	    MERGESORT_REVERSE(T, f1, f2-z);\
	  f1 = f2;\
	}\
      }\
      if (sense > 0)\
	MERGESORT_REVERSE(T, f1, f2-z);\
      f1 = f2;\
      if (f2 < last || cmp(thunk, f2 - z, f2) > 0)\
	p2 = *MERGESORT_EVAL(p2) = f2 - list1 + list2;\
      else\
	p2 = *MERGESORT_EVAL(p2) = list2 + n*z;\
    }\
  }\
} while (0)
#define MERGESORT_PAIRWISE(T) do {\
  for (f1 = list1, p2 = list2; f1 < last; f1 += z2) {\
    p2 = *MERGESORT_EVAL(p2) = p2 + z2;\
    if (cmp(thunk, f1, f1 + z) > 0)\
      TSWAP(T, f1, f1 + z);\
  }\
} while (0)
#define MERGESORT_COPY1(T) do {\
  if (q == f1) {\
    j = (T *)b - (T *)f2;\
    TCOPYS(T, f2, tp2, j); f2 += z*j; tp2 += z*j;\
    TCOPY(T, f1, tp2); f1 += z; tp2 += z;\
  }\
  else {\
    j = (T *)b - (T *)f1;\
    TCOPYS(T, f1, tp2, j); f1 += z*j; tp2 += z*j;\
    TCOPY(T, f2, tp2); f2 += z; tp2 += z;\
  }\
} while (0)
#define MERGESORT_COPY2(T) do {\
  if (f2 < l2) {\
    j = (T *)l2 - (T *)f2;\
    TCOPYS(T, f2, tp2, j); f2 += z*j; tp2 += z*j;\
  }\
  else if (f1 < l1) {\
    j = (T *)l1 - (T *)f1;\
    TCOPYS(T, f1, tp2, j); f1 += z*j; tp2 += z*j;\
  }\
} while (0)
int mergesort_r(void *base, size_t n, size_t z,
		void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (!(z < sizeof(u_char *)/2) &&
      (z == sizeof(int16_t) ||
       z == sizeof(int32_t))) {
    int i, j, z2, sense, big;
    u_char *f1, *f2, *t, *b, *tp2, *q, *l1, *l2;
    u_char *list2, *list1, *p2, *p, *last, **p1;
    /*struct temp_t { char t[z]; };*/

    if (z < sizeof(u_char *)/2) {	/* Pointers must fit into 2*z. */
      errno = EINVAL;
      return -1;
    }
    if (n == 0)
      return 0;
    if (!(list2 = malloc(z*n + sizeof(u_char *))))
      return -1;

    list1 = base;
    z2 = z*2;
    if (n <= 5) {
      insertsort_r(list1, n, z, thunk, cmp);
      return 0;
    }
    i = 4 + (n & 1);
    insertsort_r(list1 + z*(n-i), i, z, thunk, cmp);
    last = list1 + z*(n-i);
    *MERGESORT_EVAL(list2 + (last - list1)) = list2 + z*n;
#ifdef MERGESORT_NATURAL_MODE
    if (z == sizeof(int16_t))
      MERGESORT_NATURAL(int16_t);
    else
      MERGESORT_NATURAL(int32_t);
#else	/* pairwise merge only. */
    if (z == sizeof(int16_t))
      MERGESORT_PAIRWISE(int16_t);
    else
      MERGESORT_PAIRWISE(int32_t);
#endif
    last = list2 + z*n;
    i = big = 0;
    while (*MERGESORT_EVAL(list2) != last) {
      l2 = list1;
      p1 = MERGESORT_EVAL(list1);
      for (tp2=p2=list2; p2!=last; p1=MERGESORT_EVAL(l2)) {
	p2 = *MERGESORT_EVAL(p2);
	f1 = l2;
	f2 = l1 = list1 + (p2 - list2);
	if (p2 != last)
	  p2 = *MERGESORT_EVAL(p2);
	l2 = list1 + (p2 - list2);
	while (f1 < l1 && f2 < l2) {
	  if (cmp(thunk, f1, f2) <= 0) {
	    q = f2;
	    b = f1;
	    t = l1;
	    sense = -1;
	  }
	  else {
	    q = f1;
	    b = f2;
	    t = l2;
	    sense = 0;
	  }
	  if (!big) {	/* here i = 0 */
	    while ((b += z) < t && cmp(thunk, q, b) > sense)
	      if (++i == 6) {
		big = 1;
		goto EXPONENTIAL;
	      }
	  } else {
	  EXPONENTIAL:
	    for (i=z; ; i <<= 1)
	      if ((p = (b + i)) >= t) {
		if ((p = t - z) > b &&
		    cmp(thunk, q, p) <= sense)
		  t = p;
		else
		  b = p;
		break;
	      }
	      else if (cmp(thunk, q, p) <= sense) {
		t = p;
		if (i == z)
		  big = 0;
		goto FASTCASE;
	      } else
		b = p;
	    while (t > b+z) {
	      i = (((t - b)/z) >> 1) * z;
	      if (cmp(thunk, q, p = b + i) <= sense)
		t = p;
	      else
		b = p;
	    }
	    goto COPY;
	  FASTCASE:
	    while (i > z)
	      if (cmp(thunk, q, p = b + (i >>= 1)) <= sense)
		t = p;
	      else
		b = p;
	  COPY:
	    b = t;
	  }
	  if (z == sizeof(int16_t))
	    MERGESORT_COPY1(int16_t);
	  else
	    MERGESORT_COPY1(int32_t);
	}
	if (z == sizeof(int16_t))
	  MERGESORT_COPY2(int16_t);
	else
	  MERGESORT_COPY2(int32_t);
	*p1 = l2;
      }
      tp2 = list1;	/* swap list1, list2 */
      list1 = list2;
      list2 = tp2;
      last = list2 + n*z;
    }
    if (base == list2) {
      memmove(list2, list1, n*z);
      list2 = list1;
    }
    free(list2);
  }
  else {
    struct temp_t { char t[z]; } *a = malloc(z*n), *b = (struct temp_t *)base;
    if (!a) return -1;
    _mergesort_r(a, &a[n], b, &b[n], z, n, thunk, cmp);
    free(a);
  }
  return 0;
}
#endif

/*
  References:
  [1] http://www.drmaciver.com/2010/01/understanding-timsort-1adaptive-mergesort/, 2010.
 */
#define TIMMERGE(T) do {\
  T *b = (T *)base, *p1 = (T *)a1, *p2 = (T *)a2, *s = (T *)storage, *a = s;\
  size_t i1 = 0, i2 = 0;\
\
  while (i1 < l1 && i2 < l2) {\
    if (cmp(thunk, &p1[i1], &p2[i2]) <= 0) {\
      *a = p1[i1];\
      i1++;\
    }\
    else {\
      *a = p2[i2];\
      i2++;\
    }\
    a++;\
  }\
  memcpy(a, p1 + i1, sizeof(T)*(l1 - i1));\
  memcpy(a, p2 + i2, sizeof(T)*(l2 - i2));\
  memcpy(b, s, sizeof(T)*(l1 + l2));\
} while (0)
static
void timmerge_r(void *base,
		void *a1, size_t l1,
		void *a2, size_t l2, size_t z, void *storage,
		void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (z == sizeof(int8_t))
    TIMMERGE(int8_t);
  else if (z == sizeof(int16_t))
    TIMMERGE(int16_t);
  else if (z == sizeof(int32_t))
    TIMMERGE(int32_t);
  else if (z == sizeof(int64_t))
    TIMMERGE(int64_t);
  else {
    struct temp_t { char t[z]; };
    TIMMERGE(struct temp_t);
  }
}
void _timsort_r(void *base, size_t n, size_t z, void *storage,
		void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (n /*<= 7*/< 34) {
    insertsort_r(base, n, z, thunk, cmp);
    return;
  }
  struct temp_t { char t[z]; } *b = (struct temp_t *)base, *s = (struct temp_t *)storage;
  size_t m = n>>1;

  _timsort_r(b, m, z, s, thunk, cmp);
  _timsort_r(b + m, n - m, z, s + m, thunk, cmp);
  timmerge_r(b, b, m, b + m, n - m, z, s, thunk, cmp);
}
int timsort_r(void *base, size_t n, size_t z,
	      void *thunk, int (*cmp)(void *, const void *, const void *))
{
  void *storage = malloc(z*n);
  if (!storage) return -1;
  _timsort_r(base, n, z, storage, thunk, cmp);
  free(storage);
  return 0;
}

/* Quick sort variations */
#if 0
/* Two-way partitioning */
void qsort_r(void *base, size_t n, size_t z,
	     void *thunk, int (*cmp)(void *, const void *, const void *))
{
  const int u = n-1;
  int i = 0, j = n;
  struct temp_t { char t[z]; } *b = (struct temp_t *)base;

  for (;;) {
    do i++; while (i <= u && cmp(thunk, &b[i], b) < 0);
    do j--; while (cmp(thunk, &b[j], b) > 0);
    if (i > j)
      break;
    FSWAP(&b[i], &b[j]);
  }
  FSWAP(&b[0], &b[j]);
  if (0 < j-1) qsort_r(&b[0], j, z, thunk, cmp);
  if (j+1 < u) qsort_r(&b[j+1], u-j, z, thunk, cmp);
}
#elif 0
/* Two-way partitioning and Insertion */
void qsort_r(void *base, size_t n, size_t z,
	     void *thunk, int (*cmp)(void *, const void *, const void *))
{
  const int u = n-1;
  int i = 0, j = n;
  struct temp_t { char t[z]; } *b = (struct temp_t *)base;

  if (n < 7) {
    for (i=1; i<(int)n; i++)
      for (j=i; j>0 && cmp(thunk, &b[j-1], &b[j]) > 0; j--)
	FSWAP(&b[j-1], &b[j]);
    return;
  }
  for (;;) {
    do i++; while (i <= u && cmp(thunk, &b[i], b) < 0);
    do j--; while (cmp(thunk, &b[j], b) > 0);
    if (i > j)
      break;
    FSWAP(&b[i], &b[j]);
  }
  FSWAP(&b[0], &b[j]);
  if (0 < j-1) qsort_r(&b[0], j, z, thunk, cmp);
  if (j+1 < u) qsort_r(&b[j+1], u-j, z, thunk, cmp);
}
#elif 0
/* Two-way partitioning, Insertion, Median pivot, Swap count and Same values */
void qsort_r(void *base, size_t n, size_t z,
	     void *thunk, int (*cmp)(void *, const void *, const void *))
{
  int i, j, k, l, m, t;
  struct temp_t { char t[z]; } *b = (struct temp_t *)base;
  int c, h;

 loop:
  if (n < 7) {
    for (i=1; i<(int)n; i++)
      for (j=i; j>0 && cmp(thunk, &b[j-1], &b[j]) > 0; j--)
	FSWAP(&b[j-1], &b[j]);
    return;
  }
  m = n/2;
  if (n > 7) {
    int l, u;
    l = 0;
    u = n-1;
    if (n > 40) {
      int s = n/4, d = n/8;
      l = MED3(&b[l], &b[l+d], &b[l+s], thunk, cmp) - b;
      m = MED3(&b[m-d], &b[m], &b[m+d], thunk, cmp) - b;
      u = MED3(&b[u-s], &b[u-d], &b[u], thunk, cmp) - b;
    }
    m = MED3(&b[l], &b[m], &b[u], thunk, cmp) - b;
  }
  FSWAP(&b[0], &b[m]);
  for (c=0, i=k=1, j=l=n-1; ; i++, j--) {
    int v;
    while (i <= j && (v=cmp(thunk, &b[i], b)) <= 0) {
      if (v == 0) {
	FSWAP(&b[k], &b[i]);
	c++;
	k++;
      }
      i++;
    }
    while (i <= j && (v=cmp(thunk, &b[j], b)) >= 0) {
      if (v == 0) {
	FSWAP(&b[j], &b[l]);
	c++;
	l--;
      }
      j--;
    }
    if (i > j) break;
    FSWAP(&b[i], &b[j]);
    c++;
  }
  if (c == 0) {
    for (i=1; i<(int)n; i++)
      for (j=i; j>0 && cmp(thunk, &b[j-1], &b[j]) > 0; j--)
	FSWAP(&b[j-1], &b[j]);
    return;
  }
  t = n-1;
  h = min(k, i-k);
  FSWAPS(&b[0], &b[i-h], h);
  h = min(l-j, t-l);
  FSWAPS(&b[i], &b[t-h+1], h);
  if ((h=i-k) > 1) qsort_r(b, h, z, thunk, cmp);
#if 0
  if ((h=l-j) > 1) qsort_r(&b[t-h+1], h, z, thunk, cmp);
#else
  if ((h=l-j) > 1) {
    b = &b[t-h+1];
    n = h;
    goto loop;
  }
#endif
}
#elif 0
/* Two-way partitioning, Insertion, Median pivot, Swap count and Same values */
void qsort_r(void *base, size_t n, size_t z,
	     void *thunk, int (*cmp)(void *, const void *, const void *))
{
  int i, j, k, l, m, t;
  struct temp_t { char t[z]; } /*tmp,*/ *b = (struct temp_t *)base;
  int c, h;

 loop:
  if (n < 7) {
#if !0
    for (i=1; i<(int)n; i++)
      for (j=i; j>0 && cmp(thunk, &b[j-1], &b[j]) > 0; j--)
	FSWAP(&b[j-1], &b[j]);
#else /* not rapid */
    for (i=1; i<(int)n; i++) {
      tmp = b[i];
      for (j=i; j>0 && cmp(thunk, &b[j-1], &tmp) > 0; j--)
	b[j] = b[j-1];
      b[j] = tmp;
    }
#endif
    return;
  }
  m = n/2;
  if (n > 7) {
    int l, u;
    l = 0;
    u = n-1;
    if (n > 40) {
      int s = n/4, d = n/8;
      l = MED3(&b[l], &b[l+d], &b[l+s], thunk, cmp) - b;
      m = MED3(&b[m-d], &b[m], &b[m+d], thunk, cmp) - b;
      u = MED3(&b[u-s], &b[u-d], &b[u], thunk, cmp) - b;
    }
    m = MED3(&b[l], &b[m], &b[u], thunk, cmp) - b;
  }
  FSWAP(&b[0], &b[m]);
  for (c=0, i=k=1, j=l=n-1; ; i++, j--) {
    int v;
    while (i <= j && (v=cmp(thunk, &b[i], b)) <= 0) {
      if (v == 0) {
	FSWAP(&b[k], &b[i]);
	c++;
	k++;
      }
      i++;
    }
    while (i <= j && (v=cmp(thunk, &b[j], b)) >= 0) {
      if (v == 0) {
	FSWAP(&b[j], &b[l]);
	c++;
	l--;
      }
      j--;
    }
    if (i > j) break;
    FSWAP(&b[i], &b[j]);
    c++;
  }
  if (c == 0) {
#if !0
    for (i=1; i<(int)n; i++)
      for (j=i; j>0 && cmp(thunk, &b[j-1], &b[j]) > 0; j--)
	FSWAP(&b[j-1], &b[j]);
#else /* not rapid */
    for (i=1; i<(int)n; i++) {
      tmp = b[i];
      for (j=i; j>0 && cmp(thunk, &b[j-1], &tmp) > 0; j--)
	b[j] = b[j-1];
      b[j] = tmp;
    }
#endif
    return;
  }
  t = n-1;
  h = min(k, i-k);
  FSWAPS(&b[0], &b[i-h], h);
  h = min(l-j, t-l);
  FSWAPS(&b[i], &b[t-h+1], h);
  if ((h=i-k) > 1) qsort_r(b, h, z, thunk, cmp);
#if 0
  if ((h=l-j) > 1) qsort_r(&b[t-h+1], h, z, thunk, cmp);
#else
  if ((h=l-j) > 1) {
    b = &b[t-h+1];
    n = h;
    goto loop;
  }
#endif
}
#elif 0
/* Two-way partitioning, Insertion, Median pivot, Swap count, Same values and Recover */
void qsort_r(void *base, size_t n, size_t z,
	     void *thunk, int (*cmp)(void *, const void *, const void *))
{
  int i, j, k, l, m, t;
  struct temp_t { char t[z]; } /*tmp,*/ *b = (struct temp_t *)base;
  int c, h;

 loop:
  if (n < 7) {
#if !0
    for (i=1; i<(int)n; i++)
      for (j=i; j>0 && cmp(thunk, &b[j-1], &b[j]) > 0; j--)
	FSWAP(&b[j-1], &b[j]);
#else /* not rapid */
    for (i=1; i<(int)n; i++) {
      tmp = b[i];
      for (j=i; j>0 && cmp(thunk, &b[j-1], &tmp) > 0; j--)
	b[j] = b[j-1];
      b[j] = tmp;
    }
#endif
    return;
  }
  m = n/2;
  if (n > 7) {
    int l, u;
    l = 0;
    u = n-1;
    if (n > 40) {
      int s = n/4, d = n/8;
      l = MED3(&b[l], &b[l+d], &b[l+s], thunk, cmp) - b;
      m = MED3(&b[m-d], &b[m], &b[m+d], thunk, cmp) - b;
      u = MED3(&b[u-s], &b[u-d], &b[u], thunk, cmp) - b;
    }
    m = MED3(&b[l], &b[m], &b[u], thunk, cmp) - b;
  }
  FSWAP(&b[0], &b[m]);
  for (c=0, i=k=1, j=l=n-1; ; i++, j--) {
    int v;
    while (i <= j && (v=cmp(thunk, &b[i], b)) <= 0) {
      if (v == 0) {
	FSWAP(&b[k], &b[i]);
	c++;
	k++;
      }
      i++;
    }
    while (i <= j && (v=cmp(thunk, &b[j], b)) >= 0) {
      if (v == 0) {
	FSWAP(&b[j], &b[l]);
	c++;
	l--;
      }
      j--;
    }
    if (i > j) break;
    FSWAP(&b[i], &b[j]);
    c++;
  }
  t = n-1;
  h = min(k, i-k);
  FSWAPS(&b[0], &b[i-h], h);
  h = min(l-j, t-l);
  FSWAPS(&b[i], &b[t-h+1], h);
  if (c == 0) {
    int i, j, r = n/4 + 1;

    for (i=1; i<(int)n; i++)
      for (j=i; j>0 && cmp(thunk, &b[j-1], &b[j]) > 0; j--) {
	FSWAP(&b[j-1], &b[j]);
	if (++c > r) goto recover;
      }
    return;
  }
 recover:
  if ((h=i-k) > 1) qsort_r(b, h, z, thunk, cmp);
#if 0
  if ((h=l-j) > 1) qsort_r(&b[t-h+1], h, z, thunk, cmp);
#else
  if ((h=l-j) > 1) {
    b = &b[t-h+1];
    n = h;
    goto loop;
  }
#endif
}
#elif 0
/* Pointer version */
/* Two-way partitioning */
void qsort_r(void *base, size_t n, size_t z,
	     void *thunk, int (*cmp)(void *, const void *, const void *))
{
  int i = 0, j = n, u = n-1;
  struct temp_t { char t[z]; } *b = (struct temp_t *)base, *p = b, *q = &b[j];

  for (;;) {
    do i++; while (i < (int)n && cmp(thunk, (p=&b[i]), b) < 0);
    do j--; while (cmp(thunk, (q=&b[j]), b) > 0);
    if (i > j) break;
    FSWAP(p, q);
  }
  FSWAP(b, q);
  if (0 < j-1) qsort_r(b, j, z, thunk, cmp);
  if (j+1 < u) qsort_r(q+1, u-j, z, thunk, cmp);
}
#elif 0
/* Two-way partitioning and Insertion */
void qsort_r(void *base, size_t n, size_t z,
	     void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base, *p = b, *q = b+n, *t = q-1;

  if (n < 7) {
    for (p=b+1; p<b+n; p++)
      for (q=p; q>b && cmp(thunk, q-1, q) > 0; q--)
	FSWAP((q-1), q);
    return;
  }
  for (;;) {
    do p++; while (p <= t && cmp(thunk, p, b) < 0);
    do q--; while (cmp(thunk, q, b) > 0);
    if (p > q) break;
    FSWAP(p, q);
  }
  FSWAP(b, q);
  if (b < q-1) qsort_r(b, q-b, z, thunk, cmp);
  q++;
  if (q < t) qsort_r(q, t-q+1, z, thunk, cmp);
}
#elif 0
/* Two-way partitioning, Insertion and Median pivot */
void qsort_r(void *base, size_t n, size_t z,
	     void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base, *p = b, *q = b+n, *t = q-1, *m;

  if (n < 7) {
    for (p=b+1; p<b+n; p++)
      for (q=p; q>b && cmp(thunk, q-1, q) > 0; q--)
	FSWAP((q-1), q);
    return;
  }
  m = b + n/2;
  if (n > 7) {
    struct temp_t *l, *u;
    l = b;
    u = b + (n - 1);
    if (n > 40) {
      int s = n/4, d = n/8;
      l = MED3(l, l+d, l+s, thunk, cmp);
      m = MED3(m-d, m, m+d, thunk, cmp);
      u = MED3(u-s, u-d, u, thunk, cmp);
    }
    m = MED3(l, m, u, thunk, cmp);
  }
  FSWAP(b, m);
  for (;;) {
    do p++; while (p <= t && cmp(thunk, p, b) < 0);
    do q--; while (cmp(thunk, q, b) > 0);
    if (p > q) break;
    FSWAP(p, q);
  }
  FSWAP(b, q);
  if (b < q-1) qsort_r(b, q-b, z, thunk, cmp);
  q++;
  if (q < t) qsort_r(q, t-q+1, z, thunk, cmp);
}
#elif 0
/* Two-way partitioning, Insertion, Median pivot and Swap count */
void qsort_r(void *base, size_t n, size_t z,
	     void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base, *p = b, *q = b+n, *t = q-1, *m;
  int c = 0;

  if (n < 7) {
    for (p=b+1; p<b+n; p++)
      for (q=p; q>b && cmp(thunk, q-1, q) > 0; q--)
	FSWAP((q-1), q);
    return;
  }
  m = b + n/2;
  if (n > 7) {
    struct temp_t *l, *u;
    l = b;
    u = b + (n - 1);
    if (n > 40) {
      int s = n/4, d = n/8;
      l = MED3(l, l+d, l+s, thunk, cmp);
      m = MED3(m-d, m, m+d, thunk, cmp);
      u = MED3(u-s, u-d, u, thunk, cmp);
    }
    m = MED3(l, m, u, thunk, cmp);
  }
  FSWAP(b, m);
  for (;;) {
    do p++; while (p <= t && cmp(thunk, p, b) < 0);
    do q--; while (cmp(thunk, q, b) > 0);
    if (p > q) break;
    FSWAP(p, q);
    c++;
  }
  if (c == 0) {
    for (p=b+1; p<b+n; p++)
      for (q=p; q>b && cmp(thunk, q-1, q) > 0; q--)
	FSWAP((q-1), q);
    return;
  }
  FSWAP(b, q);
  if (b < q-1) qsort_r(b, q-b, z, thunk, cmp);
  q++;
  if (q < t) qsort_r(q, t-q+1, z, thunk, cmp);
}
#elif 0
/* Two-way partitioning, Insertion, Median pivot, Swap count and Same values */
void qsort_r(void *base, size_t n, size_t z,
	     void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base, *p, *q, *r, *s, *t, *m;
  int h, c;

 loop:
  if (n < 7) {
    for (p=b+1; p<b+n; p++)
      for (q=p; q>b && cmp(thunk, q-1, q) > 0; q--)
	FSWAP((q-1), q);
    return;
  }
  m = b + n/2;
  if (n > 7) {
    struct temp_t *l, *u;
    l = b;
    u = b + (n-1);
    if (n > 40) {
      int s = n/4, d = n/8;
      l = MED3(l, l+d, l+s, thunk, cmp);
      m = MED3(m-d, m, m+d, thunk, cmp);
      u = MED3(u-s, u-d, u, thunk, cmp);
    }
    m = MED3(l, m, u, thunk, cmp);
  }
  FSWAP(b, m);
  for (c=0, p=r=b+1, q=s=b+(n-1); ; p++, q--) {
    int v;
    while (p <= q && (v=cmp(thunk, p, b)) <= 0) {
      if (v == 0) {
	FSWAP(r, p);
	c++;
	r++;
      }
      p++;
    }
    while (p <= q && (v=cmp(thunk, q, b)) >= 0) {
      if (v == 0) {
	FSWAP(q, s);
	c++;
	s--;
      }
      q--;
    }
    if (p > q) break;
    FSWAP(p, q);
    c++;
  }
  if (c == 0) {
    for (p=b+1; p<b+n; p++)
      for (q=p; q>b && cmp(thunk, q-1, q) > 0; q--)
	FSWAP((q-1), q);
    return;
  }
  t = b+n-1;
  h = min(r-b, p-r);
  FSWAPS(b, p-h, h);
  h = min(s-q, t-s);
  FSWAPS(p, t-h+1, h);
  if ((h=p-r) > 1) qsort_r(b, h, z, thunk, cmp);
#if 0
  if ((h=s-q) > 1) qsort_r(t-h+1, h, z, thunk, cmp);
#else
  if ((h=s-q) > 1) {
    b = t-h+1;
    n = h;
    goto loop;
  }
#endif
}
#elif 0
/* Two-way partitioning, Insertion, Median pivot, Swap count and Same values */
void qsort_r(void *base, size_t n, size_t z,
	     void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base, *p, *q, *r, *s, *t, *m;
  int h, c;

 loop:
  if (n < 7) {
#if !0
    for (p=b+1; p<b+n; p++)
      for (q=p; q>b && cmp(thunk, q-1, q) > 0; q--)
	FSWAP((q-1), q);
#else /* not rapid */
    for (p=b+1; p<b+n; p++) {
      tmp = *p;
      for (q=p; q>b && cmp(thunk, q-1, &tmp) > 0; q--)
	*q = *(q-1);
      *q = tmp;
    }
#endif
    return;
  }
  m = b + n/2;
  if (n > 7) {
    struct temp_t *l, *u;
    l = b;
    u = b + (n-1);
    if (n > 40) {
      int s = n/4, d = n/8;
      l = MED3(l, l+d, l+s, thunk, cmp);
      m = MED3(m-d, m, m+d, thunk, cmp);
      u = MED3(u-s, u-d, u, thunk, cmp);
    }
    m = MED3(l, m, u, thunk, cmp);
  }
  FSWAP(b, m);
  for (c=0, p=r=b+1, q=s=b+(n-1); ; p++, q--) {
    int v;
    while (p <= q && (v=cmp(thunk, p, b)) <= 0) {
      if (v == 0) {
	FSWAP(r, p);
	c++;
	r++;
      }
      p++;
    }
    while (p <= q && (v=cmp(thunk, q, b)) >= 0) {
      if (v == 0) {
	FSWAP(q, s);
	c++;
	s--;
      }
      q--;
    }
    if (p > q) break;
    FSWAP(p, q);
    c++;
  }
  if (c == 0) {
#if !0
    for (p=b+1; p<b+n; p++)
      for (q=p; q>b && cmp(thunk, q-1, q) > 0; q--)
	FSWAP((q-1), q);
#else /* not rapid */
    for (p=b+1; p<b+n; p++) {
      tmp = *p;
      for (q=p; q>b && cmp(thunk, q-1, &tmp) > 0; q--)
	*q = *(q-1);
      *q = tmp;
    }
#endif
    return;
  }
  t = b+n-1;
  h = min(r-b, p-r);
  FSWAPS(b, p-h, h);
  h = min(s-q, t-s);
  FSWAPS(p, t-h+1, h);
  if ((h=p-r) > 1) qsort_r(b, h, z, thunk, cmp);
#if 0
  if ((h=s-q) > 1) qsort_r(t-h+1, h, z, thunk, cmp);
#else
  if ((h=s-q) > 1) {
    b = t-h+1;
    n = h;
    goto loop;
  }
#endif
}
#elif 0
/* Two-way partitioning, Insertion, Median pivot, Swap count, Same values and Recover */
void qsort_r(void *base, size_t n, size_t z,
	     void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base, *p, *q, *r, *s, *t, *m;
  int h, c;

 loop:
  if (n < 7) {
#if !0
    for (p=b+1; p<b+n; p++)
      for (q=p; q>b && cmp(thunk, q-1, q) > 0; q--)
	FSWAP((q-1), q);
#else /* not rapid */
    for (p=b+1; p<b+n; p++) {
      tmp = *p;
      for (q=p; q>b && cmp(thunk, q-1, &tmp) > 0; q--)
	*q = *(q-1);
      *q = tmp;
    }
#endif
    return;
  }
  m = b + n/2;
  if (n > 7) {
    struct temp_t *l, *u;
    l = b;
    u = b + (n-1);
    if (n > 40) {
      int s = n/4, d = n/8;
      l = MED3(l, l+d, l+s, thunk, cmp);
      m = MED3(m-d, m, m+d, thunk, cmp);
      u = MED3(u-s, u-d, u, thunk, cmp);
    }
    m = MED3(l, m, u, thunk, cmp);
  }
  FSWAP(b, m);
  for (c=0, p=r=b+1, q=s=b+(n-1); ; p++, q--) {
    int v;
    while (p <= q && (v=cmp(thunk, p, b)) <= 0) {
      if (v == 0) {
	FSWAP(r, p);
	c++;
	r++;
      }
      p++;
    }
    while (p <= q && (v=cmp(thunk, q, b)) >= 0) {
      if (v == 0) {
	FSWAP(q, s);
	c++;
	s--;
      }
      q--;
    }
    if (p > q) break;
    FSWAP(p, q);
    c++;
  }
  t = b+n-1;
  h = min(r-b, p-r);
  FSWAPS(b, p-h, h);
  h = min(s-q, t-s);
  FSWAPS(p, t-h+1, h);
  if (c == 0) {
    struct temp_t *p, *q;
    int r = n/4 + 1;

    for (p=b+1; p<b+n; p++)
      for (q=p; q>b && cmp(thunk, q-1, q) > 0; q--) {
	FSWAP((q-1), q);
	if (++c > r) goto recover;
      }
    return;
  }
 recover:
  if ((h=p-r) > 1) qsort_r(b, h, z, thunk, cmp);
#if 0
  if ((h=s-q) > 1) qsort_r(t-h+1, h, z, thunk, cmp);
#else
  if ((h=s-q) > 1) {
    b = t-h+1;
    n = h;
    goto loop;
  }
#endif
}
#elif 0
/* Two-way partitioning, Insertion, Median pivot, Swap count and Same values */
#define QUICKSORT(T) do {\
  T *b = (T *)base, *p, *q, *r, *s, *t, *m;\
  int h, c;\
\
 loop_##T:\
  if (n < 7) {\
    INSERTSORT(T, base);\
    return;\
  }\
  m = b + n/2;\
  if (n > 7) {\
    T *l, *u;\
    l = b;\
    u = b + (n-1);\
    if (n > 40) {\
      int s = n/4, d = n/8;\
      l = MED3(l, l+d, l+s, thunk, cmp);\
      m = MED3(m-d, m, m+d, thunk, cmp);\
      u = MED3(u-s, u-d, u, thunk, cmp);\
    }\
    m = MED3(l, m, u, thunk, cmp);\
  }\
  TSWAP(T, b, m);\
  for (c=0, p=r=b+1, q=s=b+(n-1); ; p++, q--) {\
    int v;\
    while (p <= q && (v=cmp(thunk, p, b)) <= 0) {\
      if (v == 0) {\
	TSWAP(T, r, p);\
	c++;\
	r++;\
      }\
      p++;\
    }\
    while (p <= q && (v=cmp(thunk, q, b)) >= 0) {\
      if (v == 0) {\
	TSWAP(T, q, s);\
	c++;\
	s--;\
      }\
      q--;\
    }\
    if (p > q) break;\
    TSWAP(T, p, q);\
    c++;\
  }\
  if (c == 0) {\
    INSERTSORT(T, base);\
    return;\
  }\
  t = b+n-1;\
  h = min(r-b, p-r);\
  TSWAPS(T, b, p-h, h);\
  h = min(s-q, t-s);\
  TSWAPS(T, p, t-h+1, h);\
  if ((h=p-r) > 1) qsort_r(b, h, z, thunk, cmp);\
  /*\
  if ((h=s-q) > 1) qsort_r(t-h+1, h, z, thunk, cmp);\
  */\
  /**/\
  if ((h=s-q) > 1) {\
    base = b = t-h+1;\
    n = h;\
    goto loop_##T;\
  }\
  /**/\
} while (0)
void qsort_r(void *base, size_t n, size_t z,
	     void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (z == sizeof(int8_t))
    QUICKSORT(int8_t);
  else if (z == sizeof(int16_t))
    QUICKSORT(int16_t);
  else if (z == sizeof(int32_t))
    QUICKSORT(int32_t);
  else if (z == sizeof(int64_t))
    QUICKSORT(int64_t);
  else {
    typedef struct { char t[z]; } temp_t;
    QUICKSORT(temp_t);
  }
}
#else
/* Two-way partitioning, Insertion, Median pivot, Swap count, Same values and Recover */
#define QUICKSORT(T) do {\
  T *b = (T *)base, *p, *q, *r, *s, *t, *m;\
  int h, c;\
\
 loop_##T:\
  if (n < 7) {\
    INSERTSORT(T, base);\
    return;\
  }\
  m = b + n/2;\
  if (n > 7) {\
    T *l, *u;\
    l = b;\
    u = b + (n-1);\
    if (n > 40) {\
      int s = n/4, d = n/8;\
      l = MED3(l, l+d, l+s, thunk, cmp);\
      m = MED3(m-d, m, m+d, thunk, cmp);\
      u = MED3(u-s, u-d, u, thunk, cmp);\
    }\
    m = MED3(l, m, u, thunk, cmp);\
  }\
  TSWAP(T, b, m);\
  for (c=0, p=r=b+1, q=s=b+(n-1); ; p++, q--) {\
    int v;\
    while (p <= q && (v=cmp(thunk, p, b)) <= 0) {\
      if (v == 0) {\
	TSWAP(T, r, p);\
	c++;\
	r++;\
      }\
      p++;\
    }\
    while (p <= q && (v=cmp(thunk, q, b)) >= 0) {\
      if (v == 0) {\
	TSWAP(T, q, s);\
	c++;\
	s--;\
      }\
      q--;\
    }\
    if (p > q) break;\
    TSWAP(T, p, q);\
    c++;\
  }\
  t = b+n-1;\
  h = min(r-b, p-r);\
  TSWAPS(T, b, p-h, h);\
  h = min(s-q, t-s);\
  TSWAPS(T, p, t-h+1, h);\
  if (c == 0) {\
    T *b = (T *)base, *p, *q;\
    int r = n/4 + 1;\
\
    for (p=b+1; p<b+n; p++)\
      for (q=p; q>b && cmp(thunk, q-1, q) > 0; q--) {\
        TSWAP(T, (q-1), q);\
        if (++c > r) goto recover_##T;\
      }\
    return;\
  }\
 recover_##T:\
  if ((h=p-r) > 1) qsort_r(b, h, z, thunk, cmp);\
  /*\
  if ((h=s-q) > 1) qsort_r(t-h+1, h, z, thunk, cmp);\
  */\
  /**/\
  if ((h=s-q) > 1) {\
    base = b = t-h+1;\
    n = h;\
    goto loop_##T;\
  }\
  /**/\
} while (0)
void qsort_r(void *base, size_t n, size_t z,
	     void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (z == sizeof(int8_t))
    QUICKSORT(int8_t);
  else if (z == sizeof(int16_t))
    QUICKSORT(int16_t);
  else if (z == sizeof(int32_t))
    QUICKSORT(int32_t);
  else if (z == sizeof(int64_t))
    QUICKSORT(int64_t);
  else {
    typedef struct { char t[z]; } temp_t;
    QUICKSORT(temp_t);
  }
}
#endif

/* Intro sort variations */

#if 0
/* Quick sort(Two-way partitioning, Insertion and Median pivot) and Heap sort */
static
void _introsort_r(void *base, size_t n, size_t z, size_t depth,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base, *p = b, *q = b+n, *t = q-1, *m;

  if (n < 7) {
    insertsort_r(b, n, z, thunk, cmp);
    return;
  }
  if (depth == 0) {
    heapsort_r(b, n, z, thunk, cmp);
    return;
  }
  m = b + n/2;
  if (n > 7) {
    struct temp_t *l, *u;
    l = b;
    u = b + (n - 1);
    if (n > 40) {
      int s = n/4, d = n/8;
      l = MED3(l, l+d, l+s, thunk, cmp);
      m = MED3(m-d, m, m+d, thunk, cmp);
      u = MED3(u-s, u-d, u, thunk, cmp);
    }
    m = MED3(l, m, u, thunk, cmp);
  }
  FSWAP(b, m);
  while (!0) {
    do p++; while (p <= t && cmp(thunk, p, b) < 0);
    do q--; while (cmp(thunk, q, b) > 0);
    if (p > q) break;
    FSWAP(p, q);
  }
  FSWAP(b, q);
  if (b < q-1) _introsort_r(b, q-b, z, depth-1, thunk, cmp);
  q++;
  if (q < t) _introsort_r(q, t-q+1, z, depth-1, thunk, cmp);
}
#elif 0
/* Quick sort(Two-way partitioning, Insertion, Median pivot, Swap count and Same values) and Heap sort */
static
void _introsort_r(void *base, size_t n, size_t z, size_t depth,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base, *p, *q, *r, *s, *t, *m;
  int h, c;

 loop:
  if (n < 7) {
    insertsort_r(b, n, z, thunk, cmp);
    return;
  }
  if (depth == 0) {
    heapsort_r(b, n, z, thunk, cmp);
    return;
  }
  m = b + n/2;
  if (n > 7) {
    struct temp_t *l, *u;
    l = b;
    u = b + (n-1);
    if (n > 40) {
      int s = n/4, d = n/8;
      l = MED3(l, l+d, l+s, thunk, cmp);
      m = MED3(m-d, m, m+d, thunk, cmp);
      u = MED3(u-s, u-d, u, thunk, cmp);
    }
    m = MED3(l, m, u, thunk, cmp);
  }
  FSWAP(b, m);
  for (c=0, p=r=b+1, q=s=b+(n-1); ; p++, q--) {
    int v;
    while (p <= q && (v=cmp(thunk, p, b)) <= 0) {
      if (v == 0) {
	FSWAP(r, p);
	c++;
	r++;
      }
      p++;
    }
    while (p <= q && (v=cmp(thunk, q, b)) >= 0) {
      if (v == 0) {
	FSWAP(q, s);
	c++;
	s--;
      }
      q--;
    }
    if (p > q) break;
    FSWAP(p, q);
    c++;
  }
  if (c == 0) {
    insertsort_r(b, n, z, thunk, cmp);
    return;
  }
  t = b+n-1;
  h = min(r-b, p-r);
  FSWAPS(b, p-h, h);
  h = min(s-q, t-s);
  FSWAPS(p, t-h+1, h);
  if ((h=p-r) > 1) _introsort_r(b, h, z, depth-1, thunk, cmp);
#if 0
  if ((h=s-q) > 1) _introsort_r(t-h+1, h, z, depth-1, thunk, cmp);
#else
  if ((h=s-q) > 1) {
    b = t-h+1;
    n = h;
    depth--;
    goto loop;
  }
#endif
}
#elif 0
/* Quick sort(Two-way partitioning, Insertion, Median pivot, Swap count and Same values) and Heap sort */
#define INTROSORT(T) do {\
  T *b = (T *)base, *p, *q, *r, *s, *t, *m;\
  int h, c;\
\
 loop_##T:\
  if (n < 7) {\
    INSERTSORT(T, base);\
    return;\
  }\
  if (depth == 0) {\
    HEAPSORT(T);\
    return;\
  }\
  m = b + n/2;\
  if (n > 7) {\
    T *l, *u;\
    l = b;\
    u = b + (n-1);\
    if (n > 40) {\
      int s = n/4, d = n/8;\
      l = MED3(l, l+d, l+s, thunk, cmp);\
      m = MED3(m-d, m, m+d, thunk, cmp);\
      u = MED3(u-s, u-d, u, thunk, cmp);\
    }\
    m = MED3(l, m, u, thunk, cmp);\
  }\
  TSWAP(T, b, m);\
  for (c=0, p=r=b+1, q=s=b+(n-1); ; p++, q--) {\
    int v;\
    while (p <= q && (v=cmp(thunk, p, b)) <= 0) {\
      if (v == 0) {\
	TSWAP(T, r, p);\
	c++;\
	r++;\
      }\
      p++;\
    }\
    while (p <= q && (v=cmp(thunk, q, b)) >= 0) {\
      if (v == 0) {\
	TSWAP(T, q, s);\
	c++;\
	s--;\
      }\
      q--;\
    }\
    if (p > q) break;\
    TSWAP(T, p, q);\
    c++;\
  }\
  if (c == 0) {\
    INSERTSORT(T, base);\
    return;\
  }\
  t = b+n-1;\
  h = min(r-b, p-r);\
  TSWAPS(T, b, p-h, h);\
  h = min(s-q, t-s);\
  TSWAPS(T, p, t-h+1, h);\
  if ((h=p-r) > 1) _introsort_r(b, h, z, depth-1, thunk, cmp);\
  /*\
  if ((h=s-q) > 1) _introsort_r(t-h+1, h, z, depth-1, thunk, cmp);\
  */\
  /**/\
  if ((h=s-q) > 1) {\
    base = b = t-h+1;\
    n = h;\
    depth--;\
    goto loop_##T;\
  }\
  /**/\
} while (0)
static
void _introsort_r(void *base, size_t n, size_t z, size_t depth,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (z == sizeof(int8_t))
    INTROSORT(int8_t);
  else if (z == sizeof(int16_t))
    INTROSORT(int16_t);
  else if (z == sizeof(int32_t))
    INTROSORT(int32_t);
  else if (z == sizeof(int64_t))
    INTROSORT(int64_t);
  else {
    typedef struct { char t[z]; } temp_t;
    INTROSORT(temp_t);
  }
}
#else
/* Quick sort(Two-way partitioning, Insertion, Median pivot, Swap count, Same values and Recover) and Heap sort */
#define INTROSORT(T) do {\
  T *b = (T *)base, *p, *q, *r, *s, *t, *m;\
  int h, c;\
\
 loop_##T:\
  if (n < 7) {\
    INSERTSORT(T, base);\
    return;\
  }\
  if (depth == 0) {\
    HEAPSORT(T);\
    return;\
  }\
  m = b + n/2;\
  if (n > 7) {\
    T *l, *u;\
    l = b;\
    u = b + (n-1);\
    if (n > 40) {\
      int s = n/4, d = n/8;\
      l = MED3(l, l+d, l+s, thunk, cmp);\
      m = MED3(m-d, m, m+d, thunk, cmp);\
      u = MED3(u-s, u-d, u, thunk, cmp);\
    }\
    m = MED3(l, m, u, thunk, cmp);\
  }\
  TSWAP(T, b, m);\
  for (c=0, p=r=b+1, q=s=b+(n-1); ; p++, q--) {\
    int v;\
    while (p <= q && (v=cmp(thunk, p, b)) <= 0) {\
      if (v == 0) {\
	TSWAP(T, r, p);\
	c++;\
	r++;\
      }\
      p++;\
    }\
    while (p <= q && (v=cmp(thunk, q, b)) >= 0) {\
      if (v == 0) {\
	TSWAP(T, q, s);\
	c++;\
	s--;\
      }\
      q--;\
    }\
    if (p > q) break;\
    TSWAP(T, p, q);\
    c++;\
  }\
  t = b+n-1;\
  h = min(r-b, p-r);\
  TSWAPS(T, b, p-h, h);\
  h = min(s-q, t-s);\
  TSWAPS(T, p, t-h+1, h);\
  if (c == 0) {\
    T *b = (T *)base, *p, *q;\
    int r = n/4 + 1;\
\
    for (p=b+1; p<b+n; p++)\
      for (q=p; q>b && cmp(thunk, q-1, q) > 0; q--) {\
        TSWAP(T, (q-1), q);\
        if (++c > r) goto recover_##T;\
      }\
    return;\
  }\
 recover_##T:\
  if ((h=p-r) > 1) _introsort_r(b, h, z, depth-1, thunk, cmp);\
  /*\
  if ((h=s-q) > 1) _introsort_r(t-h+1, h, z, depth-1, thunk, cmp);\
  */\
  /**/\
  if ((h=s-q) > 1) {\
    base = b = t-h+1;\
    n = h;\
    depth--;\
    goto loop_##T;\
  }\
  /**/\
} while (0)
static
void _introsort_r(void *base, size_t n, size_t z, size_t depth,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  if (z == sizeof(int8_t))
    INTROSORT(int8_t);
  else if (z == sizeof(int16_t))
    INTROSORT(int16_t);
  else if (z == sizeof(int32_t))
    INTROSORT(int32_t);
  else if (z == sizeof(int64_t))
    INTROSORT(int64_t);
  else {
    typedef struct { char t[z]; } temp_t;
    INTROSORT(temp_t);
  }
}
#endif
void introsort_r(void *base, size_t n, size_t z,
		 void *thunk, int (*cmp)(void *, const void *, const void *))
{
  size_t depth;

  introsort_r_limiter(n, &depth);
  _introsort_r(base, n, z, depth, thunk, cmp);
}

inline
unsigned int number_of_processors(void)
{
  unsigned int n_cpu = 1;

#if defined(PTW32_VERSION) || defined(__hpux)
  n_cpu = pthread_num_processors_np();
#elif defined(__APPLE__) || defined(__FreeBSD__)
  int count;
  size_t size = sizeof(count);
  n_cpu = sysctlbyname("hw.ncpu", &count, &size, NULL, 0) ? n_cpu : count;
#elif defined(_SC_NPROCESSORS_ONLN)
  int const count = sysconf(_SC_NPROCESSORS_ONLN);
  n_cpu = (count > 0) ? count : n_cpu;
#else
  n_cpu = get_nprocs();
#endif
  return n_cpu;
}

/* Parallel exchange sort variations */
#if 0 /* very dull */
/* OpenMP version odd-even sort */
/*
  References:
  [1] Clay Breshears, ``The Art of Concurrency,'' ISBN 978-0-596-52153-0,
  Oreilly Media Inc, 2009.
 */
void poddevensort_r(void *base, size_t n, size_t z,
		    void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base;
  size_t i;
  int flag0 = 0, flag1 = 0, trips = 0;

  if (n < 2) return;
  while (!flag1) {
    flag0 = flag1 = !0;
#if defined(_OPENMP)
#pragma omp for
#endif
    for (i=0; i<n-1; i+=2) {
      if (cmp(thunk, &b[i], &b[i+1]) > 0) {
	FSWAP(&b[i], &b[i+1]);
	flag0 = 0;
      }
    }
    if (!flag0 || !trips)
#if defined(_OPENMP)
#pragma omp for
#endif
      for (i=1; i<n-1; i+=2) {
	if (cmp(thunk, &b[i], &b[i+1]) > 0) {
	  FSWAP(&b[i], &b[i+1]);
	  flag1 = 0;
	}
      }
    trips = !0;
  }
}
#elif 0
/* Thread version odd-even sort */
struct thr_poddevensort_r_arg {
  void *base;
  size_t n, z;
  void *thunk;
  int (*cmp)(void *, const void *, const void *);
  size_t n0, n1;
  int *flag;
};
#define PODDEVENSORT(T) do {\
  T *b = (T *)base;\
  size_t i;\
\
  for (i=a->n0; i<a->n1; i+=2)\
    if (cmp(thunk, &b[i], &b[i+1]) > 0) {\
      TSWAP(T, &b[i], &b[i+1]);\
      *a->flag = 0;\
    }\
  for (i=a->n0+1; i<a->n1; i+=2)\
    if (cmp(thunk, &b[i], &b[i+1]) > 0) {\
      TSWAP(T, &b[i], &b[i+1]);\
      *a->flag = 0;\
    }\
  return NULL;\
} while (0)
static
void *thr_poddevensort_r(void *arg)
{
  struct thr_poddevensort_r_arg *a = (struct thr_poddevensort_r_arg *)arg;
#define base a->base
#define n a->n
#define z a->z
#define thunk a->thunk
#define cmp a->cmp
  if (z == sizeof(int8_t))
    PODDEVENSORT(int8_t);
  else if (z == sizeof(int16_t))
    PODDEVENSORT(int16_t);
  else if (z == sizeof(int32_t))
    PODDEVENSORT(int32_t);
  else if (z == sizeof(int64_t))
    PODDEVENSORT(int64_t);
  else {
    struct temp_t { char t[z]; };
    PODDEVENSORT(struct temp_t);
  }
#undef base
#undef n
#undef z
#undef thunk
#undef cmp
}
void poddevensort_r(void *base, size_t n, size_t z,
		    void *thunk, int (*cmp)(void *, const void *, const void *))
{
  size_t i, k;
  int flag = 0;

  if (n < 2) return;
  while (!flag) {
    size_t m = min(n-1, number_of_processors());
    pthread_t thrs[m];
    struct thr_poddevensort_r_arg args[m];

    flag = !0;
    for (i=0, k=0; k<m; k++) {
      struct thr_poddevensort_r_arg arg = {
	base, n, z, thunk, cmp, i, i + (n-1)/m + ((k < (n-1)%m)?1:0), &flag,
      };

      args[k] = arg;
      pthread_create(&thrs[k], NULL, thr_poddevensort_r, &args[k]);
      i = arg.n1;
    }
    for (k=0; k<m; k++)
      pthread_join(thrs[k], NULL);
  }
}
#else
/* Thread pool version odd-even sort */
struct thr_poddevensort_r_arg {
  void *base;
  size_t n, z;
  void *thunk;
  int (*cmp)(void *, const void *, const void *);
  size_t n0, n1;
  int *flag;
};
#define PODDEVENSORT(T) do {\
  T *b = (T *)base;\
  size_t i;\
\
  for (i=a->n0; i<a->n1; i+=2)\
    if (cmp(thunk, &b[i], &b[i+1]) > 0) {\
      TSWAP(T, &b[i], &b[i+1]);\
      *a->flag = 0;\
    }\
  for (i=a->n0+1; i<a->n1; i+=2)\
    if (cmp(thunk, &b[i], &b[i+1]) > 0) {\
      TSWAP(T, &b[i], &b[i+1]);\
      *a->flag = 0;\
    }\
  return NULL;\
} while (0)
static
void *thr_poddevensort_r(void *arg)
{
  struct thr_poddevensort_r_arg *a = (struct thr_poddevensort_r_arg *)arg;
#define base a->base
#define n a->n
#define z a->z
#define thunk a->thunk
#define cmp a->cmp
  if (z == sizeof(int8_t))
    PODDEVENSORT(int8_t);
  else if (z == sizeof(int16_t))
    PODDEVENSORT(int16_t);
  else if (z == sizeof(int32_t))
    PODDEVENSORT(int32_t);
  else if (z == sizeof(int64_t))
    PODDEVENSORT(int64_t);
  else {
    typedef struct { char t[z]; } temp_t;
    PODDEVENSORT(temp_t);
  }
#undef base
#undef n
#undef z
#undef thunk
#undef cmp
}
void poddevensort_r(void *base, size_t n, size_t z,
		    void *thunk, int (*cmp)(void *, const void *, const void *))
{
  unsigned int n_cpu = number_of_processors();
  thread_pool_t *pool;
  size_t i, k;
  int flag = 0;

  if (n < 2) return;
  pool = thread_pool_create(n_cpu, NULL);
  while (!flag) {
    size_t m = min(n-1, n_cpu);
    struct thr_poddevensort_r_arg args[m];

    flag = !0;
    for (i=0, k=0; k<m; k++) {
      struct thr_poddevensort_r_arg arg = {
	base, n, z, thunk, cmp, i, i + (n-1)/m + ((k < (n-1)%m)?1:0), &flag,
      };

      args[k] = arg;
      thread_pool_queue(pool, thr_poddevensort_r, &args[k]);
      i = arg.n1;
    }
    thread_pool_wait(pool);
  }
  thread_pool_join(pool);
}
#endif

#if 0
/* OpenMP version shell sort */
/*
  References:
  [1] Clay Breshears, ``The Art of Concurrency,'' ISBN 978-0-596-52153-0,
  Oreilly Media Inc, 2009.
 */
void pshellsort_r(void *base, size_t n, size_t z,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  struct temp_t { char t[z]; } *b = (struct temp_t *)base;
  size_t h, i, k;
  int j;

  if (n < 2) return;
  for (h=1; h<n; h=h*3+1); h /= 3;
#if defined(_OPENMP)
#pragma omp parallel firstprivate(h)
  {
#endif
    while (h != 1) {
#if defined(_OPENMP)
#pragma omp for
#endif
      for (k=0; k<h; k++)
	for (i=k; i<n; i+=h)
	  for (j=i; j>h && cmp(thunk, &b[j-h], &b[j]) > 0; j-=h)
	    FSWAP(&b[j-h], &b[j]);
      h /= 3;
    }
#if defined(_OPENMP)
  }
#endif
  insertsort_r(base, n, z, thunk, cmp);
}
#elif 0
/* Thread version shell sort */
struct thr_pshellsort_r_arg {
  void *base;
  size_t n, z;
  void *thunk;
  int (*cmp)(void *, const void *, const void *);
  size_t k0, k1, h;
};
#define PSHELLSORT(T) do {\
  T *b = (T *)base;\
  size_t i, k;\
  int j;\
\
  for (k=a->k0; k<a->k1; k++)\
    for (i=k; i<n; i+=a->h)\
      for (j=i; j>a->h && cmp(thunk, &b[j-a->h], &b[j]) > 0; j-=a->h)\
	TSWAP(T, &b[j-a->h], &b[j]);\
  return NULL;\
} while (0)
static
void *thr_pshellsort_r(void *arg)
{
  struct thr_pshellsort_r_arg *a = (struct thr_pshellsort_r_arg *)arg;
#define base a->base
#define n a->n
#define z a->z
#define thunk a->thunk
#define cmp a->cmp
  if (z == sizeof(int8_t))
    PSHELLSORT(int8_t);
  else if (z == sizeof(int16_t))
    PSHELLSORT(int16_t);
  else if (z == sizeof(int32_t))
    PSHELLSORT(int32_t);
  else if (z == sizeof(int64_t))
    PSHELLSORT(int64_t);
  else {
    struct temp_t { char t[z]; };
    PSHELLSORT(struct temp_t);
  }
#undef base
#undef n
#undef z
#undef thunk
#undef cmp
}
void pshellsort_r(void *base, size_t n, size_t z,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  size_t h, i, k;

  if (n < 2) return;
  for (h=1; h<n; h=h*3+1); h /= 3;
  {
    while (h != 1) {
      size_t m = min(h, number_of_processors());
      pthread_t thrs[m];
      struct thr_pshellsort_r_arg args[m];

      for (i=0, k=0; k<m; k++) {
	struct thr_pshellsort_r_arg arg = {
	  base, n, z, thunk, cmp, i, i + h/m + ((k < h%m)?1:0), h,
	};

	args[k] = arg;
	pthread_create(&thrs[k], NULL, thr_pshellsort_r, &args[k]);
	i = arg.k1;
      }
      for (k=0; k<m; k++)
	pthread_join(thrs[k], NULL);
      h /= 3;
    }
  }
  insertsort_r(base, n, z, thunk, cmp);
}
#else
/* Thread pool version shell sort */
struct thr_pshellsort_r_arg {
  void *base;
  size_t n, z;
  void *thunk;
  int (*cmp)(void *, const void *, const void *);
  size_t k0, k1, h;
};
#define PSHELLSORT(T) do {\
  T *b = (T *)base;\
  size_t i, k;\
  int j;\
\
  for (k=a->k0; k<a->k1; k++)\
    for (i=k; i<n; i+=a->h)\
      for (j=i; j>a->h && cmp(thunk, &b[j-a->h], &b[j]) > 0; j-=a->h)\
	TSWAP(T, &b[j-a->h], &b[j]);\
  return NULL;\
} while (0)
static
void *thr_pshellsort_r(void *arg)
{
  struct thr_pshellsort_r_arg *a = (struct thr_pshellsort_r_arg *)arg;
#define base a->base
#define n a->n
#define z a->z
#define thunk a->thunk
#define cmp a->cmp
  if (z == sizeof(int8_t))
    PSHELLSORT(int8_t);
  else if (z == sizeof(int16_t))
    PSHELLSORT(int16_t);
  else if (z == sizeof(int32_t))
    PSHELLSORT(int32_t);
  else if (z == sizeof(int64_t))
    PSHELLSORT(int64_t);
  else {
    struct temp_t { char t[z]; };
    PSHELLSORT(struct temp_t);
  }
#undef base
#undef n
#undef z
#undef thunk
#undef cmp
}
void pshellsort_r(void *base, size_t n, size_t z,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  unsigned int n_cpu = number_of_processors();
  thread_pool_t *pool;
  size_t h, i, k;

  if (n < 2) return;
  pool = thread_pool_create(n_cpu, NULL);
  for (h=1; h<n; h=h*3+1); h /= 3;
  {
    while (h != 1) {
      size_t m = min(h, n_cpu);
      struct thr_pshellsort_r_arg args[m];

      for (i=0, k=0; k<m; k++) {
	struct thr_pshellsort_r_arg arg = {
	  base, n, z, thunk, cmp, i, i + h/m + ((k < h%m)?1:0), h,
	};

	args[k] = arg;
	thread_pool_queue(pool, thr_pshellsort_r, &args[k]);
	i = arg.k1;
      }
      thread_pool_wait(pool);
      h /= 3;
    }
  }
  thread_pool_join(pool);
  insertsort_r(base, n, z, thunk, cmp);
}
#endif

/* Parallel merge sort variations */
#if !0
/* Thread version merge sort */
struct thr_pmergesort_r_arg {
  void *area, *an, *base, *bn;
  size_t z, n;
  void *thunk;
  int (*cmp)(void *, const void *, const void *);
  size_t *n_threshold;
};
#if !0
#define PMERGESORT(T) do {\
  T *a0 = (T *)area, *am = a0, *b0 = (T *)base, *bm = b0, *ai, *aj, *bk/*, *l*/;\
  int m = n>>1;\
\
  if (n < 34) {\
    INSERTSORT(T, area);\
    return NULL;\
  }\
  am += m;\
  bm += m;\
\
  struct thr_pmergesort_r_arg arg1 = {\
    a0, am, b0, bm, z, m, thunk, cmp, a->n_threshold,\
  }, arg2 = {\
    am, an, bm, bn, z, n - m, thunk, cmp, a->n_threshold,\
  };\
  pthread_t thr;\
  int flag;\
\
  if (!(flag = (m > *a->n_threshold &&\
		!pthread_create(&thr, NULL, thr_pmergesort_r, &arg1))))\
    thr_pmergesort_r(&arg1);\
  thr_pmergesort_r(&arg2);\
  if (flag)\
    pthread_join(thr, NULL);\
\
  memcpy(b0, a0, (char *)am - (char *)a0);/*for (bk = b0, l=a0; l!=am; l++, bk++) *bk = *l;*/\
  ai = a0;\
  aj = am;\
  bk = b0;\
  while (bk != bm && aj != an) {\
    if (cmp(thunk, aj, bk) < 0) {\
      *ai = *aj;\
      ++ai;\
      ++aj;\
    }\
    else {\
      *ai = *bk;\
      ++ai;\
      ++bk;\
    }\
  }\
  memcpy(ai, bk, (char *)bm - (char *)bk);/*for (l=bk; l!=bm; l++, ai++) *ai = *l;*/\
  return NULL;\
} while (0)
static
void *thr_pmergesort_r(void *arg)
{
  struct thr_pmergesort_r_arg *a = (struct thr_pmergesort_r_arg *)arg;
#define area a->area
#define an a->an
#define base a->base
#define bn a->bn
#define z a->z
#define n a->n
#define thunk a->thunk
#define cmp a->cmp
  if (z == sizeof(int8_t))
    PMERGESORT(int8_t);
  else if (z == sizeof(int16_t))
    PMERGESORT(int16_t);
  else if (z == sizeof(int32_t))
    PMERGESORT(int32_t);
  else if (z == sizeof(int64_t))
    PMERGESORT(int64_t);
  else {
    typedef struct { char t[z]; } temp_t;
    PMERGESORT(temp_t);
  }
#undef area
#undef an
#undef base
#undef bn
#undef z
#undef n
#undef thunk
#undef cmp
}
int pmergesort_r(void *base, size_t n, size_t z,
		 void *thunk, int (*cmp)(void *, const void *, const void *))
{
  unsigned int n_cpu = number_of_processors();
  if (n_cpu > 1) {
    size_t n_threshold = n/n_cpu;
    struct temp_t { char t[z]; } *a = malloc(z*n), *b = (struct temp_t *)base;
    if (!a) return -1;
    struct thr_pmergesort_r_arg arg = {
      b, &b[n], a, &a[n], z, n, thunk, cmp, &n_threshold,
    };

    thr_pmergesort_r(&arg);
    free(a);
  }
  else
    mergesort_r(base, n, z, thunk, cmp);
  return 0;
}
#else
#define PMERGESORT(T) do {\
  T *a0 = (T *)area, *am = a0, *b0 = (T *)base, *bm = b0, *ai, *aj, *bk/*, *l*/;\
  int m = n>>1;\
\
  if (n < 34) {\
    INSERTSORT(T, base);\
    return NULL;\
  }\
  am += m;\
  bm += m;\
\
  struct thr_pmergesort_r_arg arg1 = {\
    b0, bm, a0, am, z, m, thunk, cmp, a->n_threshold,\
  }, arg2 = {\
    bm, bn, am, an, z, n - m, thunk, cmp, a->n_threshold,\
  };\
  pthread_t thr;\
  int flag;\
\
  if (!(flag = (m > *a->n_threshold &&\
		!pthread_create(&thr, NULL, thr_pmergesort_r, &arg1))))\
    thr_pmergesort_r(&arg1);\
  thr_pmergesort_r(&arg2);\
  if (flag)\
    pthread_join(thr, NULL);\
\
  ai = a0;\
  aj = am;\
  bk = b0;\
  while (ai != am && aj != an) {\
    if (cmp(thunk, aj, ai) < 0) {\
      *bk = *aj;\
      ++bk;\
      ++aj;\
    }\
    else {\
      *bk = *ai;\
      ++bk;\
      ++ai;\
    }\
  }\
  memcpy(bk, ai, (char *)am - (char *)ai);/*for (l=ai; l!=am; l++, bk++) *bk = *l;*/\
  memcpy((char *)bk + ((char *)am - (char *)ai), aj, (char *)an - (char *)aj);/*for (l=aj; l!=an; l++, bk++) *bk = *l;*/\
  return NULL;\
} while (0)
static
void *thr_pmergesort_r(void *arg)
{
  struct thr_pmergesort_r_arg *a = (struct thr_pmergesort_r_arg *)arg;
#define area a->area
#define an a->an
#define base a->base
#define bn a->bn
#define z a->z
#define n a->n
#define thunk a->thunk
#define cmp a->cmp
  if (z == sizeof(int8_t))
    PMERGESORT(int8_t);
  else if (z == sizeof(int16_t))
    PMERGESORT(int16_t);
  else if (z == sizeof(int32_t))
    PMERGESORT(int32_t);
  else if (z == sizeof(int64_t))
    PMERGESORT(int64_t);
  else {
    struct temp_t { char t[z]; };
    PMERGESORT(struct temp_t);
  }
#undef area
#undef an
#undef base
#undef bn
#undef z
#undef n
#undef thunk
#undef cmp
}
int pmergesort_r(void *base, size_t n, size_t z,
		 void *thunk, int (*cmp)(void *, const void *, const void *))
{
  unsigned int n_cpu = number_of_processors();
  if (n_cpu > 1) {
    size_t n_threshold = n/n_cpu;
    struct temp_t { char t[z]; } *a = malloc(z*n), *b = (struct temp_t *)base;
    if (!a) return -1;
    struct thr_pmergesort_r_arg arg = {
      a, &a[n], b, &b[n], z, n, thunk, cmp, &n_threshold,
    };

    thr_pmergesort_r(&arg);
    free(a);
  }
  else
    mergesort_r(base, n, z, thunk, cmp);
  return 0;
}
#endif
#else
/* Thread pool version merge sort */
struct thr_pmergesort_r_arg {
  void *area, *an, *base, *bn;
  size_t z, n;
  void *thunk;
  int (*cmp)(void *, const void *, const void *);
  size_t *n_threshold;
  thread_pool_t *thread_pool;
};
#if 0
static
void *thr_pmergesort_r(void *arg)
{
  struct thr_pmergesort_r_arg *a = (struct thr_pmergesort_r_arg *)arg;
#define area a->area
#define an a->an
#define base a->base
#define bn a->bn
#define z a->z
#define n a->n
#define thunk a->thunk
#define cmp a->cmp
  struct temp_t { char t[z]; } *a0 = (struct temp_t *)area, *am = a0, *b0 = (struct temp_t *)base, *bm = b0, *ai, *aj, *bk, *l;
  int m = n>>1;

  if (n < 34) {
    insertsort_r(b0, n, z, thunk, cmp);
    return NULL;
  }
  am += m;
  bm += m;

  struct thr_pmergesort_r_arg arg1 = {
    b0, bm, a0, am, z, m, thunk, cmp, a->n_threshold, a->thread_pool,
  }, arg2 = {
    bm, bn, am, an, z, n - m, thunk, cmp, a->n_threshold, a->thread_pool,
  };
  int flag;

  if (!(flag = (m > *a->n_threshold &&
                !thread_pool_queue(a->thread_pool, thr_pmergesort_r, &arg1))))
    thr_pmergesort_r(&arg1);
  thr_pmergesort_r(&arg2);
  if (flag)
    thread_pool_wait(a->thread_pool);

  ai = a0;
  aj = am;
  bk = b0;
  while (ai != am && aj != an) {
    if (cmp(thunk, aj, ai) < 0) {
      *bk = *aj;
      ++bk;
      ++aj;
    }
    else {
      *bk = *ai;
      ++bk;
      ++ai;
    }
  }
  for (l=ai; l!=am; l++, bk++) *bk = *l;
  for (l=aj; l!=an; l++, bk++) *bk = *l;
  return NULL;
#undef area
#undef an
#undef base
#undef bn
#undef z
#undef n
#undef thunk
#undef cmp
}
#else
#define PMERGESORT(T) do {\
  T *a0 = (T *)area, *am = a0, *b0 = (T *)base, *bm = b0, *ai, *aj, *bk, *l;\
  int m = n>>1;\
\
  if (n < 34) {\
    INSERTSORT(T, base);\
    return NULL;\
  }\
  am += m;\
  bm += m;\
\
  struct thr_pmergesort_r_arg arg1 = {\
    b0, bm, a0, am, z, m, thunk, cmp, a->n_threshold, a->thread_pool,\
  }, arg2 = {\
    bm, bn, am, an, z, n - m, thunk, cmp, a->n_threshold, a->thread_pool,\
  };\
  int flag;\
\
  if (!(flag = (m > *a->n_threshold &&\
		!thread_pool_queue(a->thread_pool, thr_pmergesort_r, &arg1))))\
    thr_pmergesort_r(&arg1);\
  thr_pmergesort_r(&arg2);\
  if (flag)\
    thread_pool_wait(a->thread_pool);\
\
  ai = a0;\
  aj = am;\
  bk = b0;\
  while (ai != am && aj != an) {\
    if (cmp(thunk, aj, ai) < 0) {\
      *bk = *aj;\
      ++bk;\
      ++aj;\
    }\
    else {\
      *bk = *ai;\
      ++bk;\
      ++ai;\
    }\
  }\
  for (l=ai; l!=am; l++, bk++) *bk = *l;\
  for (l=aj; l!=an; l++, bk++) *bk = *l;\
  return NULL;\
} while (0)
static
void *thr_pmergesort_r(void *arg)
{
  struct thr_pmergesort_r_arg *a = (struct thr_pmergesort_r_arg *)arg;
#define area a->area
#define an a->an
#define base a->base
#define bn a->bn
#define z a->z
#define n a->n
#define thunk a->thunk
#define cmp a->cmp
  if (z == sizeof(int8_t))
    PMERGESORT(int8_t);
  else if (z == sizeof(int16_t))
    PMERGESORT(int16_t);
  else if (z == sizeof(int32_t))
    PMERGESORT(int32_t);
  else if (z == sizeof(int64_t))
    PMERGESORT(int64_t);
  else {
    struct temp_t { char t[z]; };
    PMERGESORT(struct temp_t);
  }
#undef area
#undef an
#undef base
#undef bn
#undef z
#undef n
#undef thunk
#undef cmp
}
#endif
int pmergesort_r(void *base, size_t n, size_t z,
		 void *thunk, int (*cmp)(void *, const void *, const void *))
{
  unsigned int n_cpu = number_of_processors();
  if (n_cpu > 1) {
    size_t n_threshold = n/n_cpu;
    struct temp_t { char t[z]; } *a = malloc(z*n), *b = (struct temp_t *)base;
    if (!a) return -1;
    thread_pool_t *pool;
    struct thr_pmergesort_r_arg arg = {
      a, &a[n], b, &b[n], z, n, thunk, cmp, &n_threshold, (pool = thread_pool_create(n_cpu, NULL))
    };

    thr_pmergesort_r(&arg);
    thread_pool_join(pool);
    free(a);
  }
  else
    mergesort_r(base, n, z, thunk, cmp);
  return 0;
}
#endif

#if !0
/* Thread version tim sort */
struct thr_ptimsort_r_arg {
  void *base;
  size_t n, z;
  void *storage;
  void *thunk;
  int (*cmp)(void *, const void *, const void *);
  size_t *n_threshold;
};
void *thr_ptimsort_r(void *arg)
{
  struct thr_ptimsort_r_arg *a = (struct thr_ptimsort_r_arg *)arg;
  if (a->n /*<= 7*/< 34) {
    insertsort_r(a->base, a->n, a->z, a->thunk, a->cmp);
    return NULL;
  }
  struct temp_t { char t[a->z]; } *b = (struct temp_t *)a->base, *s = (struct temp_t *)a->storage;
  size_t m = a->n>>1;
  struct thr_ptimsort_r_arg arg1 = {
    b, m, a->z, s, a->thunk, a->cmp, a->n_threshold,
  }, arg2 = {
    b + m, a->n - m, a->z, s + m, a->thunk, a->cmp, a->n_threshold,
  };
  pthread_t thr;
  int flag;

  if (!(flag = (m > *a->n_threshold &&
		!pthread_create(&thr, NULL, thr_ptimsort_r, &arg1))))
    thr_ptimsort_r(&arg1);
  thr_ptimsort_r(&arg2);
  if (flag)
    pthread_join(thr, NULL);
  timmerge_r(b, b, m, b + m, a->n - m, a->z, s, a->thunk, a->cmp);
  return NULL;
}
int ptimsort_r(void *base, size_t n, size_t z,
	       void *thunk, int (*cmp)(void *, const void *, const void *))
{
  unsigned int n_cpu = number_of_processors();
  if (n_cpu > 1) {
    size_t n_threshold = n/n_cpu;
    void *storage = malloc(z*n);
    if (!storage) return -1;
    struct thr_ptimsort_r_arg arg = {
      base, n, z, storage, thunk, cmp, &n_threshold,
    };

    thr_ptimsort_r(&arg);
    free(storage);
  }
  else
    timsort_r(base, n, z, thunk, cmp);
  return 0;
}
#else
/* Thread pool version tim sort */
struct thr_ptimsort_r_arg {
  void *base;
  size_t n, z;
  void *storage;
  void *thunk;
  int (*cmp)(void *, const void *, const void *);
  size_t *n_threshold;
  thread_pool_t *thread_pool;
};
void *thr_ptimsort_r(void *arg)
{
  struct thr_ptimsort_r_arg *a = (struct thr_ptimsort_r_arg *)arg;
  if (a->n /*<= 7*/< 34) {
    insertsort_r(a->base, a->n, a->z, a->thunk, a->cmp);
    return NULL;
  }
  struct temp_t { char t[a->z]; } *b = (struct temp_t *)a->base, *s = (struct temp_t *)a->storage;
  size_t m = a->n>>1;
  struct thr_ptimsort_r_arg arg1 = {
    b, m, a->z, s, a->thunk, a->cmp, a->n_threshold, a->thread_pool,
  }, arg2 = {
    b + m, a->n - m, a->z, s + m, a->thunk, a->cmp, a->n_threshold, a->thread_pool,
  };
  int flag;

  if (!(flag = (m > *a->n_threshold &&
		!thread_pool_queue(a->thread_pool, thr_ptimsort_r, &arg1))))
    thr_ptimsort_r(&arg1);
  thr_ptimsort_r(&arg2);
  if (flag)
    thread_pool_wait(a->thread_pool);
  timmerge_r(b, b, m, b + m, a->n - m, a->z, s, a->thunk, a->cmp);
  return NULL;
}
int ptimsort_r(void *base, size_t n, size_t z,
	       void *thunk, int (*cmp)(void *, const void *, const void *))
{
  unsigned int n_cpu = number_of_processors();
  if (n_cpu > 1) {
    size_t n_threshold = n/n_cpu;
    void *storage = malloc(z*n);
    if (!storage) return -1;
    thread_pool_t *pool;
    struct thr_ptimsort_r_arg arg = {
      base, n, z, storage, thunk, cmp, &n_threshold, (pool = thread_pool_create(n_cpu, NULL)),
    };

    thr_ptimsort_r(&arg);
    thread_pool_join(pool);
    free(storage);
  }
  else
    timsort_r(base, n, z, thunk, cmp);
  return 0;
}
#endif

/* Parallel intro sort variations */
#if 0 /* very dull */
/* Thread version intro sort */
struct thr_pintrosort_r_arg {
  void *base;
  size_t n, z, depth;
  void *thunk;
  int (*cmp)(void *, const void *, const void *);
  size_t *n_thread, m_thread;
  pthread_t *thrs;
  pthread_mutex_t *mutex;
  struct thr_pintrosort_r_arg *args;
};
#define PINTROSORT(T) do {\
  T *b = (T *)base, *p, *q, *r, *s, *t, *m;\
  int h, c;\
\
 loop_##T:\
  if (n < 7) {\
    INSERTSORT(T, base);\
    return NULL;\
  }\
  if (depth == 0) {\
    HEAPSORT(T);\
    return NULL;\
  }\
  m = b + n/2;\
  if (n > 7) {\
    T *l, *u;\
    l = b;\
    u = b + (n-1);\
    if (n > 40) {\
      int s = n/4, d = n/8;\
      l = MED3(l, l+d, l+s, thunk, cmp);\
      m = MED3(m-d, m, m+d, thunk, cmp);\
      u = MED3(u-s, u-d, u, thunk, cmp);\
    }\
    m = MED3(l, m, u, thunk, cmp);\
  }\
  TSWAP(T, b, m);\
  for (c=0, p=r=b+1, q=s=b+(n-1); ; p++, q--) {\
    int v;\
    while (p <= q && (v=cmp(thunk, p, b)) <= 0) {\
      if (v == 0) {\
	TSWAP(T, r, p);\
	c++;\
	r++;\
      }\
      p++;\
    }\
    while (p <= q && (v=cmp(thunk, q, b)) >= 0) {\
      if (v == 0) {\
	TSWAP(T, q, s);\
	c++;\
	s--;\
      }\
      q--;\
    }\
    if (p > q) break;\
    TSWAP(T, p, q);\
    c++;\
  }\
  t = b+n-1;\
  h = min(r-b, p-r);\
  TSWAPS(T, b, p-h, h);\
  h = min(s-q, t-s);\
  TSWAPS(T, p, t-h+1, h);\
  if (c == 0) {\
    T *b = (T *)base, *p, *q;\
    int r = n/4 + 1;\
\
    for (p=b+1; p<b+n; p++)\
      for (q=p; q>b && cmp(thunk, q-1, q) > 0; q--) {\
        TSWAP(T, (q-1), q);\
        if (++c > r) goto recover_##T;\
      }\
    return NULL;\
  }\
 recover_##T:\
  if ((h=p-r) > 1) {\
    int flag;\
    struct thr_pintrosort_r_arg arg = {\
      b, h, z, depth-1, thunk, cmp, a->n_thread, a->m_thread, a->thrs, a->mutex, a->args,\
    };\
\
    pthread_mutex_lock(a->mutex);\
    a->args[*a->n_thread] = arg;\
    if ((flag = (*a->n_thread < a->m_thread && !pthread_create(&a->thrs[*a->n_thread], NULL, thr_pintrosort_r, &a->args[*a->n_thread]))))\
      (*a->n_thread)++;\
    pthread_mutex_unlock(a->mutex);\
    if (!flag) thr_pintrosort_r(&arg);\
  }\
  if ((h=s-q) > 1) {\
    /*\
    int flag;\
    struct thr_pintrosort_r_arg arg = {\
      t-h+1, h, z, depth-1, thunk, cmp, a->n_thread, a->m_thread, a->thrs, a->mutex, a->args,\
    };\
\
    pthread_mutex_lock(a->mutex);\
    a->args[*a->n_thread] = arg;\
    if ((flag = (*a->n_thread < a->m_thread && !pthread_create(&a->thrs[*a->n_thread], NULL, thr_pintrosort_r, &a->args[*a->n_thread]))))\
      (*a->n_thread)++;\
    pthread_mutex_unlock(a->mutex);\
    if (!flag) thr_pintrosort_r(&arg);\
    */\
    /**/\
    base = b = t-h+1;\
    n = h;\
    depth--;\
    goto loop_##T;\
    /**/\
  }\
  return NULL;\
} while (0)
static
void *thr_pintrosort_r(void *arg)
{
  struct thr_pintrosort_r_arg *a = (struct thr_pintrosort_r_arg *)arg;
#define base a->base
#define n a->n
#define z a->z
#define depth a->depth
#define thunk a->thunk
#define cmp a->cmp
  if (z == sizeof(int8_t))
    PINTROSORT(int8_t);
  else if (z == sizeof(int16_t))
    PINTROSORT(int16_t);
  else if (z == sizeof(int32_t))
    PINTROSORT(int32_t);
  else if (z == sizeof(int64_t))
    PINTROSORT(int64_t);
  else {
    typedef struct { char t[z]; } temp_t;
    PINTROSORT(temp_t);
  }
#undef base
#undef n
#undef z
#undef depth
#undef thunk
#undef cmp
}
void pintrosort_r(void *base, size_t n, size_t z,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  unsigned int n_cpu = number_of_processors();
  if (n_cpu > 1) {
    size_t depth;

    introsort_r_limiter(n, &depth);

    pthread_t thrs[16*n_cpu];
    pthread_mutex_t mutex;
    size_t n_thread = 0, i;
    struct thr_pintrosort_r_arg args[sizeof(thrs)/sizeof(thrs[0])], arg = {
      base, n, z, depth, thunk, cmp, &n_thread, sizeof(thrs)/sizeof(thrs[0]), thrs, &mutex, args,
    };

    pthread_mutex_init(&mutex, NULL);
    thr_pintrosort_r(&arg);
    for (i=0; i<n_thread; i++)
      pthread_join(thrs[i], NULL);
  }
  else
    introsort_r(base, n, z, thunk, cmp);
}
#else
/* Thread pool version intro sort */
struct thr_pintrosort_r_arg {
  void *base;
  size_t n, z, depth;
  void *thunk;
  int (*cmp)(void *, const void *, const void *);
  size_t *n_threshold;
  thread_pool_t *thread_pool;
  int free_arg;
};
#define PINTROSORT(T) do {\
  T *b = (T *)base, *p, *q, *r, *s, *t, *m;\
  int h, c;\
\
 loop_##T:\
  if (n < 7) {\
    INSERTSORT(T, base);\
    goto free_return_##T;\
  }\
  if (depth == 0) {\
    HEAPSORT(T);\
    goto free_return_##T;\
  }\
  m = b + n/2;\
  if (n > 7) {\
    T *l, *u;\
    l = b;\
    u = b + (n-1);\
    if (n > 40) {\
      int s = n/4, d = n/8;\
      l = MED3(l, l+d, l+s, thunk, cmp);\
      m = MED3(m-d, m, m+d, thunk, cmp);\
      u = MED3(u-s, u-d, u, thunk, cmp);\
    }\
    m = MED3(l, m, u, thunk, cmp);\
  }\
  TSWAP(T, b, m);\
  for (c=0, p=r=b+1, q=s=b+(n-1); ; p++, q--) {\
    int v;\
    while (p <= q && (v=cmp(thunk, p, b)) <= 0) {\
      if (v == 0) {\
	TSWAP(T, r, p);\
	c++;\
	r++;\
      }\
      p++;\
    }\
    while (p <= q && (v=cmp(thunk, q, b)) >= 0) {\
      if (v == 0) {\
	TSWAP(T, q, s);\
	c++;\
	s--;\
      }\
      q--;\
    }\
    if (p > q) break;\
    TSWAP(T, p, q);\
    c++;\
  }\
  t = b+n-1;\
  h = min(r-b, p-r);\
  TSWAPS(T, b, p-h, h);\
  h = min(s-q, t-s);\
  TSWAPS(T, p, t-h+1, h);\
  if (c == 0) {\
    T *b = (T *)base, *p, *q;\
    int r = n/4 + 1;\
\
    for (p=b+1; p<b+n; p++)\
      for (q=p; q>b && cmp(thunk, q-1, q) > 0; q--) {\
        TSWAP(T, (q-1), q);\
        if (++c > r) goto recover_##T;\
      }\
    goto free_return_##T;\
  }\
 recover_##T:\
  if ((h=p-r) > 1) {\
    struct thr_pintrosort_r_arg arg = {\
      b, h, z, depth-1, thunk, cmp, a->n_threshold, a->thread_pool, !0,\
    }, *thr_arg;\
\
    if (!(h > *a->n_threshold &&\
          (thr_arg = malloc(sizeof(struct thr_pintrosort_r_arg))) &&\
          (*thr_arg = arg, !0) && !thread_pool_queue(a->thread_pool, thr_pintrosort_r, thr_arg))) {\
      arg.free_arg = 0;\
      thr_pintrosort_r(&arg);\
    }\
  }\
  if ((h=s-q) > 1) {\
    /*\
    struct thr_pintrosort_r_arg arg = {\
      t-h+1, h, z, depth-1, thunk, cmp, a->n_threshold, a->thread_pool, !0,\
    }, *thr_arg;\
\
    if (!(h > *a->n_threshold &&\
          (thr_arg = malloc(sizeof(struct thr_pintrosort_r_arg))) &&\
	  (*thr_arg = arg, !0) && !thread_pool_queue(a->thread_pool, thr_pintrosort_r, thr_arg))) {\
      arg.free_arg = 0;\
      thr_pintrosort_r(&arg);\
    }\
    */\
    /**/\
    base = b = t-h+1;\
    n = h;\
    depth--;\
    goto loop_##T;\
    /**/\
  }\
 free_return_##T:\
  if (a->free_arg) free(arg);\
  return NULL;\
} while (0)
static
void *thr_pintrosort_r(void *arg)
{
  struct thr_pintrosort_r_arg *a = (struct thr_pintrosort_r_arg *)arg;
#define base a->base
#define n a->n
#define z a->z
#define depth a->depth
#define thunk a->thunk
#define cmp a->cmp
  if (z == sizeof(int8_t))
    PINTROSORT(int8_t);
  else if (z == sizeof(int16_t))
    PINTROSORT(int16_t);
  else if (z == sizeof(int32_t))
    PINTROSORT(int32_t);
  else if (z == sizeof(int64_t))
    PINTROSORT(int64_t);
  else {
    typedef struct { char t[z]; } temp_t;
    PINTROSORT(temp_t);
  }
#undef base
#undef n
#undef z
#undef depth
#undef thunk
#undef cmp
}
void pintrosort_r(void *base, size_t n, size_t z,
		  void *thunk, int (*cmp)(void *, const void *, const void *))
{
  unsigned int n_cpu = number_of_processors();
  if (n_cpu > 1) {
    size_t depth, n_threshold;

    pintrosort_r_limiter(n, &depth, &n_threshold);

    thread_pool_t *pool;
    struct thr_pintrosort_r_arg arg = {
      base, n, z, depth, thunk, cmp, &n_threshold, (pool = thread_pool_create(n_cpu/*-1*/, NULL)), 0,
    };
    int flag;

    if (!(flag = !thread_pool_queue(pool, thr_pintrosort_r, &arg)))
      introsort_r(base, n, z, thunk, cmp);
    if (flag)
      thread_pool_join(pool);
  }
  else
    introsort_r(base, n, z, thunk, cmp);
}
#endif

/* linear search and append */

void *lsearch_r(const void *key, const void *base, size_t *n, size_t z,
		void *thunk, int (*cmp)(void *, const void *, const void *))
{
  const char *rv, *tail;

  tail = (const char *)base + *n*z;
  for (rv=(const char *)base; rv<tail; rv+=z)
    if (!cmp(thunk, key, (void *)rv))
      return (void *)rv;
  ++*n;
  memcpy((void *)tail, key, z);
  return (void *)tail;
}
void *lfind_r(const void *key, const void *base, size_t *n, size_t z,
	      void *thunk, int (*cmp)(void *, const void *, const void *))
{
  const char *rv, *tail;

  tail = (const char *)base + *n*z;
  for (rv=(const char *)base; rv<tail; rv+=z)
    if (!cmp(thunk, key, (void *)rv))
      return (void *)rv;
  return NULL;
}

/* binary search of a sorted table */

void *bsearch_r(const void *key, const void *base, size_t n, size_t z,
		void *thunk, int (*cmp)(void *, const void *, const void *))
{
  const char *b = base;
  size_t i;

  for (i=n; i!=0; i>>=1) {
    const void *p = b + (i >> 1) * z;
    int r = (*cmp)(thunk, key, p);

    if (!r)
      return (void *)p;
    else if (r > 0) {
      b = (char *)p + z;
      i--;
    }
  }
  return NULL;
}
void *lbsearch_r(const void *key, const void *base, size_t n, size_t z,
		 void *thunk, int (*cmp)(void *, const void *, const void *))
{
  const char *b = base;
  size_t i;

  for (i=n; i!=0; i>>=1) {
    const void *p = b + (i >> 1) * z;
    int r = (*cmp)(thunk, key, p);

    if (r > 0) {
      b = (char *)p + z;
      i--;
    }
  }
  return (void *)b;
}
void *ubsearch_r(const void *key, const void *base, size_t n, size_t z,
		 void *thunk, int (*cmp)(void *, const void *, const void *))
{
  const char *b = base;
  size_t i;

  for (i=n; i!=0; i>>=1) {
    const void *p = b + (i >> 1) * z;
    int r = (*cmp)(thunk, key, p);

    if (r >= 0) {
      b = (char *)p + z;
      i--;
    }
  }
  return (void *)b;
}

#undef min

#if defined(COMPARISON_WITH_BSD)
#define __FBSDID(x)
#define I_AM_QSORT_R
#define qsort_r bsd_qsort_r
#include "qsort-fbsd.c"
#undef qsort_r
#undef swap
#define mergesort bsd_mergesort
#include "merge-fbsd.c"
#undef mergesort
#endif
