/*
  sort_r-bench.cpp - bench mark for reentrant sort variations such as qsort_r

  Copyright (C) 2010 Taiji Yamada <taiji@aihara.co.jp>

  Distributed under the Boost Software License, Version 1.0. See
  http://www.boost.org/LICENSE_1_0.txt for more details.
*/
#include <iostream>
#include <iomanip>
#include <cstdlib>
#include <cmath>
#include <algorithm>
#include <functional>
#include <boost/array.hpp>
#include <boost/multi_array.hpp>
#include <boost/accumulators/accumulators.hpp>
#include <boost/accumulators/statistics.hpp>
#include "StopWatch.hpp"
#include <search.h>	/* lfind, etc */
#include <netinet/in.h>
#include "sort_r.h"
#include "sort.hpp"
#include "bubble_sort.cpp"
#include "comb_sort.cpp"
#include "icomb_sort.cpp"
#include "heap_sort1.cpp"
#include "heap_sort2.cpp"
#include "heap_sort3.cpp"
#include "partial_sort.cpp"
#include "partial_sort2.cpp"
#include "merge_sort.cpp"
#include "pmerge_sort.cpp"
#include "pmerge_sort2.cpp"
#include "mmerge_sort.cpp"
#include "quick_sort.cpp"
#include "pquick_sort.cpp"
#include "pquick_sort2.cpp"
#include "intro_sort.cpp"
using namespace std;
using namespace boost;

/*
 * interpoint distance
 */
double vec_dist(int N, double a[], double b[])
{
  int i;
  double d = 0;

  for (i=0; i<N; i++)
    d += (a[i] - b[i])*(a[i] - b[i]);
  return sqrt(d);
}
/*
 * nD Utilities
 */
double **mat(int M, int N)
{
  int i;
  double **m;

#if defined(UNEFFICIENT)
  m = (double **)malloc(sizeof(double *)*M);
#if defined(UNSAFE)
  for (i=0; i<M; i++)
    m[i] = (double *)malloc(sizeof(double)*N);
#else
  if (!m)
    return NULL;
  for (i=0; i<M; i++) {
    m[i] = (double *)malloc(sizeof(double)*N);
    if (!m[i]) {
      do free(m[--i]); while (i != 0);
      free(m);
      return NULL;
    }
  }
#endif
#else
  m = (double **)malloc(sizeof(double *)*M);
#if defined(UNSAFE)
  m[0] = (double *)malloc(sizeof(double)*M*N);
#else
  if (!m)
    return NULL;
  m[0] = (double *)malloc(sizeof(double)*M*N);
  if (!m[0]) {
    free(m);
    return NULL;
  }
#endif
  for (i=1; i<M; i++) {
    m[i] = m[i-1] + N;
  }
#endif
  return m;
}
void free_mat(int M, double **m)
{
#if defined(UNEFFICIENT)
  int i;

  for (i=0; i<M; i++)
    free(m[i]);
#else
  free(m[0]);
#endif
  free(m);
}
#define deg2rad(d) ((d)*M_PI/180)
void sphere(int n, int n_point_dimension, double center[], double radius[], double **points)
{
  int i, j, k, n_point_sphere = (int)pow((double)n_point_dimension, n-1);

  for (k=0; k<n_point_sphere; k++) {
    int quotient = k;
    double s[n-1], c[n-1];

    for (i=0; i<n-1; i++) {
      double rad = deg2rad(360.0*(quotient % n_point_dimension)/n_point_dimension+
			   90.0/n_point_dimension);
      s[i] = sin(rad);
      c[i] = cos(rad);
      quotient /= n_point_dimension;
    }
    for (j=0; j<n; j++) {
      points[k][j] = radius[j];
      for (i=0; i<n-1; i++)
	points[k][j] *= ((i>j)?1:((i==j)?c[i]:s[i]));
    }
    for (j=0; j<n; j++)
      points[k][j] += center[j];
  }
}
/*
 * t[0..L-1][0..M-1][0..N-1]: rank-3 tensor
 */
double ***r3tensor(int L, int M, int N)
{
  int i, j;
  double ***t;

  t = (double ***)malloc(sizeof(double **)*L);
  if (!t)
    return NULL;
  t[0] = (double **)malloc(sizeof(double *)*L*M);
  if (!t[0]) {
    free(t);
    return NULL;
  }
  t[0][0] = (double *)malloc(sizeof(double)*L*M*N);
  if (!t[0][0]) {
    free(t[0]);
    free(t);
    return NULL;
  }
  for (j=1; j<M; j++)
    t[0][j] = t[0][j-1] + N;
  for (i=1; i<L; i++) {
    t[i] = t[i-1] + M;
    t[i][0] = t[i-1][0] + M*N;
    for (j=1; j<M; j++)
      t[i][j] = t[i][j-1] + N;
  }
  return t;
}
void free_r3tensor(double ***t)
{
  free(t[0][0]);
  free(t[0]);
  free(t);
}

#ifndef N_ELEMENT
#define N_ELEMENT 100000
#endif
#ifndef N_POINT_DIMENSION
#define N_POINT_DIMENSION 360
#endif
#ifndef N_POINT_SPHERE
#define N_POINT_SPHERE (int)pow((double)N_POINT_DIMENSION, 3-1)
#endif
#ifndef N_AVERAGE
#define N_AVERAGE 10
#endif

#if !defined(DETAIL_BENCH)
#define do_bench_sort(cp,op,cmp) do {\
  int i, c;\
  accumulators::accumulator_set<double, accumulators::features<accumulators::tag::mean> > acc_ctime, acc_utime;\
  if (!flag) { cout << left << setw(48) << "operation" << "\t" << setw(10) << "ctime(sec)" << "\t" << "utime(sec)" << endl; flag = !0; }\
  for (c=0; c<N_AVERAGE; c++) {\
    for (i=0; i<n; i++)\
      cp;\
    sw.start();\
    op;\
    sw.stop();\
    acc_ctime(sw.get_ctime());\
    acc_utime(sw.get_utime());\
  }\
  for (i=0; i<n-1; i++)\
    if (cmp) break;\
  cout << left << setw(48) << #op << "\t" << ((i==n-1)?"":"!") << setw(10) << accumulators::extract::mean(acc_ctime) << "\t" << accumulators::extract::mean(acc_utime) << endl;\
} while (0)
#else
#define do_bench_sort(cp,op,cmp) do {\
  int i, k, c, m = n, d[] = { 1000, 500,200,100, 50,20,10, 5,2,1, };\
  accumulators::accumulator_set<double, accumulators::features<accumulators::tag::mean> > acc_ctime;\
  if (!flag) { cout << left << setw(48) << "operation" << "\t"; for (k=0; k<(int)(sizeof(d)/sizeof(d[0])); k++) cout << setw(10) << m/d[k] << "\t"; cout << endl; flag = !0; }\
  cout << left << setw(48) << #op << "\t";\
  for (k=0; k<(int)(sizeof(d)/sizeof(d[0])); k++) {\
    n = m/d[k];\
    for (c=0; c<N_AVERAGE; c++) {\
      for (i=0; i<n; i++)\
        cp;\
      sw.start();\
      op;\
      sw.stop();\
      acc_ctime(sw.get_ctime());\
    }\
    for (i=0; i<n-1; i++)\
      if (cmp) break;\
    cout << ((i==n-1)?"":"!") << setw(10) << accumulators::extract::mean(acc_ctime) << "\t" << flush;\
  }\
  cout << endl;\
  n = m;\
} while (0)
#endif /* !defined(DETAIL_BENCH) */

struct with_in6_addr {
  struct in6_addr addr;
  char c;
};

#define T_CMP(T) ((*(T *)a < *(T *)b) ? -1 : (*(T *)a > *(T *)b) ? 1 : 0)
#define ARI_T_CMP_R(T) ((((T *)thunk)[*(int *)a] < ((T *)thunk)[*(int *)b]) ? -1 : (((T *)thunk)[*(int *)a] > ((T *)thunk)[*(int *)b]) ? 1 : 0)
int c_cmp(const void *a, const void *b)	{ return T_CMP(char); }
int s_cmp(const void *a, const void *b)	{ return T_CMP(short); }
int i_cmp(const void *a, const void *b)	{ return T_CMP(int); }
int d_cmp(const void *a, const void *b)	{ return T_CMP(double); }
int d3_cmp(const void *a, const void *b)
{
  return (((const double *)a)[3] < ((const double *)b)[3]) ? -1 : (((const double *)a)[3] > ((const double *)b)[3]) ? 1 : 0;
}
int in6_cmp(const void *a, const void *b)
{
  return memcmp((const void *)&((struct with_in6_addr *)a)->addr, (const void *)&((struct with_in6_addr *)b)->addr, sizeof(struct in6_addr));
}

int c_cmp_r(void *thunk, const void *a, const void *b)	{ return T_CMP(char); }
int s_cmp_r(void *thunk, const void *a, const void *b)	{ return T_CMP(short); }
int i_cmp_r(void *thunk, const void *a, const void *b)	{ return T_CMP(int); }
int d_cmp_r(void *thunk, const void *a, const void *b)	{ return T_CMP(double); }
int d3_cmp_r(void *thunk, const void *a, const void *b)
{
  return (((const double *)a)[3] < ((const double *)b)[3]) ? -1 : (((const double *)a)[3] > ((const double *)b)[3]) ? 1 : 0;
}
int in6_cmp_r(void *thunk, const void *a, const void *b)
{
  return memcmp((const void *)&((struct with_in6_addr *)a)->addr, (const void *)&((struct with_in6_addr *)b)->addr, sizeof(struct in6_addr));
}
int ari_c_cmp_r(void *thunk, const void *a, const void *b)	{ return ARI_T_CMP_R(char); }
int ari_s_cmp_r(void *thunk, const void *a, const void *b)	{ return ARI_T_CMP_R(short); }
int ari_i_cmp_r(void *thunk, const void *a, const void *b)	{ return ARI_T_CMP_R(int); }
int ari_d_cmp_r(void *thunk, const void *a, const void *b)	{ return ARI_T_CMP_R(double); }
int ari_d3_cmp_r(void *thunk, const void *a, const void *b)
{
  return (((double **)thunk)[*(int *)a][3] < ((double **)thunk)[*(int *)b][3]) ? -1 : (((double **)thunk)[*(int *)a][3] > ((double **)thunk)[*(int *)b][3]) ? 1 : 0;
}
int ari_in6_cmp_r(void *thunk, const void *a, const void *b)
{
  return memcmp((const void *)&((struct with_in6_addr *)thunk)[*(int *)a].addr, (const void *)&((struct with_in6_addr *)thunk)[*(int *)b].addr, sizeof(struct in6_addr));
}

bool operator<(const struct with_in6_addr &a, const struct with_in6_addr &b)
{
  return memcmp(&a.addr, &b.addr, sizeof(a.addr.s6_addr)) < 0;
}
struct in6_comp : public binary_function<struct with_in6_addr, struct with_in6_addr, bool>
{
  bool operator()(const struct with_in6_addr &a, const struct with_in6_addr &b) const { return a < b; }
};
template <typename T, int i>
struct ith_comp { bool operator()(const T &a, const T &b) { return a[i] < b[i]; } };
template <typename T, int i>
struct ari_ith_comp { ari_ith_comp(T *p):v(p){} bool operator()(const int &a, const int &b) { return (*v)[a][i] < (*v)[b][i]; } private: T *v; };
template <typename T>
struct ari_comp { ari_comp(T *p):v(p){} bool operator()(const int &a, const int &b) { return v[a] < v[b]; } private: T *v; };


#define T_RCMP(T) ((*(T *)a > *(T *)b) ? -1 : (*(T *)a < *(T *)b) ? 1 : 0)
#define ARI_T_RCMP_R(T) ((((T *)thunk)[*(int *)a] > ((T *)thunk)[*(int *)b]) ? -1 : (((T *)thunk)[*(int *)a] < ((T *)thunk)[*(int *)b]) ? 1 : 0)
int c_rcmp(const void *a, const void *b)	{ return T_RCMP(char); }
int s_rcmp(const void *a, const void *b)	{ return T_RCMP(short); }
int i_rcmp(const void *a, const void *b)	{ return T_RCMP(int); }
int d_rcmp(const void *a, const void *b)	{ return T_RCMP(double); }
int d3_rcmp(const void *a, const void *b)
{
  return (((const double *)a)[3] > ((const double *)b)[3]) ? -1 : (((const double *)a)[3] < ((const double *)b)[3]) ? 1 : 0;
}
int in6_rcmp(const void *a, const void *b)
{
  return -1*memcmp((const void *)&((struct with_in6_addr *)a)->addr, (const void *)&((struct with_in6_addr *)b)->addr, sizeof(struct in6_addr));
}
int c_rcmp_r(void *thunk, const void *a, const void *b)	{ return T_RCMP(char); }
int s_rcmp_r(void *thunk, const void *a, const void *b)	{ return T_RCMP(short); }
int i_rcmp_r(void *thunk, const void *a, const void *b)	{ return T_RCMP(int); }
int d_rcmp_r(void *thunk, const void *a, const void *b)	{ return T_RCMP(double); }
int d3_rcmp_r(void *thunk, const void *a, const void *b)
{
  return (((const double *)a)[3] > ((const double *)b)[3]) ? -1 : (((const double *)a)[3] < ((const double *)b)[3]) ? 1 : 0;
}
int in6_rcmp_r(void *thunk, const void *a, const void *b)
{
  return -1*memcmp((const void *)&((struct with_in6_addr *)a)->addr, (const void *)&((struct with_in6_addr *)b)->addr, sizeof(struct in6_addr));
}
int ari_c_rcmp_r(void *thunk, const void *a, const void *b)	{ return ARI_T_RCMP_R(char); }
int ari_s_rcmp_r(void *thunk, const void *a, const void *b)	{ return ARI_T_RCMP_R(short); }
int ari_i_rcmp_r(void *thunk, const void *a, const void *b)	{ return ARI_T_RCMP_R(int); }
int ari_d_rcmp_r(void *thunk, const void *a, const void *b)	{ return ARI_T_RCMP_R(double); }
int ari_d3_rcmp_r(void *thunk, const void *a, const void *b)
{
  return (((double **)thunk)[*(int *)a][3] > ((double **)thunk)[*(int *)b][3]) ? -1 : (((double **)thunk)[*(int *)a][3] < ((double **)thunk)[*(int *)b][3]) ? 1 : 0;
}
int ari_in6_rcmp_r(void *thunk, const void *a, const void *b)
{
  return -1*memcmp((const void *)&((struct with_in6_addr *)thunk)[*(int *)a].addr, (const void *)&((struct with_in6_addr *)thunk)[*(int *)b].addr, sizeof(struct in6_addr));
}

bool operator>(const struct with_in6_addr &a, const struct with_in6_addr &b)
{
  return memcmp(&a.addr, &b.addr, sizeof(a.addr.s6_addr)) > 0;
}
struct in6_rcomp : public binary_function<struct with_in6_addr, struct with_in6_addr, bool>
{
  bool operator()(const struct with_in6_addr &a, const struct with_in6_addr &b) const { return a > b; }
};
template <typename T, int i>
struct ith_rcomp { bool operator()(const T &a, const T &b) { return a[i] > b[i]; } };
template <typename T, int i>
struct ari_ith_rcomp { ari_ith_rcomp(T *p):v(p){} bool operator()(const int &a, const int &b) { return (*v)[a][i] > (*v)[b][i]; } private: T *v; };
template <typename T>
struct ari_rcomp { ari_rcomp(T *p):v(p){} bool operator()(const int &a, const int &b) { return v[a] > v[b]; } private: T *v; };

#if !defined(DETAIL_BENCH)
void markbench_sort_char()
{
  int n = N_ELEMENT;
  multi_array<char, 2> x0(extents[N_AVERAGE][n]);
  char *x = new char[n];
  int i, j, *ix = new int[n];
  typedef ari_comp<char> ari_c_comp;
#if defined(__BLOCKS__)
  int (^c_cmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(char *)a < *(char *)b) ? -1 : (*(char *)a > *(char *)b) ? 1 : 0;
  };
  int (^ari_c_cmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (x[*(int *)a] < x[*(int *)b]) ? -1 : (x[*(int *)a] > x[*(int *)b]) ? 1 : 0;
  };
#endif
  int flag = 0;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++)
      x0[j][i] = (char)random();

#if !defined(NO_SORT_VALUES)
  cout << "# number: " << n << ",\t" << "type of compare: char" << ",\t" << "type of swap: char" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubblesort_r(x,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddevensort_r(x,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertsort_r(x,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],poddevensort_r(x,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shellsort_r(x,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pshellsort_r(x,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubble_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],foam_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddeven_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insert_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertion_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_oddeven_sort(x,x+n),x[i] > x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shell_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_shell_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],combsort_r(x,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icombsort_r(x,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],comb_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icomb_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],heapsort(x,n,sizeof(x[0]),c_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heapsort_r(x,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],heapsort_b(x,n,sizeof(x[0]),c_cmp_b),x[i] > x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],partialsort_r(x,n,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],heap_sort1(x,x+n),x[i] > x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],heap_sort2(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heap_sort3(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort(x,x+n,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort2(x,x+n,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],partial_sort(x,x+n,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
#if defined(ENABLE_BSD_MERGESORT_AT_SMALL_VALUE)
  do_bench_sort(x[i] = x0[c][i],mergesort(x,n,sizeof(x[0]),c_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_mergesort(x,n,sizeof(x[0]),c_cmp),x[i] > x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],mergesort_r(x,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
#if defined(ENABLE_BSD_MERGESORT_AT_SMALL_VALUE)
  do_bench_sort(x[i] = x0[c][i],mergesort_b(x,n,sizeof(x[0]),c_cmp_b),x[i] > x[i+1]);
#endif
#endif
  do_bench_sort(x[i] = x0[c][i],timsort_r(x,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],merge_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pmerge_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],mmerge_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],stable_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pmergesort_r(x,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],ptimsort_r(x,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_stable_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],qsort(x,n,sizeof(x[0]),c_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_qsort_r(x,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],qsort_r(x,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],qsort_b(x,n,sizeof(x[0]),c_cmp_b),x[i] > x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],sort(x,x+n),x[i] > x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],quick_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort2(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],introsort_r(x,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],intro_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pintrosort_r(x,n,sizeof(x[0]),0,c_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],psort_b(x,n,sizeof(x[0]),c_cmp_b),x[i] > x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_VALUES) */

#if !defined(NO_SORT_INDICE)
  cout << "# number: " << n << ",\t" << "type of compare: char" << ",\t" << "type of swap: index" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubblesort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddevensort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertsort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],poddevensort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shellsort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pshellsort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubble_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],foam_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddeven_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insert_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertion_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_oddeven_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shell_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_shell_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],combsort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icombsort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],comb_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icomb_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_b(ix,n,sizeof(ix[0]),ari_c_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,n,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort1(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort2(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort3(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+n,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort2(ix,ix+n,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+n,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_b(ix,n,sizeof(ix[0]),ari_c_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],timsort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],merge_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmerge_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],mmerge_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],stable_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmergesort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],ptimsort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_stable_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bsd_qsort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_b(ix,n,sizeof(ix[0]),ari_c_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],quick_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort2(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],introsort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],intro_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pintrosort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],psort_b(ix,n,sizeof(ix[0]),ari_c_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_INDICE) */

  delete[] x;
  delete[] ix;
}
void markbench_sort_short()
{
  int n = N_ELEMENT;
  multi_array<short, 2> x0(extents[N_AVERAGE][n]);
  short *x = new short[n];
  int i, j, *ix = new int[n];
  typedef ari_comp<short> ari_s_comp;
#if defined(__BLOCKS__)
  int (^s_cmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(short *)a < *(short *)b) ? -1 : (*(short *)a > *(short *)b) ? 1 : 0;
  };
  int (^ari_s_cmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (x[*(int *)a] < x[*(int *)b]) ? -1 : (x[*(int *)a] > x[*(int *)b]) ? 1 : 0;
  };
#endif
  int flag = 0;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++)
      x0[j][i] = (short)random();

#if !defined(NO_SORT_VALUES)
  cout << "# number: " << n << ",\t" << "type of compare: short" << ",\t" << "type of swap: short" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubblesort_r(x,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddevensort_r(x,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertsort_r(x,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],poddevensort_r(x,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shellsort_r(x,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pshellsort_r(x,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubble_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],foam_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddeven_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insert_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertion_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_oddeven_sort(x,x+n),x[i] > x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shell_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_shell_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],combsort_r(x,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icombsort_r(x,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],comb_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icomb_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],heapsort(x,n,sizeof(x[0]),s_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heapsort_r(x,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],heapsort_b(x,n,sizeof(x[0]),s_cmp_b),x[i] > x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],partialsort_r(x,n,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],heap_sort1(x,x+n),x[i] > x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],heap_sort2(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heap_sort3(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort(x,x+n,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort2(x,x+n,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],partial_sort(x,x+n,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
#if defined(ENABLE_BSD_MERGESORT_AT_SMALL_VALUE)
  do_bench_sort(x[i] = x0[c][i],mergesort(x,n,sizeof(x[0]),s_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_mergesort(x,n,sizeof(x[0]),s_cmp),x[i] > x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],mergesort_r(x,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
#if defined(ENABLE_BSD_MERGESORT_AT_SMALL_VALUE)
  do_bench_sort(x[i] = x0[c][i],mergesort_b(x,n,sizeof(x[0]),s_cmp_b),x[i] > x[i+1]);
#endif
#endif
  do_bench_sort(x[i] = x0[c][i],timsort_r(x,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],merge_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pmerge_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],mmerge_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],stable_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pmergesort_r(x,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],ptimsort_r(x,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_stable_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],qsort(x,n,sizeof(x[0]),s_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_qsort_r(x,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],qsort_r(x,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],qsort_b(x,n,sizeof(x[0]),s_cmp_b),x[i] > x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],sort(x,x+n),x[i] > x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],quick_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort2(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],introsort_r(x,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],intro_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pintrosort_r(x,n,sizeof(x[0]),0,s_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],psort_b(x,n,sizeof(x[0]),s_cmp_b),x[i] > x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_VALUES) */

#if !defined(NO_SORT_INDICE)
  cout << "# number: " << n << ",\t" << "type of compare: short" << ",\t" << "type of swap: index" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubblesort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddevensort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertsort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],poddevensort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shellsort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pshellsort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubble_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],foam_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddeven_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insert_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertion_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_oddeven_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shell_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_shell_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],combsort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icombsort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],comb_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icomb_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_b(ix,n,sizeof(ix[0]),ari_s_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,n,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort1(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort2(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort3(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+n,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+n,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+n,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_b(ix,n,sizeof(ix[0]),ari_s_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],timsort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],merge_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmerge_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],mmerge_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],stable_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmergesort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],ptimsort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_stable_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bsd_qsort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_b(ix,n,sizeof(ix[0]),ari_s_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],quick_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort2(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],introsort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],intro_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pintrosort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],psort_b(ix,n,sizeof(ix[0]),ari_s_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_INDICE) */

  delete[] x;
  delete[] ix;
}
void markbench_sort_int()
{
  int n = N_ELEMENT;
  multi_array<int, 2> x0(extents[N_AVERAGE][n]);
  int *x = new int[n];
  int i, j, *ix = new int[n];
  typedef ari_comp<int> ari_i_comp;
#if defined(__BLOCKS__)
  int (^i_cmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(int *)a < *(int *)b) ? -1 : (*(int *)a > *(int *)b) ? 1 : 0;
  };
  int (^ari_i_cmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (x[*(int *)a] < x[*(int *)b]) ? -1 : (x[*(int *)a] > x[*(int *)b]) ? 1 : 0;
  };
#endif
  int flag = 0;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++)
      x0[j][i] = (int)random();

#if !defined(NO_SORT_VALUES)
  cout << "# number: " << n << ",\t" << "type of compare: int" << ",\t" << "type of swap: int" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubblesort_r(x,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddevensort_r(x,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertsort_r(x,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],poddevensort_r(x,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shellsort_r(x,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pshellsort_r(x,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubble_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],foam_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddeven_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insert_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertion_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_oddeven_sort(x,x+n),x[i] > x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shell_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_shell_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],combsort_r(x,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icombsort_r(x,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],comb_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icomb_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],heapsort(x,n,sizeof(x[0]),i_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heapsort_r(x,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],heapsort_b(x,n,sizeof(x[0]),i_cmp_b),x[i] > x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],partialsort_r(x,n,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],heap_sort1(x,x+n),x[i] > x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],heap_sort2(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heap_sort3(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort(x,x+n,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort2(x,x+n,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],partial_sort(x,x+n,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],mergesort(x,n,sizeof(x[0]),i_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_mergesort(x,n,sizeof(x[0]),i_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],mergesort_r(x,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],mergesort_b(x,n,sizeof(x[0]),i_cmp_b),x[i] > x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],timsort_r(x,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],merge_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pmerge_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],mmerge_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],stable_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pmergesort_r(x,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],ptimsort_r(x,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_stable_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],qsort(x,n,sizeof(x[0]),i_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_qsort_r(x,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],qsort_r(x,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],qsort_b(x,n,sizeof(x[0]),i_cmp_b),x[i] > x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],sort(x,x+n),x[i] > x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],quick_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort2(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],introsort_r(x,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],intro_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pintrosort_r(x,n,sizeof(x[0]),0,i_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],psort_b(x,n,sizeof(x[0]),i_cmp_b),x[i] > x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_VALUES) */

#if !defined(NO_SORT_INDICE)
  cout << "# number: " << n << ",\t" << "type of compare: int" << ",\t" << "type of swap: index" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubblesort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddevensort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertsort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],poddevensort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shellsort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pshellsort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubble_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],foam_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddeven_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insert_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertion_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_oddeven_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shell_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_shell_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],combsort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icombsort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],comb_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icomb_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_b(ix,n,sizeof(ix[0]),ari_i_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,n,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort1(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort2(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort3(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+n,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort2(ix,ix+n,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+n,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_b(ix,n,sizeof(ix[0]),ari_i_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],timsort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],merge_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmerge_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],mmerge_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],stable_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmergesort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],ptimsort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_stable_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bsd_qsort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_b(ix,n,sizeof(ix[0]),ari_i_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],quick_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort2(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],introsort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],intro_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pintrosort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],psort_b(ix,n,sizeof(ix[0]),ari_i_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_INDICE) */

  delete[] x;
  delete[] ix;
}
void markbench_sort_double()
{
  int n = N_ELEMENT;
  multi_array<double, 2> x0(extents[N_AVERAGE][n]);
  double *x = new double[n];
  int i, j, *ix = new int[n];
  typedef ari_comp<double> ari_d_comp;
#if defined(__BLOCKS__)
  int (^d_cmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(double *)a < *(double *)b) ? -1 : (*(double *)a > *(double *)b) ? 1 : 0;
  };
  int (^ari_d_cmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (x[*(int *)a] < x[*(int *)b]) ? -1 : (x[*(int *)a] > x[*(int *)b]) ? 1 : 0;
  };
#endif
  int flag = 0;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++)
      x0[j][i] = (double)random()/((1<<31)-1);

#if !defined(NO_SORT_VALUES)
  cout << "# number: " << n << ",\t" << "type of compare: double" << ",\t" << "type of swap: double" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubblesort_r(x,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddevensort_r(x,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertsort_r(x,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],poddevensort_r(x,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shellsort_r(x,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pshellsort_r(x,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubble_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],foam_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddeven_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insert_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertion_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_oddeven_sort(x,x+n),x[i] > x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shell_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_shell_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],combsort_r(x,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icombsort_r(x,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],comb_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icomb_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],heapsort(x,n,sizeof(x[0]),d_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heapsort_r(x,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],heapsort_b(x,n,sizeof(x[0]),d_cmp_b),x[i] > x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],partialsort_r(x,n,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],heap_sort1(x,x+n),x[i] > x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],heap_sort2(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heap_sort3(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort(x,x+n,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort2(x,x+n,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],partial_sort(x,x+n,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],mergesort(x,n,sizeof(x[0]),d_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_mergesort(x,n,sizeof(x[0]),d_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],mergesort_r(x,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],mergesort_b(x,n,sizeof(x[0]),d_cmp_b),x[i] > x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],timsort_r(x,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],merge_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pmerge_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],mmerge_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],stable_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pmergesort_r(x,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],ptimsort_r(x,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_stable_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],qsort(x,n,sizeof(x[0]),d_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_qsort_r(x,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],qsort_r(x,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],qsort_b(x,n,sizeof(x[0]),d_cmp_b),x[i] > x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],sort(x,x+n),x[i] > x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],quick_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort2(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],introsort_r(x,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],intro_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pintrosort_r(x,n,sizeof(x[0]),0,d_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],psort_b(x,n,sizeof(x[0]),d_cmp_b),x[i] > x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_VALUES) */

#if !defined(NO_SORT_INDICE)
  cout << "# number: " << n << ",\t" << "type of compare: double" << ",\t" << "type of swap: index" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubblesort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddevensort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],poddevensort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shellsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pshellsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubble_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],foam_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddeven_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insert_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertion_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_oddeven_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shell_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_shell_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],combsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icombsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],comb_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icomb_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_b(ix,n,sizeof(ix[0]),ari_d_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,n,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort1(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort2(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort3(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+n,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort2(ix,ix+n,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+n,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_b(ix,n,sizeof(ix[0]),ari_d_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],timsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],merge_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmerge_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],mmerge_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],stable_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmergesort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],ptimsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_stable_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bsd_qsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_b(ix,n,sizeof(ix[0]),ari_d_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],quick_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort2(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],introsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],intro_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pintrosort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],psort_b(ix,n,sizeof(ix[0]),ari_d_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_INDICE) */

  delete[] x;
  delete[] ix;
}
void markbench_sort_doubles()
{
  int n = N_POINT_SPHERE;
  double ***x0 = r3tensor(N_AVERAGE, n, 4), **x = mat(n, 4), c[] = { 0, 0, 0 }, r[] = { 1, 1, 1 }, p[3];
  vector<array<double, 4> > y(n);
  int i, j, *ix = new int[n];
  typedef ith_comp<array<double, 4>, 3> d3_comp;
  typedef ari_ith_comp<vector<array<double, 4> >, 3> ari_d3_comp;
#if defined(__BLOCKS__)
  int (^d3_cmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (((double **)a)[3] < ((double **)b)[3]) ? -1 : (((double **)a)[3] > ((double **)b)[3]) ? 1 : 0;
  };
  int (^ari_d3_cmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (x[*(int *)a][3] < x[*(int *)b][3]) ? -1 : (x[*(int *)a][3] > x[*(int *)b][3]) ? 1 : 0;
  };
#endif
  int flag = 0;
  StopWatch sw;
  
  for (j=0; j<N_AVERAGE; j++) {
    sphere(3, N_POINT_DIMENSION, c, r, x0[j]);
    for (i=0; i<3; i++)
      p[i] = (double)random()/((1<<31)-1);
    for (i=0; i<n; i++)
      x0[j][i][3] = vec_dist(3, p, x0[j][i]);
  }

#if !defined(NO_SORT_VALUES)
  cout << "# number: " << n << ",\t" << "type of compare: doubles" << ",\t" << "type of swap: doubles" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),bubblesort_r(x[0],n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),oddevensort_r(x[0],n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),insertsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),poddevensort_r(x[0],n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),shellsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),pshellsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],bubble_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],foam_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],oddeven_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],insert_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],insertion_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],parallel_oddeven_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],shell_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],parallel_shell_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),combsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),icombsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],comb_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],icomb_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),heapsort(x[0],n,sizeof(x[0][0])*4,d3_cmp),x[i][3] > x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),heapsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
#if defined(__BLOCKS__)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),heapsort_b(x[0],n,sizeof(x[0][0])*4,d3_cmp_b),x[i][3] > x[i+1][3]);
#endif
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),partialsort_r(x[0],n,n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],heap_sort1(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  //do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],heap_sort2(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],heap_sort3(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],my_partial_sort(y.begin(),y.begin()+n,y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],my_partial_sort2(y.begin(),y.begin()+n,y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],partial_sort(y.begin(),y.begin()+n,y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),mergesort(x[0],n,sizeof(x[0][0])*4,d3_cmp),x[i][3] > x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),bsd_mergesort(x[0],n,sizeof(x[0][0])*4,d3_cmp),x[i][3] > x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),mergesort_r(x[0],n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
#if defined(__BLOCKS__)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),mergesort_b(x[0],n,sizeof(x[0][0])*4,d3_cmp_b),x[i][3] > x[i+1][3]);
#endif
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),timsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],merge_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],pmerge_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],mmerge_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],stable_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),pmergesort_r(x[0],n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),ptimsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],parallel_stable_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),qsort(x[0],n,sizeof(x[0][0])*4,d3_cmp),x[i][3] > x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),bsd_qsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),qsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
#if defined(__BLOCKS__)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),qsort_b(x[0],n,sizeof(x[0][0])*4,d3_cmp_b),x[i][3] > x[i+1][3]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  //do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],quick_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],pquick_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],pquick_sort2(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),introsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],intro_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),pintrosort_r(x[0],n,sizeof(x[0][0])*4,0,d3_cmp_r),x[i][3] > x[i+1][3]);
#if defined(__BLOCKS__)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),psort_b(x[0],n,sizeof(x[0][0])*4,d3_cmp_b),x[i][3] > x[i+1][3]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],parallel_sort(y.begin(),y.begin()+n,d3_comp()),y[i][3] > y[i+1][3]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_VALUES) */

#if !defined(NO_SORT_INDICE)
  cout << "# number: " << n << ",\t" << "type of compare: doubles" << ",\t" << "type of swap: index" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),bubblesort_r(ix,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),oddevensort_r(ix,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),insertsort_r(ix,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),poddevensort_r(ix,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),shellsort_r(ix,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),pshellsort_r(ix,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],bubble_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],foam_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],oddeven_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],insert_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],insertion_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],parallel_oddeven_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],shell_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],parallel_shell_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),combsort_r(ix,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),icombsort_r(ix,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],comb_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],icomb_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),heapsort_r(ix,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
#if defined(__BLOCKS__)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),heapsort_b(ix,n,sizeof(ix[0]),ari_d3_cmp_b),x[ix[i]][3] > x[ix[i+1]][3]);
#endif
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),partialsort_r(ix,n,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],heap_sort1(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  //do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],heap_sort2(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],heap_sort3(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],my_partial_sort(ix,ix+n,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],my_partial_sort2(ix,ix+n,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],partial_sort(ix,ix+n,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),mergesort_r(ix,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
#if defined(__BLOCKS__)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),mergesort_b(ix,n,sizeof(ix[0]),ari_d3_cmp_b),x[ix[i]][3] > x[ix[i+1]][3]);
#endif
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),timsort_r(ix,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],merge_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],pmerge_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],mmerge_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],stable_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),pmergesort_r(ix,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),ptimsort_r(ix,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],parallel_stable_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),bsd_qsort_r(ix,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),qsort_r(ix,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
#if defined(__BLOCKS__)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),qsort_b(ix,n,sizeof(ix[0]),ari_d3_cmp_b),x[ix[i]][3] > x[ix[i+1]][3]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  //do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],quick_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],pquick_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],pquick_sort2(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),introsort_r(ix,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],intro_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),pintrosort_r(ix,n,sizeof(ix[0]),x,ari_d3_cmp_r),x[ix[i]][3] > x[ix[i+1]][3]);
#if defined(__BLOCKS__)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),psort_b(ix,n,sizeof(ix[0]),ari_d3_cmp_b),x[ix[i]][3] > x[ix[i+1]][3]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],parallel_sort(ix,ix+n,ari_d3_comp(&y)),y[ix[i]][3] > y[ix[i+1]][3]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_INDICE) */

  free_r3tensor(x0);
  free_mat(n, x);
  delete[] ix;
}
void markbench_sort_array_indice_doubles()
{
  int n = N_POINT_SPHERE;
  double ***x0 = r3tensor(N_AVERAGE, n, 4), *x = new double[n], c[] = { 0, 0, 0 }, r[] = { 1, 1, 1 }, p[3];
  int i, j, *ix = new int[n];
  typedef ari_comp<double> ari_d_comp;
#if defined(__BLOCKS__)
  int (^ari_d_cmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (x[*(int *)a] < x[*(int *)b]) ? -1 : (x[*(int *)a] > x[*(int *)b]) ? 1 : 0;
  };
#endif
  int flag = 0;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++) {
    sphere(3, N_POINT_DIMENSION, c, r, x0[j]);
    for (i=0; i<3; i++)
      p[i] = (double)random()/((1<<31)-1);
    for (i=0; i<n; i++)
      x0[j][i][3] = vec_dist(3, p, x0[j][i]);
  }

  cout << "# number: " << n << ",\t" << "type of compare: double" << ",\t" << "type of swap: index" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],bubblesort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],oddevensort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],insertsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],poddevensort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],shellsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],pshellsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],bubble_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],foam_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],oddeven_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],insert_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],insertion_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],parallel_oddeven_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],shell_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],parallel_shell_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],combsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],icombsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],comb_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],icomb_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],heapsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],heapsort_b(ix,n,sizeof(ix[0]),ari_d_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],partialsort_r(ix,n,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i][3],heap_sort1(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i][3],heap_sort2(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],heap_sort3(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],my_partial_sort(ix,ix+n,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],my_partial_sort2(ix,ix+n,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],partial_sort(ix,ix+n,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],mergesort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],mergesort_b(ix,n,sizeof(ix[0]),ari_d_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],timsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i][3],merge_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],pmerge_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],mmerge_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],stable_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],pmergesort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],ptimsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],parallel_stable_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],bsd_qsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],qsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],qsort_b(ix,n,sizeof(ix[0]),ari_d_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i][3],quick_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],pquick_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],pquick_sort2(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],introsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],intro_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],pintrosort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],psort_b(ix,n,sizeof(ix[0]),ari_d_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],parallel_sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */

  free_r3tensor(x0);
  delete[] x;
  delete[] ix;
}
void markbench_sort_in6()
{
  int n = N_ELEMENT;
  multi_array<struct with_in6_addr, 2> x0(extents[N_AVERAGE][n]);
  struct with_in6_addr *x = new struct with_in6_addr[n];
  int i, j, k, *ix = new int[n];
  typedef ari_comp<struct with_in6_addr> ari_in6_comp;
#if defined(__BLOCKS__)
  int (^in6_cmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return memcmp((const void *)&((struct with_in6_addr *)a)->addr, (const void *)&((struct with_in6_addr *)b)->addr, sizeof(struct in6_addr));
  };
  int (^ari_in6_cmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return memcmp((const void *)&x[*(int *)a].addr, (const void *)&x[*(int *)b].addr, sizeof(struct in6_addr));
  };
#endif
  int flag = 0;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++) {
      uint16_t *p = (uint16_t *)&x0[j][i].addr.s6_addr;
      for (k=0; k<(int)(sizeof(struct in6_addr)/sizeof(uint16_t)); k++)
	p[k] = (uint16_t)random();
      x0[j][i].c = i;
    }

#if !defined(NO_SORT_VALUES)
  cout << "# number: " << n << ",\t" << "type of compare: in6" << ",\t" << "type of swap: with_in6" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubblesort_r(x,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddevensort_r(x,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertsort_r(x,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],poddevensort_r(x,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shellsort_r(x,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pshellsort_r(x,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubble_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],foam_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddeven_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insert_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertion_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_oddeven_sort(x,x+n),x[i] > x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shell_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_shell_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],combsort_r(x,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icombsort_r(x,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],comb_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icomb_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],heapsort(x,n,sizeof(x[0]),in6_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heapsort_r(x,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],heapsort_b(x,n,sizeof(x[0]),in6_cmp_b),x[i] > x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],partialsort_r(x,n,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],heap_sort1(x,x+n),x[i] > x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],heap_sort2(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heap_sort3(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort(x,x+n,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort2(x,x+n,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],partial_sort(x,x+n,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],mergesort(x,n,sizeof(x[0]),in6_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_mergesort(x,n,sizeof(x[0]),in6_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],mergesort_r(x,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],mergesort_b(x,n,sizeof(x[0]),in6_cmp_b),x[i] > x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],timsort_r(x,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],merge_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pmerge_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],mmerge_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],stable_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pmergesort_r(x,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],ptimsort_r(x,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_stable_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],qsort(x,n,sizeof(x[0]),in6_cmp),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_qsort_r(x,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],qsort_r(x,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],qsort_b(x,n,sizeof(x[0]),in6_cmp_b),x[i] > x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],sort(x,x+n),x[i] > x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],quick_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort(x,x+n),x[i] > x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort2(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],introsort_r(x,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],intro_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */

  do_bench_sort(x[i] = x0[c][i],pintrosort_r(x,n,sizeof(x[0]),0,in6_cmp_r),x[i] > x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],psort_b(x,n,sizeof(x[0]),in6_cmp_b),x[i] > x[i+1]);
#endif
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_sort(x,x+n),x[i] > x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_VALUES) */

#if !defined(NO_SORT_INDICE)
  cout << "# number: " << n << ",\t" << "type of compare: in6" << ",\t" << "type of swap: index" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubblesort_r(ix,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddevensort_r(ix,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertsort_r(ix,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],poddevensort_r(ix,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shellsort_r(ix,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pshellsort_r(ix,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubble_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],foam_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddeven_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insert_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertion_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_oddeven_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shell_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_shell_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],combsort_r(ix,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icombsort_r(ix,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],comb_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icomb_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_r(ix,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_b(ix,n,sizeof(ix[0]),ari_in6_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,n,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort1(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort2(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort3(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+n,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+n,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+n,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_r(ix,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_b(ix,n,sizeof(ix[0]),ari_in6_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],timsort_r(ix,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],merge_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmerge_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],mmerge_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],stable_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmergesort_r(ix,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],ptimsort_r(ix,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_stable_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bsd_qsort_r(ix,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_r(ix,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_b(ix,n,sizeof(ix[0]),ari_in6_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],quick_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort2(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],introsort_r(ix,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],intro_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pintrosort_r(ix,n,sizeof(ix[0]),x,ari_in6_cmp_r),x[ix[i]] > x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],psort_b(ix,n,sizeof(ix[0]),ari_in6_cmp_b),x[ix[i]] > x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_sort(ix,ix+n,ari_in6_comp(x)),x[ix[i]] > x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_INDICE) */

  delete[] x;
  delete[] ix;
}
#endif /* !defined(DETAIL_BENCH) */
void markbench_rsort_char()
{
  int n = N_ELEMENT;
  multi_array<char, 2> x0(extents[N_AVERAGE][n]);
  char *x = new char[n];
  int i, j, *ix = new int[n];
  typedef ari_rcomp<char> ari_c_rcomp;
#if defined(__BLOCKS__)
  int (^c_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(char *)a > *(char *)b) ? -1 : (*(char *)a < *(char *)b) ? 1 : 0;
  };
  int (^ari_c_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (x[*(int *)a] > x[*(int *)b]) ? -1 : (x[*(int *)a] < x[*(int *)b]) ? 1 : 0;
  };
#endif
  int flag = 0;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++)
      x0[j][i] = (char)random();

#if !defined(NO_SORT_VALUES)
  cout << "# number: " << n << ",\t" << "type of compare: char" << ",\t" << "type of swap: char" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubblesort_r(x,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddevensort_r(x,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertsort_r(x,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],poddevensort_r(x,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shellsort_r(x,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pshellsort_r(x,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubble_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],foam_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddeven_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insert_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertion_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_oddeven_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shell_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_shell_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],combsort_r(x,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icombsort_r(x,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],comb_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icomb_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],heapsort(x,n,sizeof(x[0]),c_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heapsort_r(x,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],heapsort_b(x,n,sizeof(x[0]),c_rcmp_b),x[i] < x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],partialsort_r(x,n,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],heap_sort1(x,x+n,greater<char>()),x[i] < x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],heap_sort2(x,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heap_sort3(x,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort(x,x+n,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort2(x,x+n,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],partial_sort(x,x+n,x+n,greater<char>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
#if defined(ENABLE_BSD_MERGESORT_AT_SMALL_VALUE)
  do_bench_sort(x[i] = x0[c][i],mergesort(x,n,sizeof(x[0]),c_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_mergesort(x,n,sizeof(x[0]),c_rcmp),x[i] < x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],mergesort_r(x,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
#if defined(ENABLE_BSD_MERGESORT_AT_SMALL_VALUE)
  do_bench_sort(x[i] = x0[c][i],mergesort_b(x,n,sizeof(x[0]),c_rcmp_b),x[i] < x[i+1]);
#endif
#endif
  do_bench_sort(x[i] = x0[c][i],timsort_r(x,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],merge_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pmerge_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],mmerge_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],stable_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pmergesort_r(x,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],ptimsort_r(x,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
  do_bench_sort(x[i] = x0[c][i],parallel_stable_sort(x,x+n,greater<char>()),x[i] < x[i+1]);

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],qsort(x,n,sizeof(x[0]),c_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_qsort_r(x,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],qsort_r(x,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],qsort_b(x,n,sizeof(x[0]),c_rcmp_b),x[i] < x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],sort(x,x+n,greater<char>()),x[i] < x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],quick_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort2(x,x+n,greater<char>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],introsort_r(x,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],intro_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pintrosort_r(x,n,sizeof(x[0]),0,c_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],psort_b(x,n,sizeof(x[0]),c_rcmp_b),x[i] < x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_sort(x,x+n,greater<char>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_VALUES) */

#if !defined(NO_SORT_INDICE)
  cout << "# number: " << n << ",\t" << "type of compare: char" << ",\t" << "type of swap: index" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubblesort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddevensort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertsort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],poddevensort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shellsort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pshellsort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubble_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],foam_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddeven_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insert_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertion_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_oddeven_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shell_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_shell_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],combsort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icombsort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],comb_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icomb_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_b(ix,n,sizeof(ix[0]),ari_c_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,n,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort1(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort2(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort3(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+n,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort2(ix,ix+n,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+n,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_b(ix,n,sizeof(ix[0]),ari_c_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],timsort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],merge_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmerge_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],mmerge_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],stable_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmergesort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],ptimsort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_stable_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bsd_qsort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_b(ix,n,sizeof(ix[0]),ari_c_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],quick_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort2(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],introsort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],intro_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pintrosort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],psort_b(ix,n,sizeof(ix[0]),ari_c_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_INDICE) */

  delete[] x;
  delete[] ix;
}
void markbench_rsort_short()
{
  int n = N_ELEMENT;
  multi_array<short, 2> x0(extents[N_AVERAGE][n]);
  short *x = new short[n];
  int i, j, *ix = new int[n];
  typedef ari_rcomp<short> ari_s_rcomp;
#if defined(__BLOCKS__)
  int (^s_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(short *)a > *(short *)b) ? -1 : (*(short *)a < *(short *)b) ? 1 : 0;
  };
  int (^ari_s_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (x[*(int *)a] > x[*(int *)b]) ? -1 : (x[*(int *)a] < x[*(int *)b]) ? 1 : 0;
  };
#endif
  int flag = 0;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++)
      x0[j][i] = (short)random();

#if !defined(NO_SORT_VALUES)
  cout << "# number: " << n << ",\t" << "type of compare: short" << ",\t" << "type of swap: short" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubblesort_r(x,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddevensort_r(x,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertsort_r(x,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],poddevensort_r(x,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shellsort_r(x,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pshellsort_r(x,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubble_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],foam_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddeven_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insert_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertion_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_oddeven_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shell_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_shell_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],combsort_r(x,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icombsort_r(x,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],comb_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icomb_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],heapsort(x,n,sizeof(x[0]),s_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heapsort_r(x,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],heapsort_b(x,n,sizeof(x[0]),s_rcmp_b),x[i] < x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],partialsort_r(x,n,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],heap_sort1(x,x+n,greater<short>()),x[i] < x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],heap_sort2(x,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heap_sort3(x,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort(x,x+n,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort2(x,x+n,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],partial_sort(x,x+n,x+n,greater<short>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
#if defined(ENABLE_BSD_MERGESORT_AT_SMALL_VALUE)
  do_bench_sort(x[i] = x0[c][i],mergesort(x,n,sizeof(x[0]),s_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_mergesort(x,n,sizeof(x[0]),s_rcmp),x[i] < x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],mergesort_r(x,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
#if defined(ENABLE_BSD_MERGESORT_AT_SMALL_VALUE)
  do_bench_sort(x[i] = x0[c][i],mergesort_b(x,n,sizeof(x[0]),s_rcmp_b),x[i] < x[i+1]);
#endif
#endif
  do_bench_sort(x[i] = x0[c][i],timsort_r(x,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],merge_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pmerge_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],mmerge_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],stable_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pmergesort_r(x,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],ptimsort_r(x,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_stable_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],qsort(x,n,sizeof(x[0]),s_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_qsort_r(x,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],qsort_r(x,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],qsort_b(x,n,sizeof(x[0]),s_rcmp_b),x[i] < x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],sort(x,x+n,greater<short>()),x[i] < x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],quick_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort2(x,x+n,greater<short>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],introsort_r(x,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],intro_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pintrosort_r(x,n,sizeof(x[0]),0,s_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],psort_b(x,n,sizeof(x[0]),s_rcmp_b),x[i] < x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_sort(x,x+n,greater<short>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_VALUES) */

#if !defined(NO_SORT_INDICE)
  cout << "# number: " << n << ",\t" << "type of compare: short" << ",\t" << "type of swap: index" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubblesort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddevensort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertsort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],poddevensort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shellsort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pshellsort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubble_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],foam_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddeven_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insert_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertion_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_oddeven_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shell_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_shell_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],combsort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icombsort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],comb_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icomb_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_b(ix,n,sizeof(ix[0]),ari_s_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,n,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort1(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort2(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort3(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+n,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort2(ix,ix+n,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+n,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_b(ix,n,sizeof(ix[0]),ari_s_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],timsort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],merge_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmerge_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],mmerge_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],stable_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmergesort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],ptimsort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_stable_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bsd_qsort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_b(ix,n,sizeof(ix[0]),ari_s_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],quick_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort2(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],introsort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],intro_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pintrosort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],psort_b(ix,n,sizeof(ix[0]),ari_s_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_INDICE) */

  delete[] x;
  delete[] ix;
}
void markbench_rsort_int()
{
  int n = N_ELEMENT;
  multi_array<int, 2> x0(extents[N_AVERAGE][n]);
  int *x = new int[n];
  int i, j, *ix = new int[n];
  typedef ari_rcomp<int> ari_i_rcomp;
#if defined(__BLOCKS__)
  int (^i_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(int *)a > *(int *)b) ? -1 : (*(int *)a < *(int *)b) ? 1 : 0;
  };
  int (^ari_i_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (x[*(int *)a] > x[*(int *)b]) ? -1 : (x[*(int *)a] < x[*(int *)b]) ? 1 : 0;
  };
#endif
  int flag = 0;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++)
      x0[j][i] = (int)random();

#if !defined(NO_SORT_VALUES)
  cout << "# number: " << n << ",\t" << "type of compare: int" << ",\t" << "type of swap: int" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubblesort_r(x,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddevensort_r(x,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertsort_r(x,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],poddevensort_r(x,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shellsort_r(x,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pshellsort_r(x,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubble_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],foam_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddeven_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insert_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertion_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_oddeven_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shell_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_shell_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],combsort_r(x,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icombsort_r(x,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],comb_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icomb_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],heapsort(x,n,sizeof(x[0]),i_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heapsort_r(x,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],heapsort_b(x,n,sizeof(x[0]),i_rcmp_b),x[i] < x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],partialsort_r(x,n,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],heap_sort1(x,x+n,greater<int>()),x[i] < x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],heap_sort2(x,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heap_sort3(x,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort(x,x+n,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort2(x,x+n,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],partial_sort(x,x+n,x+n,greater<int>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],mergesort(x,n,sizeof(x[0]),i_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_mergesort(x,n,sizeof(x[0]),i_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],mergesort_r(x,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],mergesort_b(x,n,sizeof(x[0]),i_rcmp_b),x[i] < x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],timsort_r(x,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],merge_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pmerge_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],mmerge_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],stable_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pmergesort_r(x,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],ptimsort_r(x,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_stable_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],qsort(x,n,sizeof(x[0]),i_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_qsort_r(x,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],qsort_r(x,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],qsort_b(x,n,sizeof(x[0]),i_rcmp_b),x[i] < x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],sort(x,x+n,greater<int>()),x[i] < x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],quick_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort2(x,x+n,greater<int>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],introsort_r(x,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],intro_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pintrosort_r(x,n,sizeof(x[0]),0,i_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],psort_b(x,n,sizeof(x[0]),i_rcmp_b),x[i] < x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_sort(x,x+n,greater<int>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_VALUES) */

#if !defined(NO_SORT_INDICE)
  cout << "# number: " << n << ",\t" << "type of compare: int" << ",\t" << "type of swap: index" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubblesort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddevensort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertsort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],poddevensort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shellsort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pshellsort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubble_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],foam_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddeven_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insert_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertion_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_oddeven_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shell_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_shell_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],combsort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icombsort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],comb_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icomb_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_b(ix,n,sizeof(ix[0]),ari_i_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,n,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort1(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort2(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort3(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+n,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort2(ix,ix+n,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+n,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_b(ix,n,sizeof(ix[0]),ari_i_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],timsort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],merge_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmerge_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],mmerge_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],stable_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmergesort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],ptimsort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_stable_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bsd_qsort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_b(ix,n,sizeof(ix[0]),ari_i_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],quick_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort2(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],introsort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],intro_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pintrosort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],psort_b(ix,n,sizeof(ix[0]),ari_i_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_INDICE) */

  delete[] x;
  delete[] ix;
}
void markbench_rsort_double()
{
  int n = N_ELEMENT;
  multi_array<double, 2> x0(extents[N_AVERAGE][n]);
  double *x = new double[n];
  int i, j, *ix = new int[n];
  typedef ari_rcomp<double> ari_d_rcomp;
#if defined(__BLOCKS__)
  int (^d_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(double *)a > *(double *)b) ? -1 : (*(double *)a < *(double *)b) ? 1 : 0;
  };
  int (^ari_d_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (x[*(int *)a] > x[*(int *)b]) ? -1 : (x[*(int *)a] < x[*(int *)b]) ? 1 : 0;
  };
#endif
  int flag = 0;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++)
      x0[j][i] = (double)random()/((1<<31)-1);

#if !defined(NO_SORT_VALUES)
  cout << "# number: " << n << ",\t" << "type of compare: double" << ",\t" << "type of swap: double" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubblesort_r(x,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddevensort_r(x,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertsort_r(x,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],poddevensort_r(x,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shellsort_r(x,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pshellsort_r(x,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubble_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],foam_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddeven_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insert_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertion_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_oddeven_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shell_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_shell_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],combsort_r(x,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icombsort_r(x,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],comb_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icomb_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],heapsort(x,n,sizeof(x[0]),d_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heapsort_r(x,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],heapsort_b(x,n,sizeof(x[0]),d_rcmp_b),x[i] < x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],partialsort_r(x,n,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],heap_sort1(x,x+n,greater<double>()),x[i] < x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],heap_sort2(x,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heap_sort3(x,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort(x,x+n,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort2(x,x+n,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],partial_sort(x,x+n,x+n,greater<double>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],mergesort(x,n,sizeof(x[0]),d_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_mergesort(x,n,sizeof(x[0]),d_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],mergesort_r(x,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],mergesort_b(x,n,sizeof(x[0]),d_rcmp_b),x[i] < x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],timsort_r(x,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],merge_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pmerge_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],mmerge_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],stable_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pmergesort_r(x,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],ptimsort_r(x,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_stable_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],qsort(x,n,sizeof(x[0]),d_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_qsort_r(x,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],qsort_r(x,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],qsort_b(x,n,sizeof(x[0]),d_rcmp_b),x[i] < x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],sort(x,x+n,greater<double>()),x[i] < x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],quick_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort2(x,x+n,greater<double>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],introsort_r(x,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],intro_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pintrosort_r(x,n,sizeof(x[0]),0,d_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],psort_b(x,n,sizeof(x[0]),d_rcmp_b),x[i] < x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_sort(x,x+n,greater<double>()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_VALUES) */

#if !defined(NO_SORT_INDICE)
  cout << "# number: " << n << ",\t" << "type of compare: double" << ",\t" << "type of swap: index" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubblesort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddevensort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],poddevensort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shellsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pshellsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubble_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],foam_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddeven_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insert_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertion_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_oddeven_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shell_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_shell_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],combsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icombsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],comb_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icomb_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_b(ix,n,sizeof(ix[0]),ari_d_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,n,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort1(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort2(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort3(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+n,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort2(ix,ix+n,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+n,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_b(ix,n,sizeof(ix[0]),ari_d_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],timsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],merge_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmerge_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],mmerge_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],stable_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmergesort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],ptimsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_stable_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bsd_qsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_b(ix,n,sizeof(ix[0]),ari_d_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],quick_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort2(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],introsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],intro_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pintrosort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],psort_b(ix,n,sizeof(ix[0]),ari_d_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_INDICE) */

  delete[] x;
  delete[] ix;
}
void markbench_rsort_doubles()
{
  int n = N_POINT_SPHERE;
  double ***x0 = r3tensor(N_AVERAGE, n, 4), **x = mat(n, 4), c[] = { 0, 0, 0 }, r[] = { 1, 1, 1 }, p[3];
  vector<array<double, 4> > y(n);
  int i, j, *ix = new int[n];
  typedef ith_rcomp<array<double, 4>, 3> d3_rcomp;
  typedef ari_ith_rcomp<vector<array<double, 4> >, 3> ari_d3_rcomp;
#if defined(__BLOCKS__)
  int (^d3_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (((double **)a)[3] > ((double **)b)[3]) ? -1 : (((double **)a)[3] < ((double **)b)[3]) ? 1 : 0;
  };
  int (^ari_d3_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (x[*(int *)a][3] > x[*(int *)b][3]) ? -1 : (x[*(int *)a][3] < x[*(int *)b][3]) ? 1 : 0;
  };
#endif
  int flag = 0;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++) {
    sphere(3, N_POINT_DIMENSION, c, r, x0[j]);
    for (i=0; i<3; i++)
      p[i] = (double)random()/((1<<31)-1);
    for (i=0; i<n; i++)
      x0[j][i][3] = vec_dist(3, p, x0[j][i]);
  }

#if !defined(NO_SORT_VALUES)
  cout << "# number: " << n << ",\t" << "type of compare: doubles" << ",\t" << "type of swap: doubles" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),bubblesort_r(x[0],n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),oddevensort_r(x[0],n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),insertsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),poddevensort_r(x[0],n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),shellsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),pshellsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],bubble_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],foam_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],oddeven_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],insert_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],insertion_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],parallel_oddeven_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],shell_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],parallel_shell_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),combsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),icombsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],comb_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],icomb_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),heapsort(x[0],n,sizeof(x[0][0])*4,d3_rcmp),x[i][3] < x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),heapsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
#if defined(__BLOCKS__)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),heapsort_b(x[0],n,sizeof(x[0][0])*4,d3_rcmp_b),x[i][3] < x[i+1][3]);
#endif
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),partialsort_r(x[0],n,n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],heap_sort1(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  //do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],heap_sort2(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],heap_sort3(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],my_partial_sort(y.begin(),y.begin()+n,y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],my_partial_sort2(y.begin(),y.begin()+n,y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],partial_sort(y.begin(),y.begin()+n,y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),mergesort(x[0],n,sizeof(x[0][0])*4,d3_rcmp),x[i][3] < x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),bsd_mergesort(x[0],n,sizeof(x[0][0])*4,d3_rcmp),x[i][3] < x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),mergesort_r(x[0],n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
#if defined(__BLOCKS__)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),mergesort_b(x[0],n,sizeof(x[0][0])*4,d3_rcmp_b),x[i][3] < x[i+1][3]);
#endif
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),timsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],merge_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],pmerge_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],mmerge_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],stable_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),pmergesort_r(x[0],n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),ptimsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],parallel_stable_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),qsort(x[0],n,sizeof(x[0][0])*4,d3_rcmp),x[i][3] < x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),bsd_qsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),qsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
#if defined(__BLOCKS__)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),qsort_b(x[0],n,sizeof(x[0][0])*4,d3_rcmp_b),x[i][3] < x[i+1][3]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  //do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],quick_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],pquick_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],pquick_sort2(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),introsort_r(x[0],n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],intro_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),pintrosort_r(x[0],n,sizeof(x[0][0])*4,0,d3_rcmp_r),x[i][3] < x[i+1][3]);
#if defined(__BLOCKS__)
  do_bench_sort(memcpy(x[i],x0[c][i],sizeof(x[0][0])*4),psort_b(x[0],n,sizeof(x[0][0])*4,d3_rcmp_b),x[i][3] < x[i+1][3]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (j=0; j<4; j++) y[i][j] = x0[c][i][j],parallel_sort(y.begin(),y.begin()+n,d3_rcomp()),y[i][3] < y[i+1][3]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_VALUES) */

#if !defined(NO_SORT_INDICE)
  cout << "# number: " << n << ",\t" << "type of compare: doubles" << ",\t" << "type of swap: index" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),bubblesort_r(ix,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),oddevensort_r(ix,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),insertsort_r(ix,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),poddevensort_r(ix,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),shellsort_r(ix,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),pshellsort_r(ix,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],bubble_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],foam_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],oddeven_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],insert_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],insertion_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],parallel_oddeven_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],shell_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],parallel_shell_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),combsort_r(ix,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),icombsort_r(ix,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],comb_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],icomb_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),heapsort_r(ix,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
#if defined(__BLOCKS__)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),heapsort_b(ix,n,sizeof(ix[0]),ari_d3_rcmp_b),x[ix[i]][3] < x[ix[i+1]][3]);
#endif
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),partialsort_r(ix,n,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],heap_sort1(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  //do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],heap_sort2(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],heap_sort3(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],my_partial_sort(ix,ix+n,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],my_partial_sort2(ix,ix+n,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],partial_sort(ix,ix+n,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),mergesort_r(ix,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
#if defined(__BLOCKS__)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),mergesort_b(ix,n,sizeof(ix[0]),ari_d3_rcmp_b),x[ix[i]][3] < x[ix[i+1]][3]);
#endif
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),timsort_r(ix,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],merge_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],pmerge_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],mmerge_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],stable_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),pmergesort_r(ix,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),ptimsort_r(ix,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],parallel_stable_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),bsd_qsort_r(ix,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),qsort_r(ix,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
#if defined(__BLOCKS__)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),qsort_b(ix,n,sizeof(ix[0]),ari_d3_rcmp_b),x[ix[i]][3] < x[ix[i+1]][3]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  //do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],quick_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],pquick_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],pquick_sort2(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),introsort_r(ix,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],intro_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),pintrosort_r(ix,n,sizeof(ix[0]),x,ari_d3_rcmp_r),x[ix[i]][3] < x[ix[i+1]][3]);
#if defined(__BLOCKS__)
  do_bench_sort(memcpy(x[i],x0[c][ix[i] = i],sizeof(x[0][0])*4),psort_b(ix,n,sizeof(ix[0]),ari_d3_rcmp_b),x[ix[i]][3] < x[ix[i+1]][3]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(for (ix[i] = i, j=0; j<4; j++) y[i][j] = x0[c][i][j],parallel_sort(ix,ix+n,ari_d3_rcomp(&y)),y[ix[i]][3] < y[ix[i+1]][3]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_INDICE) */

  free_r3tensor(x0);
  free_mat(n, x);
  delete[] ix;
}
void markbench_rsort_array_indice_doubles()
{
  int n = N_POINT_SPHERE;
  double ***x0 = r3tensor(N_AVERAGE, n, 4), *x = new double[n], c[] = { 0, 0, 0 }, r[] = { 1, 1, 1 }, p[3];
  int i, j, *ix = new int[n];
  typedef ari_rcomp<double> ari_d_rcomp;
#if defined(__BLOCKS__)
  int (^ari_d_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (x[*(int *)a] > x[*(int *)b]) ? -1 : (x[*(int *)a] < x[*(int *)b]) ? 1 : 0;
  };
#endif
  int flag = 0;
  StopWatch sw;
  
  for (j=0; j<N_AVERAGE; j++) {
    sphere(3, N_POINT_DIMENSION, c, r, x0[j]);
    for (i=0; i<3; i++)
      p[i] = (double)random()/((1<<31)-1);
    for (i=0; i<n; i++)
      x0[j][i][3] = vec_dist(3, p, x0[j][i]);
  }

  cout << "# number: " << n << ",\t" << "type of compare: double" << ",\t" << "type of swap: index" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],bubblesort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],oddevensort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],insertsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],poddevensort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],shellsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],pshellsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],bubble_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],foam_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],oddeven_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],insert_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],insertion_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],parallel_oddeven_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],shell_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],parallel_shell_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],combsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],icombsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],comb_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],icomb_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],heapsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],heapsort_b(ix,n,sizeof(ix[0]),ari_d_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],partialsort_r(ix,n,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i][3],heap_sort1(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i][3],heap_sort2(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],heap_sort3(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],my_partial_sort(ix,ix+n,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],my_partial_sort2(ix,ix+n,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],partial_sort(ix,ix+n,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],mergesort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],mergesort_b(ix,n,sizeof(ix[0]),ari_d_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],timsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i][3],merge_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],pmerge_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],mmerge_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],stable_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],pmergesort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],ptimsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],parallel_stable_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],bsd_qsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],qsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],qsort_b(ix,n,sizeof(ix[0]),ari_d_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i][3],quick_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],pquick_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],pquick_sort2(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],introsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],intro_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],pintrosort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],psort_b(ix,n,sizeof(ix[0]),ari_d_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i][3],parallel_sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */

  free_r3tensor(x0);
  delete[] x;
  delete[] ix;
}
void markbench_rsort_in6()
{
  int n = N_ELEMENT;
  multi_array<struct with_in6_addr, 2> x0(extents[N_AVERAGE][n]);
  struct with_in6_addr *x = new struct with_in6_addr[n];
  int i, j, k, *ix = new int[n];
  typedef ari_rcomp<struct with_in6_addr> ari_in6_rcomp;
#if defined(__BLOCKS__)
  int (^in6_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return -1*memcmp((const void *)&((struct with_in6_addr *)a)->addr, (const void *)&((struct with_in6_addr *)b)->addr, sizeof(struct in6_addr));
  };
  int (^ari_in6_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return -1*memcmp((const void *)&x[*(int *)a].addr, (const void *)&x[*(int *)b].addr, sizeof(struct in6_addr));
  };
#endif
  int flag = 0;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++) {
      uint16_t *p = (uint16_t *)&x0[j][i].addr.s6_addr;
      for (k=0; k<(int)(sizeof(struct in6_addr)/sizeof(uint16_t)); k++)
	p[k] = (uint16_t)random();
      x0[j][i].c = i;
    }

#if !defined(NO_SORT_VALUES)
  cout << "# number: " << n << ",\t" << "type of compare: in6" << ",\t" << "type of swap: with_in6" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubblesort_r(x,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddevensort_r(x,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertsort_r(x,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],poddevensort_r(x,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shellsort_r(x,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pshellsort_r(x,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][i],bubble_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],foam_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],oddeven_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insert_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],insertion_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_oddeven_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][i],shell_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],parallel_shell_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],combsort_r(x,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icombsort_r(x,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],comb_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],icomb_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],heapsort(x,n,sizeof(x[0]),in6_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heapsort_r(x,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],heapsort_b(x,n,sizeof(x[0]),in6_rcmp_b),x[i] < x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],partialsort_r(x,n,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],heap_sort1(x,x+n,in6_rcomp()),x[i] < x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],heap_sort2(x,x+n,in6_rcomp()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],heap_sort3(x,x+n,in6_rcomp()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort(x,x+n,x+n,in6_rcomp()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],my_partial_sort2(x,x+n,x+n,in6_rcomp()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],partial_sort(x,x+n,x+n,in6_rcomp()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],mergesort(x,n,sizeof(x[0]),in6_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_mergesort(x,n,sizeof(x[0]),in6_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],mergesort_r(x,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],mergesort_b(x,n,sizeof(x[0]),in6_rcmp_b),x[i] < x[i+1]);
#endif
  do_bench_sort(x[i] = x0[c][i],timsort_r(x,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][i],merge_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pmerge_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],mmerge_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],stable_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pmergesort_r(x,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],ptimsort_r(x,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_stable_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],qsort(x,n,sizeof(x[0]),in6_rcmp),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],bsd_qsort_r(x,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],qsort_r(x,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],qsort_b(x,n,sizeof(x[0]),in6_rcmp_b),x[i] < x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
  //do_bench_sort(x[i] = x0[c][i],quick_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
  do_bench_sort(x[i] = x0[c][i],pquick_sort2(x,x+n,in6_rcomp()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],introsort_r(x,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],intro_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][i],pintrosort_r(x,n,sizeof(x[0]),0,in6_rcmp_r),x[i] < x[i+1]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][i],psort_b(x,n,sizeof(x[0]),in6_rcmp_b),x[i] < x[i+1]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][i],parallel_sort(x,x+n,in6_rcomp()),x[i] < x[i+1]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_VALUES) */

#if !defined(NO_SORT_INDICE)
  cout << "# number: " << n << ",\t" << "type of compare: in6" << ",\t" << "type of swap: index" << endl;
#if !defined(NO_DULL)
#if !defined(NO_IN_C)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubblesort_r(ix,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddevensort_r(ix,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertsort_r(ix,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],poddevensort_r(ix,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shellsort_r(ix,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pshellsort_r(ix,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
#if defined(WITH_DULLEST)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bubble_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],foam_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],oddeven_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insert_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],insertion_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_oddeven_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* defined(WITH_DULLEST) */
  do_bench_sort(x[i] = x0[c][ix[i] = i],shell_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_shell_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],combsort_r(ix,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icombsort_r(ix,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],comb_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],icomb_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_r(ix,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],heapsort_b(ix,n,sizeof(ix[0]),ari_in6_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,n,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort1(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort2(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],heap_sort3(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+n,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],my_partial_sort2(ix,ix+n,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+n,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_DULL) */

#if !defined(NO_RAPID)
#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_r(ix,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],mergesort_b(ix,n,sizeof(ix[0]),ari_in6_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
  do_bench_sort(x[i] = x0[c][ix[i] = i],timsort_r(ix,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  //do_bench_sort(x[i] = x0[c][ix[i] = i],merge_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmerge_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],mmerge_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],stable_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pmergesort_r(ix,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],ptimsort_r(ix,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_stable_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],bsd_qsort_r(ix,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_r(ix,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],qsort_b(ix,n,sizeof(ix[0]),ari_in6_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  //do_bench_sort(x[i] = x0[c][ix[i] = i],quick_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_sort(x[i] = x0[c][ix[i] = i],pquick_sort2(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],introsort_r(ix,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],intro_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */

#if !defined(NO_IN_C)
  do_bench_sort(x[i] = x0[c][ix[i] = i],pintrosort_r(ix,n,sizeof(ix[0]),x,ari_in6_rcmp_r),x[ix[i]] < x[ix[i+1]]);
#if defined(__BLOCKS__)
  do_bench_sort(x[i] = x0[c][ix[i] = i],psort_b(ix,n,sizeof(ix[0]),ari_in6_rcmp_b),x[ix[i]] < x[ix[i+1]]);
#endif
#endif /* !defined(NO_IN_C) */
#if !defined(NO_IN_CXX)
  do_bench_sort(x[i] = x0[c][ix[i] = i],parallel_sort(ix,ix+n,ari_in6_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
#endif /* !defined(NO_IN_CXX) */
#endif /* !defined(NO_RAPID) */
#endif /* !defined(NO_SORT_INDICE) */

  delete[] x;
  delete[] ix;
}


#define do_bench_partialsort(cp,op,op0,cmp) do {\
  int i, c, b, l, m;\
  accumulators::accumulator_set<double, accumulators::features<accumulators::tag::mean> > acc_ctime0, acc_utime0;\
  double ctime;\
  for (c=0; c<N_AVERAGE; c++) {\
    for (i=0; i<n; i++)\
      cp;\
    sw.start();\
    op0;\
    sw.stop();\
    acc_ctime0(sw.get_ctime());\
    acc_utime0(sw.get_utime());\
  }\
  for (b=0, l=n; l!=0; l>>=1) {\
    accumulators::accumulator_set<double, accumulators::features<accumulators::tag::mean> > acc_ctime, acc_utime;\
    m = b + (l >> 1);\
    for (c=0; c<N_AVERAGE; c++) {\
      for (i=0; i<n; i++)\
        cp;\
      sw.start();\
      op;\
      sw.stop();\
      acc_ctime(sw.get_ctime());\
      acc_utime(sw.get_utime());\
      for (i=0; i<m-1; i++)\
        if (cmp) break;\
      if (m != 0 && i != m-1) cout << "error" << endl;\
    }\
    if ((ctime = accumulators::extract::mean(acc_ctime)) < accumulators::extract::mean(acc_ctime0)) {\
      b = m;\
      l--;\
    }\
  }\
  cout << left << setw(48) << #op << "\t" << (double)m/n << ":" << setw(10) << ctime << "\t" << accumulators::extract::mean(acc_ctime0) << endl;\
} while (0)

#if !defined(DETAIL_BENCH)
void markbench_partialsort_char()
{
  int n = N_ELEMENT;
  multi_array<char, 2> x0(extents[N_AVERAGE][n]);
  char *x = new char[n];
  int i, j, *ix = new int[n];
  typedef ari_comp<char> ari_c_comp;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++)
      x0[j][i] = (char)random();

  cout << "# number: " << n << ",\t" << "type of compare: char" << ",\t" << "type of swap: char" << endl;
  do_bench_partialsort(x[i] = x0[c][i],partialsort_r(x,m,n,sizeof(x[0]),0,c_cmp_r),qsort(x,n,sizeof(x[0]),c_cmp),x[i] > x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],my_partial_sort(x,x+m,x+n),sort(x,x+n),x[i] > x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],my_partial_sort2(x,x+m,x+n),sort(x,x+n),x[i] > x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],partial_sort(x,x+m,x+n),sort(x,x+n),x[i] > x[i+1]);

  cout << "# number: " << n << ",\t" << "type of compare: char" << ",\t" << "type of swap: index" << endl;
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,m,n,sizeof(ix[0]),x,ari_c_cmp_r),qsort_r(ix,n,sizeof(ix[0]),x,ari_c_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+m,ix+n,ari_c_comp(x)),sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],my_partial_sort2(ix,ix+m,ix+n,ari_c_comp(x)),sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+m,ix+n,ari_c_comp(x)),sort(ix,ix+n,ari_c_comp(x)),x[ix[i]] > x[ix[i+1]]);

  delete[] x;
  delete[] ix;
}
void markbench_partialsort_short()
{
  int n = N_ELEMENT;
  multi_array<short, 2> x0(extents[N_AVERAGE][n]);
  short *x = new short[n];
  int i, j, *ix = new int[n];
  typedef ari_comp<short> ari_s_comp;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++)
      x0[j][i] = (short)random();

  cout << "# number: " << n << ",\t" << "type of compare: short" << ",\t" << "type of swap: short" << endl;
  do_bench_partialsort(x[i] = x0[c][i],partialsort_r(x,m,n,sizeof(x[0]),0,s_cmp_r),qsort(x,n,sizeof(x[0]),s_cmp),x[i] > x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],my_partial_sort(x,x+m,x+n),sort(x,x+n),x[i] > x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],my_partial_sort2(x,x+m,x+n),sort(x,x+n),x[i] > x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],partial_sort(x,x+m,x+n),sort(x,x+n),x[i] > x[i+1]);

  cout << "# number: " << n << ",\t" << "type of compare: short" << ",\t" << "type of swap: index" << endl;
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,m,n,sizeof(ix[0]),x,ari_s_cmp_r),qsort_r(ix,n,sizeof(ix[0]),x,ari_s_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+m,ix+n,ari_s_comp(x)),sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],my_partial_sort2(ix,ix+m,ix+n,ari_s_comp(x)),sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+m,ix+n,ari_s_comp(x)),sort(ix,ix+n,ari_s_comp(x)),x[ix[i]] > x[ix[i+1]]);

  delete[] x;
  delete[] ix;
}
void markbench_partialsort_int()
{
  int n = N_ELEMENT;
  multi_array<int, 2> x0(extents[N_AVERAGE][n]);
  int *x = new int[n];
  int i, j, *ix = new int[n];
  typedef ari_comp<int> ari_i_comp;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++)
      x0[j][i] = (int)random();

  cout << "# number: " << n << ",\t" << "type of compare: int" << ",\t" << "type of swap: int" << endl;
  do_bench_partialsort(x[i] = x0[c][i],partialsort_r(x,m,n,sizeof(x[0]),0,i_cmp_r),qsort(x,n,sizeof(x[0]),i_cmp),x[i] > x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],my_partial_sort(x,x+m,x+n),sort(x,x+n),x[i] > x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],my_partial_sort2(x,x+m,x+n),sort(x,x+n),x[i] > x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],partial_sort(x,x+m,x+n),sort(x,x+n),x[i] > x[i+1]);

  cout << "# number: " << n << ",\t" << "type of compare: int" << ",\t" << "type of swap: index" << endl;
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,m,n,sizeof(ix[0]),x,ari_i_cmp_r),qsort_r(ix,n,sizeof(ix[0]),x,ari_i_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+m,ix+n,ari_i_comp(x)),sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],my_partial_sort2(ix,ix+m,ix+n,ari_i_comp(x)),sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+m,ix+n,ari_i_comp(x)),sort(ix,ix+n,ari_i_comp(x)),x[ix[i]] > x[ix[i+1]]);

  delete[] x;
  delete[] ix;
}
void markbench_partialsort_double()
{
  int n = N_ELEMENT;
  multi_array<double, 2> x0(extents[N_AVERAGE][n]);
  double *x = new double[n];
  int i, j, *ix = new int[n];
  typedef ari_comp<double> ari_d_comp;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++)
      x0[j][i] = (double)random()/((1<<31)-1);

  cout << "# number: " << n << ",\t" << "type of compare: double" << ",\t" << "type of swap: double" << endl;
  do_bench_partialsort(x[i] = x0[c][i],partialsort_r(x,m,n,sizeof(x[0]),0,d_cmp_r),qsort(x,n,sizeof(x[0]),d_cmp),x[i] > x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],my_partial_sort(x,x+m,x+n),sort(x,x+n),x[i] > x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],my_partial_sort2(x,x+m,x+n),sort(x,x+n),x[i] > x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],partial_sort(x,x+m,x+n),sort(x,x+n),x[i] > x[i+1]);

  cout << "# number: " << n << ",\t" << "type of compare: double" << ",\t" << "type of swap: index" << endl;
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,m,n,sizeof(ix[0]),x,ari_d_cmp_r),qsort_r(ix,n,sizeof(ix[0]),x,ari_d_cmp_r),x[ix[i]] > x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+m,ix+n,ari_d_comp(x)),sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],my_partial_sort2(ix,ix+m,ix+n,ari_d_comp(x)),sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+m,ix+n,ari_d_comp(x)),sort(ix,ix+n,ari_d_comp(x)),x[ix[i]] > x[ix[i+1]]);

  delete[] x;
  delete[] ix;
}

void markbench_rpartialsort_char()
{
  int n = N_ELEMENT;
  multi_array<char, 2> x0(extents[N_AVERAGE][n]);
  char *x = new char[n];
  int i, j, *ix = new int[n];
  typedef ari_rcomp<char> ari_c_rcomp;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++)
      x0[j][i] = (char)random();

  cout << "# number: " << n << ",\t" << "type of compare: char" << ",\t" << "type of swap: char" << endl;
  do_bench_partialsort(x[i] = x0[c][i],partialsort_r(x,m,n,sizeof(x[0]),0,c_rcmp_r),qsort(x,n,sizeof(x[0]),c_rcmp),x[i] < x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],my_partial_sort(x,x+m,x+n,greater<char>()),sort(x,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],my_partial_sort2(x,x+m,x+n,greater<char>()),sort(x,x+n,greater<char>()),x[i] < x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],partial_sort(x,x+m,x+n,greater<char>()),sort(x,x+n,greater<char>()),x[i] < x[i+1]);

  cout << "# number: " << n << ",\t" << "type of compare: char" << ",\t" << "type of swap: index" << endl;
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,m,n,sizeof(ix[0]),x,ari_c_rcmp_r),qsort_r(ix,n,sizeof(ix[0]),x,ari_c_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+m,ix+n,ari_c_rcomp(x)),sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],my_partial_sort2(ix,ix+m,ix+n,ari_c_rcomp(x)),sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+m,ix+n,ari_c_rcomp(x)),sort(ix,ix+n,ari_c_rcomp(x)),x[ix[i]] < x[ix[i+1]]);

  delete[] x;
  delete[] ix;
}
void markbench_rpartialsort_short()
{
  int n = N_ELEMENT;
  multi_array<short, 2> x0(extents[N_AVERAGE][n]);
  short *x = new short[n];
  int i, j, *ix = new int[n];
  typedef ari_rcomp<short> ari_s_rcomp;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++)
      x0[j][i] = (short)random();

  cout << "# number: " << n << ",\t" << "type of compare: short" << ",\t" << "type of swap: short" << endl;
  do_bench_partialsort(x[i] = x0[c][i],partialsort_r(x,m,n,sizeof(x[0]),0,s_rcmp_r),qsort(x,n,sizeof(x[0]),s_rcmp),x[i] < x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],my_partial_sort(x,x+m,x+n,greater<short>()),sort(x,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],my_partial_sort2(x,x+m,x+n,greater<short>()),sort(x,x+n,greater<short>()),x[i] < x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],partial_sort(x,x+m,x+n,greater<short>()),sort(x,x+n,greater<short>()),x[i] < x[i+1]);

  cout << "# number: " << n << ",\t" << "type of compare: short" << ",\t" << "type of swap: index" << endl;
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,m,n,sizeof(ix[0]),x,ari_s_rcmp_r),qsort_r(ix,n,sizeof(ix[0]),x,ari_s_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+m,ix+n,ari_s_rcomp(x)),sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],my_partial_sort2(ix,ix+m,ix+n,ari_s_rcomp(x)),sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+m,ix+n,ari_s_rcomp(x)),sort(ix,ix+n,ari_s_rcomp(x)),x[ix[i]] < x[ix[i+1]]);

  delete[] x;
  delete[] ix;
}
void markbench_rpartialsort_int()
{
  int n = N_ELEMENT;
  multi_array<int, 2> x0(extents[N_AVERAGE][n]);
  int *x = new int[n];
  int i, j, *ix = new int[n];
  typedef ari_rcomp<int> ari_i_rcomp;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++)
      x0[j][i] = (int)random();

  cout << "# number: " << n << ",\t" << "type of compare: int" << ",\t" << "type of swap: int" << endl;
  do_bench_partialsort(x[i] = x0[c][i],partialsort_r(x,m,n,sizeof(x[0]),0,i_rcmp_r),qsort(x,n,sizeof(x[0]),i_rcmp),x[i] < x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],my_partial_sort(x,x+m,x+n,greater<int>()),sort(x,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],my_partial_sort2(x,x+m,x+n,greater<int>()),sort(x,x+n,greater<int>()),x[i] < x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],partial_sort(x,x+m,x+n,greater<int>()),sort(x,x+n,greater<int>()),x[i] < x[i+1]);

  cout << "# number: " << n << ",\t" << "type of compare: int" << ",\t" << "type of swap: index" << endl;
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,m,n,sizeof(ix[0]),x,ari_i_rcmp_r),qsort_r(ix,n,sizeof(ix[0]),x,ari_i_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+m,ix+n,ari_i_rcomp(x)),sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],my_partial_sort2(ix,ix+m,ix+n,ari_i_rcomp(x)),sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+m,ix+n,ari_i_rcomp(x)),sort(ix,ix+n,ari_i_rcomp(x)),x[ix[i]] < x[ix[i+1]]);

  delete[] x;
  delete[] ix;
}
void markbench_rpartialsort_double()
{
  int n = N_ELEMENT;
  multi_array<double, 2> x0(extents[N_AVERAGE][n]);
  double *x = new double[n];
  int i, j, *ix = new int[n];
  typedef ari_rcomp<double> ari_d_rcomp;
  StopWatch sw;

  for (j=0; j<N_AVERAGE; j++)
    for (i=0; i<n; i++)
      x0[j][i] = (double)random()/((1<<31)-1);

  cout << "# number: " << n << ",\t" << "type of compare: double" << ",\t" << "type of swap: double" << endl;
  do_bench_partialsort(x[i] = x0[c][i],partialsort_r(x,m,n,sizeof(x[0]),0,d_rcmp_r),qsort(x,n,sizeof(x[0]),d_rcmp),x[i] < x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],my_partial_sort(x,x+m,x+n,greater<double>()),sort(x,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],my_partial_sort2(x,x+m,x+n,greater<double>()),sort(x,x+n,greater<double>()),x[i] < x[i+1]);
  do_bench_partialsort(x[i] = x0[c][i],partial_sort(x,x+m,x+n,greater<double>()),sort(x,x+n,greater<double>()),x[i] < x[i+1]);

  cout << "# number: " << n << ",\t" << "type of compare: double" << ",\t" << "type of swap: index" << endl;
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],partialsort_r(ix,m,n,sizeof(ix[0]),x,ari_d_rcmp_r),qsort_r(ix,n,sizeof(ix[0]),x,ari_d_rcmp_r),x[ix[i]] < x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],my_partial_sort(ix,ix+m,ix+n,ari_d_rcomp(x)),sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],my_partial_sort2(ix,ix+m,ix+n,ari_d_rcomp(x)),sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);
  do_bench_partialsort(x[i] = x0[c][ix[i] = i],partial_sort(ix,ix+m,ix+n,ari_d_rcomp(x)),sort(ix,ix+n,ari_d_rcomp(x)),x[ix[i]] < x[ix[i+1]]);

  delete[] x;
  delete[] ix;
}
#endif /* !defined(DETAIL_BENCH) */


#define do_bench_search(op,cmp) do {\
  int c;\
  accumulators::accumulator_set<double, accumulators::features<accumulators::tag::mean> > acc_ctime, acc_utime;\
  for (c=0; c<N_AVERAGE; c++) {\
    sw.start();\
    op;\
    sw.stop();\
    acc_ctime(sw.get_ctime());\
    acc_utime(sw.get_utime());\
  }\
  cout << ((cmp)?"":"!") << left << setw(48) << #op << "\t" << setw(10) << accumulators::extract::mean(acc_ctime) << "\t" << accumulators::extract::mean(acc_utime) << endl;\
} while (0)

#define ARI_T_NE_R(T) ((*(T *)k < ((T *)thunk)[*(int *)i]) ? -1 : (*(T *)k > ((T *)thunk)[*(int *)i]) ? 1 : 0)
int ari_c_ne_r(void *thunk, const void *k, const void *i) { return ARI_T_NE_R(char); }
int ari_s_ne_r(void *thunk, const void *k, const void *i) { return ARI_T_NE_R(short); }
int ari_i_ne_r(void *thunk, const void *k, const void *i) { return ARI_T_NE_R(int); }
int ari_d_ne_r(void *thunk, const void *k, const void *i) { return ARI_T_NE_R(double); }

template <typename T>
struct ari_pred { ari_pred(T *p, T *q):v(p),t(q){} bool operator()(const int &a) { return v[a] == *t; } private: T *v, *t; };

template <typename T>
struct ari_komp { ari_komp(T *p):v(p){} bool operator()(const int &i, const T &k) { return v[i] < k; } private: T *v; };

template <typename T>
struct ari_kowp { ari_kowp(T *p):v(p){} bool operator()(const T &k, const int &i) { return k < v[i]; } private: T *v; };

#define ARI_T_RE_R(T) ((*(T *)k > ((T *)thunk)[*(int *)i]) ? -1 : (*(T *)k < ((T *)thunk)[*(int *)i]) ? 1 : 0)
int ari_c_re_r(void *thunk, const void *k, const void *i) { return ARI_T_RE_R(char); }
int ari_s_re_r(void *thunk, const void *k, const void *i) { return ARI_T_RE_R(short); }
int ari_i_re_r(void *thunk, const void *k, const void *i) { return ARI_T_RE_R(int); }
int ari_d_re_r(void *thunk, const void *k, const void *i) { return ARI_T_RE_R(double); }

template <typename T>
struct ari_rpred { ari_rpred(T *p, T *q):v(p),t(q){} bool operator()(const int &a) { return v[a] == *t; } private: T *v, *t; };

template <typename T>
struct ari_rkomp { ari_rkomp(T *p):v(p){} bool operator()(const int &i, const T &k) { return v[i] > k; } private: T *v; };

template <typename T>
struct ari_rkowp { ari_rkowp(T *p):v(p){} bool operator()(const T &k, const int &i) { return k > v[i]; } private: T *v; };

#if !defined(DETAIL_BENCH)
void markbench_search_char()
{
  int n = N_ELEMENT;
  size_t m = n;
  char *x0 = new char[n], *x = new char[n], t, *f;
  int i, *ix = new int[n], *g;
  bool b;
  typedef ari_comp<char> ari_c_comp;
  typedef ari_pred<char> ari_c_pred;
  typedef ari_komp<char> ari_c_komp;
  typedef ari_kowp<char> ari_c_kowp;
#if defined(__BLOCKS__)
  int (^c_cmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(char *)a < *(char *)b) ? -1 : (*(char *)a > *(char *)b) ? 1 : 0;
  };
  int (^ari_c_ne_b)(const void *k, const void *i) = ^(const void *k, const void *i) {
    return (*(char *)k < x0[*(int *)i]) ? -1 : (*(char *)k > x0[*(int *)i]) ? 1 : 0;
  };
#endif
  StopWatch sw;

  for (i=0; i<n; i++)
    x[i] = x0[ix[i] = i] = (char)random();
  t = x0[random()%n];
  sort(x,x+n);
  sort(ix,ix+n,ari_c_comp(x0));

  cout << "# number: " << n << ",\t" << "type of compare: char" << ",\t" << "type of search: char" << endl;
  do_bench_search(f=(char *)lfind(&t,x0,&m,sizeof(x0[0]),c_cmp), f && *f == t);
  do_bench_search(f=(char *)lfind_r(&t,x0,&m,sizeof(x0[0]),0,c_cmp_r), f && *f == t);
  do_bench_search(f=find(x0,x0+n,t), *f == t);
  do_bench_search(f=(char *)bsearch(&t,x,n,sizeof(x[0]),c_cmp), f && *f == t);
  do_bench_search(f=(char *)lbsearch_r(&t,x,n,sizeof(x[0]),0,c_cmp_r), f && *f == t);
  do_bench_search(f=(char *)ubsearch_r(&t,x,n,sizeof(x[0]),0,c_cmp_r), f && *(f-1) == t);
  do_bench_search(f=(char *)bsearch_r(&t,x,n,sizeof(x[0]),0,c_cmp_r), f && *f == t);
#if defined(__BLOCKS__)
  do_bench_search(f=(char *)bsearch_b(&t,x,n,sizeof(x[0]),c_cmp_b), f && *f == t);
#endif
  do_bench_search(f=lower_bound(x,x+n,t), f && *f == t);
  do_bench_search(f=upper_bound(x,x+n,t), f && *(f-1) == t);
  do_bench_search(b=binary_search(x,x+n,t), b);

  cout << "# number: " << n << ",\t" << "type of compare: char" << ",\t" << "type of search: int" << endl;
  do_bench_search(g=find_if(ix,ix+n,ari_c_pred(x0,&t)), g && x0[*g] == t);
  do_bench_search(g=(int *)lfind_r(&t,ix,&m,sizeof(ix[0]),x0,ari_c_ne_r), g && x0[*g] == t);
  do_bench_search(g=(int *)lbsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_c_ne_r), g && x0[*g] == t);
  do_bench_search(g=(int *)ubsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_c_ne_r), g && x0[*(g-1)] == t);
  do_bench_search(g=(int *)bsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_c_ne_r), g && x0[*g] == t);
#if defined(__BLOCKS__)
  do_bench_search(g=(int *)bsearch_b(&t,ix,n,sizeof(ix[0]),ari_c_ne_b), g && x0[*g] == t);
#endif
  do_bench_search(g=lower_bound(ix,ix+n,t,ari_c_komp(x0)), g && x0[*g] == t);
  do_bench_search(g=upper_bound(ix,ix+n,t,ari_c_kowp(x0)), g && x0[*(g-1)] == t);

  delete[] x;
  delete[] x0;
  delete[] ix;
}
void markbench_search_short()
{
  int n = N_ELEMENT;
  size_t m = n;
  short *x0 = new short[n], *x = new short[n], t, *f;
  int i, *ix = new int[n], *g;
  bool b;
  typedef ari_comp<short> ari_s_comp;
  typedef ari_pred<short> ari_s_pred;
  typedef ari_komp<short> ari_s_komp;
  typedef ari_kowp<short> ari_s_kowp;
#if defined(__BLOCKS__)
  int (^s_cmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(short *)a < *(short *)b) ? -1 : (*(short *)a > *(short *)b) ? 1 : 0;
  };
  int (^ari_s_ne_b)(const void *k, const void *i) = ^(const void *k, const void *i) {
    return (*(short *)k < x0[*(int *)i]) ? -1 : (*(short *)k > x0[*(int *)i]) ? 1 : 0;
  };
#endif
  StopWatch sw;

  for (i=0; i<n; i++)
    x[i] = x0[ix[i] = i] = (short)random();
  t = x0[random()%n];
  sort(x,x+n);
  sort(ix,ix+n,ari_s_comp(x0));

  cout << "# number: " << n << ",\t" << "type of compare: short" << ",\t" << "type of search: short" << endl;
  do_bench_search(f=(short *)lfind(&t,x0,&m,sizeof(x0[0]),s_cmp), f && *f == t);
  do_bench_search(f=(short *)lfind_r(&t,x0,&m,sizeof(x0[0]),0,s_cmp_r), f && *f == t);
  do_bench_search(f=find(x0,x0+n,t), *f == t);
  do_bench_search(f=(short *)bsearch(&t,x,n,sizeof(x[0]),s_cmp), f && *f == t);
  do_bench_search(f=(short *)lbsearch_r(&t,x,n,sizeof(x[0]),0,s_cmp_r), f && *f == t);
  do_bench_search(f=(short *)ubsearch_r(&t,x,n,sizeof(x[0]),0,s_cmp_r), f && *(f-1) == t);
  do_bench_search(f=(short *)bsearch_r(&t,x,n,sizeof(x[0]),0,s_cmp_r), f && *f == t);
#if defined(__BLOCKS__)
  do_bench_search(f=(short *)bsearch_b(&t,x,n,sizeof(x[0]),s_cmp_b), f && *f == t);
#endif
  do_bench_search(f=lower_bound(x,x+n,t), f && *f == t);
  do_bench_search(f=upper_bound(x,x+n,t), f && *(f-1) == t);
  do_bench_search(b=binary_search(x,x+n,t), b);

  cout << "# number: " << n << ",\t" << "type of compare: short" << ",\t" << "type of search: int" << endl;
  do_bench_search(g=find_if(ix,ix+n,ari_s_pred(x0,&t)), g && x0[*g] == t);
  do_bench_search(g=(int *)lfind_r(&t,ix,&m,sizeof(ix[0]),x0,ari_s_ne_r), g && x0[*g] == t);
  do_bench_search(g=(int *)lbsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_s_ne_r), g && x0[*g] == t);
  do_bench_search(g=(int *)ubsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_s_ne_r), g && x0[*(g-1)] == t);
  do_bench_search(g=(int *)bsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_s_ne_r), g && x0[*g] == t);
#if defined(__BLOCKS__)
  do_bench_search(g=(int *)bsearch_b(&t,ix,n,sizeof(ix[0]),ari_s_ne_b), g && x0[*g] == t);
#endif
  do_bench_search(g=lower_bound(ix,ix+n,t,ari_s_komp(x0)), g && x0[*g] == t);
  do_bench_search(g=upper_bound(ix,ix+n,t,ari_s_kowp(x0)), g && x0[*(g-1)] == t);

  delete[] x;
  delete[] x0;
  delete[] ix;
}
void markbench_search_int()
{
  int n = N_ELEMENT;
  size_t m = n;
  int *x0 = new int[n], *x = new int[n], t, *f;
  int i, *ix = new int[n], *g;
  bool b;
  typedef ari_comp<int> ari_i_comp;
  typedef ari_pred<int> ari_i_pred;
  typedef ari_komp<int> ari_i_komp;
  typedef ari_kowp<int> ari_i_kowp;
#if defined(__BLOCKS__)
  int (^i_cmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(int *)a < *(int *)b) ? -1 : (*(int *)a > *(int *)b) ? 1 : 0;
  };
  int (^ari_i_ne_b)(const void *k, const void *i) = ^(const void *k, const void *i) {
    return (*(int *)k < x0[*(int *)i]) ? -1 : (*(int *)k > x0[*(int *)i]) ? 1 : 0;
  };
#endif
  StopWatch sw;

  for (i=0; i<n; i++)
    x[i] = x0[ix[i] = i] = (int)random();
  t = x0[random()%n];
  sort(x,x+n);
  sort(ix,ix+n,ari_i_comp(x0));

  cout << "# number: " << n << ",\t" << "type of compare: int" << ",\t" << "type of search: int" << endl;
  do_bench_search(f=(int *)lfind(&t,x0,&m,sizeof(x0[0]),i_cmp), f && *f == t);
  do_bench_search(f=(int *)lfind_r(&t,x0,&m,sizeof(x0[0]),0,i_cmp_r), f && *f == t);
  do_bench_search(f=find(x0,x0+n,t), *f == t);
  do_bench_search(f=(int *)bsearch(&t,x,n,sizeof(x[0]),i_cmp), f && *f == t);
  do_bench_search(f=(int *)lbsearch_r(&t,x,n,sizeof(x[0]),0,i_cmp_r), f && *f == t);
  do_bench_search(f=(int *)ubsearch_r(&t,x,n,sizeof(x[0]),0,i_cmp_r), f && *(f-1) == t);
  do_bench_search(f=(int *)bsearch_r(&t,x,n,sizeof(x[0]),0,i_cmp_r), f && *f == t);
#if defined(__BLOCKS__)
  do_bench_search(f=(int *)bsearch_b(&t,x,n,sizeof(x[0]),i_cmp_b), f && *f == t);
#endif
  do_bench_search(f=lower_bound(x,x+n,t), f && *f == t);
  do_bench_search(f=upper_bound(x,x+n,t), f && *(f-1) == t);
  do_bench_search(b=binary_search(x,x+n,t), b);

  cout << "# number: " << n << ",\t" << "type of compare: int" << ",\t" << "type of search: int" << endl;
  do_bench_search(g=find_if(ix,ix+n,ari_i_pred(x0,&t)), g && x0[*g] == t);
  do_bench_search(g=(int *)lfind_r(&t,ix,&m,sizeof(ix[0]),x0,ari_i_ne_r), g && x0[*g] == t);
  do_bench_search(g=(int *)lbsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_i_ne_r), g && x0[*g] == t);
  do_bench_search(g=(int *)ubsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_i_ne_r), g && x0[*(g-1)] == t);
  do_bench_search(g=(int *)bsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_i_ne_r), g && x0[*g] == t);
#if defined(__BLOCKS__)
  do_bench_search(g=(int *)bsearch_b(&t,ix,n,sizeof(ix[0]),ari_i_ne_b), g && x0[*g] == t);
#endif
  do_bench_search(g=lower_bound(ix,ix+n,t,ari_i_komp(x0)), g && x0[*g] == t);
  do_bench_search(g=upper_bound(ix,ix+n,t,ari_i_kowp(x0)), g && x0[*(g-1)] == t);

  delete[] x;
  delete[] x0;
  delete[] ix;
}
void markbench_search_double()
{
  int n = N_ELEMENT;
  size_t m = n;
  double *x0 = new double[n], *x = new double[n], t, *f;
  int i, *ix = new int[n], *g;
  bool b;
  typedef ari_comp<double> ari_d_comp;
  typedef ari_pred<double> ari_d_pred;
  typedef ari_komp<double> ari_d_komp;
  typedef ari_kowp<double> ari_d_kowp;
#if defined(__BLOCKS__)
  int (^d_cmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(double *)a < *(double *)b) ? -1 : (*(double *)a > *(double *)b) ? 1 : 0;
  };
  int (^ari_d_ne_b)(const void *k, const void *i) = ^(const void *k, const void *i) {
    return (*(double *)k < x0[*(int *)i]) ? -1 : (*(double *)k > x0[*(int *)i]) ? 1 : 0;
  };
#endif
  StopWatch sw;

  for (i=0; i<n; i++)
    x[i] = x0[ix[i] = i] = (double)random()/((1<<31)-1);
  t = x0[random()%n];
  sort(x,x+n);
  sort(ix,ix+n,ari_d_comp(x0));

  cout << "# number: " << n << ",\t" << "type of compare: double" << ",\t" << "type of search: double" << endl;
  do_bench_search(f=(double *)lfind(&t,x0,&m,sizeof(x0[0]),d_cmp), f && *f == t);
  do_bench_search(f=(double *)lfind_r(&t,x0,&m,sizeof(x0[0]),0,d_cmp_r), f && *f == t);
  do_bench_search(f=find(x0,x0+n,t), *f == t);
  do_bench_search(f=(double *)bsearch(&t,x,n,sizeof(x[0]),d_cmp), f && *f == t);
  do_bench_search(f=(double *)lbsearch_r(&t,x,n,sizeof(x[0]),0,d_cmp_r), f && *f == t);
  do_bench_search(f=(double *)ubsearch_r(&t,x,n,sizeof(x[0]),0,d_cmp_r), f && *(f-1) == t);
  do_bench_search(f=(double *)bsearch_r(&t,x,n,sizeof(x[0]),0,d_cmp_r), f && *f == t);
#if defined(__BLOCKS__)
  do_bench_search(f=(double *)bsearch_b(&t,x,n,sizeof(x[0]),d_cmp_b), f && *f == t);
#endif
  do_bench_search(f=lower_bound(x,x+n,t), f && *f == t);
  do_bench_search(f=upper_bound(x,x+n,t), f && *(f-1) == t);
  do_bench_search(b=binary_search(x,x+n,t), b);

  cout << "# number: " << n << ",\t" << "type of compare: double" << ",\t" << "type of search: int" << endl;
  do_bench_search(g=find_if(ix,ix+n,ari_d_pred(x0,&t)), g && x0[*g] == t);
  do_bench_search(g=(int *)lfind_r(&t,ix,&m,sizeof(ix[0]),x0,ari_d_ne_r), g && x0[*g] == t);
  do_bench_search(g=(int *)lbsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_d_ne_r), g && x0[*g] == t);
  do_bench_search(g=(int *)ubsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_d_ne_r), g && x0[*(g-1)] == t);
  do_bench_search(g=(int *)bsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_d_ne_r), g && x0[*g] == t);
#if defined(__BLOCKS__)
  do_bench_search(g=(int *)bsearch_b(&t,ix,n,sizeof(ix[0]),ari_d_ne_b), g && x0[*g] == t);
#endif
  do_bench_search(g=lower_bound(ix,ix+n,t,ari_d_komp(x0)), g && x0[*g] == t);
  do_bench_search(g=upper_bound(ix,ix+n,t,ari_d_kowp(x0)), g && x0[*(g-1)] == t);

  delete[] x;
  delete[] x0;
  delete[] ix;
}

void markbench_rsearch_char()
{
  int n = N_ELEMENT;
  size_t m = n;
  char *x0 = new char[n], *x = new char[n], t, *f;
  int i, *ix = new int[n], *g;
  bool b;
  typedef ari_rcomp<char> ari_c_rcomp;
  typedef ari_rpred<char> ari_c_rpred;
  typedef ari_rkomp<char> ari_c_rkomp;
  typedef ari_rkowp<char> ari_c_rkowp;
#if defined(__BLOCKS__)
  int (^c_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(char *)a > *(char *)b) ? -1 : (*(char *)a < *(char *)b) ? 1 : 0;
  };
  int (^ari_c_re_b)(const void *k, const void *i) = ^(const void *k, const void *i) {
    return (*(char *)k > x0[*(int *)i]) ? -1 : (*(char *)k < x0[*(int *)i]) ? 1 : 0;
  };
#endif
  StopWatch sw;

  for (i=0; i<n; i++)
    x[i] = x0[ix[i] = i] = (char)random();
  t = x0[random()%n];
  sort(x,x+n,greater<char>());
  sort(ix,ix+n,ari_c_rcomp(x0));

  cout << "# number: " << n << ",\t" << "type of compare: char" << ",\t" << "type of search: char" << endl;
  do_bench_search(f=(char *)lfind(&t,x0,&m,sizeof(x0[0]),c_rcmp), f && *f == t);
  do_bench_search(f=(char *)lfind_r(&t,x0,&m,sizeof(x0[0]),0,c_rcmp_r), f && *f == t);
  do_bench_search(f=find(x0,x0+n,t), *f == t);
  do_bench_search(f=(char *)bsearch(&t,x,n,sizeof(x[0]),c_rcmp), f && *f == t);
  do_bench_search(f=(char *)lbsearch_r(&t,x,n,sizeof(x[0]),0,c_rcmp_r), f && *f == t);
  do_bench_search(f=(char *)ubsearch_r(&t,x,n,sizeof(x[0]),0,c_rcmp_r), f && *(f-1) == t);
  do_bench_search(f=(char *)bsearch_r(&t,x,n,sizeof(x[0]),0,c_rcmp_r), f && *f == t);
#if defined(__BLOCKS__)
  do_bench_search(f=(char *)bsearch_b(&t,x,n,sizeof(x[0]),c_rcmp_b), f && *f == t);
#endif
  do_bench_search(f=lower_bound(x,x+n,t,greater<char>()), f && *f == t);
  do_bench_search(f=upper_bound(x,x+n,t,greater<char>()), f && *(f-1) == t);
  do_bench_search(b=binary_search(x,x+n,t,greater<char>()), b);

  cout << "# number: " << n << ",\t" << "type of compare: char" << ",\t" << "type of search: int" << endl;
  do_bench_search(g=find_if(ix,ix+n,ari_c_rpred(x0,&t)), g && x0[*g] == t);
  do_bench_search(g=(int *)lfind_r(&t,ix,&m,sizeof(ix[0]),x0,ari_c_re_r), g && x0[*g] == t);
  do_bench_search(g=(int *)lbsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_c_re_r), g && x0[*g] == t);
  do_bench_search(g=(int *)ubsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_c_re_r), g && x0[*(g-1)] == t);
  do_bench_search(g=(int *)bsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_c_re_r), g && x0[*g] == t);
#if defined(__BLOCKS__)
  do_bench_search(g=(int *)bsearch_b(&t,ix,n,sizeof(ix[0]),ari_c_re_b), g && x0[*g] == t);
#endif
  do_bench_search(g=lower_bound(ix,ix+n,t,ari_c_rkomp(x0)), g && x0[*g] == t);
  do_bench_search(g=upper_bound(ix,ix+n,t,ari_c_rkowp(x0)), g && x0[*(g-1)] == t);

  delete[] x;
  delete[] x0;
  delete[] ix;
}
void markbench_rsearch_short()
{
  int n = N_ELEMENT;
  size_t m = n;
  short *x0 = new short[n], *x = new short[n], t, *f;
  int i, *ix = new int[n], *g;
  bool b;
  typedef ari_rcomp<short> ari_s_rcomp;
  typedef ari_rpred<short> ari_s_rpred;
  typedef ari_rkomp<short> ari_s_rkomp;
  typedef ari_rkowp<short> ari_s_rkowp;
#if defined(__BLOCKS__)
  int (^s_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(short *)a > *(short *)b) ? -1 : (*(short *)a < *(short *)b) ? 1 : 0;
  };
  int (^ari_s_re_b)(const void *k, const void *i) = ^(const void *k, const void *i) {
    return (*(short *)k > x0[*(int *)i]) ? -1 : (*(short *)k < x0[*(int *)i]) ? 1 : 0;
  };
#endif
  StopWatch sw;

  for (i=0; i<n; i++)
    x[i] = x0[ix[i] = i] = (short)random();
  t = x0[random()%n];
  sort(x,x+n,greater<short>());
  sort(ix,ix+n,ari_s_rcomp(x0));

  cout << "# number: " << n << ",\t" << "type of compare: short" << ",\t" << "type of search: short" << endl;
  do_bench_search(f=(short *)lfind(&t,x0,&m,sizeof(x0[0]),s_rcmp), f && *f == t);
  do_bench_search(f=(short *)lfind_r(&t,x0,&m,sizeof(x0[0]),0,s_rcmp_r), f && *f == t);
  do_bench_search(f=find(x0,x0+n,t), *f == t);
  do_bench_search(f=(short *)bsearch(&t,x,n,sizeof(x[0]),s_rcmp), f && *f == t);
  do_bench_search(f=(short *)lbsearch_r(&t,x,n,sizeof(x[0]),0,s_rcmp_r), f && *f == t);
  do_bench_search(f=(short *)ubsearch_r(&t,x,n,sizeof(x[0]),0,s_rcmp_r), f && *(f-1) == t);
  do_bench_search(f=(short *)bsearch_r(&t,x,n,sizeof(x[0]),0,s_rcmp_r), f && *f == t);
#if defined(__BLOCKS__)
  do_bench_search(f=(short *)bsearch_b(&t,x,n,sizeof(x[0]),s_rcmp_b), f && *f == t);
#endif
  do_bench_search(f=lower_bound(x,x+n,t,greater<short>()), f && *f == t);
  do_bench_search(f=upper_bound(x,x+n,t,greater<short>()), f && *(f-1) == t);
  do_bench_search(b=binary_search(x,x+n,t,greater<short>()), b);

  cout << "# number: " << n << ",\t" << "type of compare: short" << ",\t" << "type of search: int" << endl;
  do_bench_search(g=find_if(ix,ix+n,ari_s_rpred(x0,&t)), g && x0[*g] == t);
  do_bench_search(g=(int *)lfind_r(&t,ix,&m,sizeof(ix[0]),x0,ari_s_re_r), g && x0[*g] == t);
  do_bench_search(g=(int *)lbsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_s_re_r), g && x0[*g] == t);
  do_bench_search(g=(int *)ubsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_s_re_r), g && x0[*(g-1)] == t);
  do_bench_search(g=(int *)bsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_s_re_r), g && x0[*g] == t);
#if defined(__BLOCKS__)
  do_bench_search(g=(int *)bsearch_b(&t,ix,n,sizeof(ix[0]),ari_s_re_b), g && x0[*g] == t);
#endif
  do_bench_search(g=lower_bound(ix,ix+n,t,ari_s_rkomp(x0)), g && x0[*g] == t);
  do_bench_search(g=upper_bound(ix,ix+n,t,ari_s_rkowp(x0)), g && x0[*(g-1)] == t);

  delete[] x;
  delete[] x0;
  delete[] ix;
}
void markbench_rsearch_int()
{
  int n = N_ELEMENT;
  size_t m = n;
  int *x0 = new int[n], *x = new int[n], t, *f;
  int i, *ix = new int[n], *g;
  bool b;
  typedef ari_rcomp<int> ari_i_rcomp;
  typedef ari_rpred<int> ari_i_rpred;
  typedef ari_rkomp<int> ari_i_rkomp;
  typedef ari_rkowp<int> ari_i_rkowp;
#if defined(__BLOCKS__)
  int (^i_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(int *)a > *(int *)b) ? -1 : (*(int *)a < *(int *)b) ? 1 : 0;
  };
  int (^ari_i_re_b)(const void *k, const void *i) = ^(const void *k, const void *i) {
    return (*(int *)k > x0[*(int *)i]) ? -1 : (*(int *)k < x0[*(int *)i]) ? 1 : 0;
  };
#endif
  StopWatch sw;

  for (i=0; i<n; i++)
    x[i] = x0[ix[i] = i] = (int)random();
  t = x0[random()%n];
  sort(x,x+n,greater<int>());
  sort(ix,ix+n,ari_i_rcomp(x0));

  cout << "# number: " << n << ",\t" << "type of compare: int" << ",\t" << "type of search: int" << endl;
  do_bench_search(f=(int *)lfind(&t,x0,&m,sizeof(x0[0]),i_rcmp), f && *f == t);
  do_bench_search(f=(int *)lfind_r(&t,x0,&m,sizeof(x0[0]),0,i_rcmp_r), f && *f == t);
  do_bench_search(f=find(x0,x0+n,t), *f == t);
  do_bench_search(f=(int *)bsearch(&t,x,n,sizeof(x[0]),i_rcmp), f && *f == t);
  do_bench_search(f=(int *)lbsearch_r(&t,x,n,sizeof(x[0]),0,i_rcmp_r), f && *f == t);
  do_bench_search(f=(int *)ubsearch_r(&t,x,n,sizeof(x[0]),0,i_rcmp_r), f && *(f-1) == t);
  do_bench_search(f=(int *)bsearch_r(&t,x,n,sizeof(x[0]),0,i_rcmp_r), f && *f == t);
#if defined(__BLOCKS__)
  do_bench_search(f=(int *)bsearch_b(&t,x,n,sizeof(x[0]),i_rcmp_b), f && *f == t);
#endif
  do_bench_search(f=lower_bound(x,x+n,t,greater<int>()), f && *f == t);
  do_bench_search(f=upper_bound(x,x+n,t,greater<int>()), f && *(f-1) == t);
  do_bench_search(b=binary_search(x,x+n,t,greater<int>()), b);

  cout << "# number: " << n << ",\t" << "type of compare: int" << ",\t" << "type of search: int" << endl;
  do_bench_search(g=find_if(ix,ix+n,ari_i_rpred(x0,&t)), g && x0[*g] == t);
  do_bench_search(g=(int *)lfind_r(&t,ix,&m,sizeof(ix[0]),x0,ari_i_re_r), g && x0[*g] == t);
  do_bench_search(g=(int *)lbsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_i_re_r), g && x0[*g] == t);
  do_bench_search(g=(int *)ubsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_i_re_r), g && x0[*(g-1)] == t);
  do_bench_search(g=(int *)bsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_i_re_r), g && x0[*g] == t);
#if defined(__BLOCKS__)
  do_bench_search(g=(int *)bsearch_b(&t,ix,n,sizeof(ix[0]),ari_i_re_b), g && x0[*g] == t);
#endif
  do_bench_search(g=lower_bound(ix,ix+n,t,ari_i_rkomp(x0)), g && x0[*g] == t);
  do_bench_search(g=upper_bound(ix,ix+n,t,ari_i_rkowp(x0)), g && x0[*(g-1)] == t);

  delete[] x;
  delete[] x0;
  delete[] ix;
}
void markbench_rsearch_double()
{
  int n = N_ELEMENT;
  size_t m = n;
  double *x0 = new double[n], *x = new double[n], t, *f;
  int i, *ix = new int[n], *g;
  bool b;
  typedef ari_rcomp<double> ari_d_rcomp;
  typedef ari_rpred<double> ari_d_rpred;
  typedef ari_rkomp<double> ari_d_rkomp;
  typedef ari_rkowp<double> ari_d_rkowp;
#if defined(__BLOCKS__)
  int (^d_rcmp_b)(const void *a, const void *b) = ^(const void *a, const void *b) {
    return (*(double *)a > *(double *)b) ? -1 : (*(double *)a < *(double *)b) ? 1 : 0;
  };
  int (^ari_d_re_b)(const void *k, const void *i) = ^(const void *k, const void *i) {
    return (*(double *)k > x0[*(int *)i]) ? -1 : (*(double *)k < x0[*(int *)i]) ? 1 : 0;
  };
#endif
  StopWatch sw;

  for (i=0; i<n; i++)
    x[i] = x0[ix[i] = i] = (double)random()/((1<<31)-1);
  t = x0[random()%n];
  sort(x,x+n,greater<double>());
  sort(ix,ix+n,ari_d_rcomp(x0));

  cout << "# number: " << n << ",\t" << "type of compare: double" << ",\t" << "type of search: double" << endl;
  do_bench_search(f=(double *)lfind(&t,x0,&m,sizeof(x0[0]),d_rcmp), f && *f == t);
  do_bench_search(f=(double *)lfind_r(&t,x0,&m,sizeof(x0[0]),0,d_rcmp_r), f && *f == t);
  do_bench_search(f=find(x0,x0+n,t), *f == t);
  do_bench_search(f=(double *)bsearch(&t,x,n,sizeof(x[0]),d_rcmp), f && *f == t);
  do_bench_search(f=(double *)lbsearch_r(&t,x,n,sizeof(x[0]),0,d_rcmp_r), f && *f == t);
  do_bench_search(f=(double *)ubsearch_r(&t,x,n,sizeof(x[0]),0,d_rcmp_r), f && *(f-1) == t);
  do_bench_search(f=(double *)bsearch_r(&t,x,n,sizeof(x[0]),0,d_rcmp_r), f && *f == t);
#if defined(__BLOCKS__)
  do_bench_search(f=(double *)bsearch_b(&t,x,n,sizeof(x[0]),d_rcmp_b), f && *f == t);
#endif
  do_bench_search(f=lower_bound(x,x+n,t,greater<double>()), f && *f == t);
  do_bench_search(f=upper_bound(x,x+n,t,greater<double>()), f && *(f-1) == t);
  do_bench_search(b=binary_search(x,x+n,t,greater<double>()), b);

  cout << "# number: " << n << ",\t" << "type of compare: double" << ",\t" << "type of search: int" << endl;
  do_bench_search(g=find_if(ix,ix+n,ari_d_rpred(x0,&t)), g && x0[*g] == t);
  do_bench_search(g=(int *)lfind_r(&t,ix,&m,sizeof(ix[0]),x0,ari_d_re_r), g && x0[*g] == t);
  do_bench_search(g=(int *)lbsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_d_re_r), g && x0[*g] == t);
  do_bench_search(g=(int *)ubsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_d_re_r), g && x0[*(g-1)] == t);
  do_bench_search(g=(int *)bsearch_r(&t,ix,n,sizeof(ix[0]),x0,ari_d_re_r), g && x0[*g] == t);
#if defined(__BLOCKS__)
  do_bench_search(g=(int *)bsearch_b(&t,ix,n,sizeof(ix[0]),ari_d_re_b), g && x0[*g] == t);
#endif
  do_bench_search(g=lower_bound(ix,ix+n,t,ari_d_rkomp(x0)), g && x0[*g] == t);
  do_bench_search(g=upper_bound(ix,ix+n,t,ari_d_rkowp(x0)), g && x0[*(g-1)] == t);

  delete[] x;
  delete[] x0;
  delete[] ix;
}
#endif /* !defined(DETAIL_BENCH) */

int main()
{
  cout << "number of concurrency:\t" << thread::hardware_concurrency() << endl;

#if !defined(DETAIL_BENCH)
  markbench_sort_char();
  markbench_sort_short();
  markbench_sort_int();
  markbench_sort_double();
  markbench_sort_doubles();
  markbench_sort_array_indice_doubles();
  markbench_sort_in6();
#endif /* !defined(DETAIL_BENCH) */

  markbench_rsort_char();
  markbench_rsort_short();
  markbench_rsort_int();
  markbench_rsort_double();
  markbench_rsort_doubles();
#if !defined(DETAIL_BENCH)
  markbench_rsort_array_indice_doubles();
#endif /* !defined(DETAIL_BENCH) */
  markbench_rsort_in6();

#if !defined(DETAIL_BENCH)
  markbench_partialsort_char();
  markbench_partialsort_short();
  markbench_partialsort_int();
  markbench_partialsort_double();

  markbench_rpartialsort_char();
  markbench_rpartialsort_short();
  markbench_rpartialsort_int();
  markbench_rpartialsort_double();
#endif /* !defined(DETAIL_BENCH) */

#if !defined(DETAIL_BENCH)
  markbench_search_char();
  markbench_search_short();
  markbench_search_int();
  markbench_search_double();

  markbench_rsearch_char();
  markbench_rsearch_short();
  markbench_rsearch_int();
  markbench_rsearch_double();
#endif /* !defined(DETAIL_BENCH) */

  return 0;
}
